
#include "commands.h"
#include "misc.h"

#include <WinUser.h>    //  for VK_



CommandBase * CommandBase::commands_[MAX_COMMANDS];
CommandBase * CommandBase::keys_[256];
int CommandBase::numCommands_;


unsigned int CommandBase::initKeysDefault()
{
  memset( keys_, 0, sizeof( keys_ ) );
  for( int i = 0; i < numCommands_; ++i ) {
    commands_[i]->key_ = 0;
  }
  unsigned int ret = 0;
  //  for each scan code
  for( unsigned int j = 0; j < 256; ++j ) {
    char name[32];
    //  get the appropriate name
    int l = 0;
    unsigned int scanCode = MapVirtualKey( j, 0 );
    if( (l = GetKeyNameText( scanCode<<16, name, 32 )) > 0 ) {
      name[l] = 0;
      //  compare to the commands
      for( int i = 0; i < numCommands_; ++i ) {
        char const * bind = commands_[i]->defBind_;
        if( !stricmp( bind, name ) ) {
          if( keys_[j] != 0 ) {
            //  more than one command bound to the same key!
            ret |= DOUBLE_MAPPED_KEY;
          }
          keys_[j] = commands_[i];
        }
      }
    }
  }
  for( int i = 0; i < numCommands_; ++i ) {
    if( !commands_[i]->key_ ) {
      ret |= UNMAPPED_COMMAND;
    }
  }
  return ret;
}

unsigned int CommandBase::initKeysText( char const * text, size_t size )
{
  for( int i = 0; i < numCommands_; ++i ) {
    commands_[i]->key_ = 0;
  }
  unsigned int ret = 0;
  //  for each scan code
  for( unsigned int j = 0; j < 256; ++j ) {
    char name[32];
    //  get the appropriate name
    //  translate the vcode to scancode
    int l = 0;
    unsigned int scanCode = MapVirtualKey( j, 0 );
    if( (l = GetKeyNameText( scanCode<<16, name, 32 )) > 0 ) {
      name[l] = 0;
      char const * line = text;
    again:
      //  find the appropriate line
      line = strnistr( line, name, size-(line-text) );
      if( !line ) {
        //  this key not bound
        continue;
      }
      if( line != text && line[-1] != '\n' ) {
        //  not beginning of a line
        ++line;
        goto again;
      }
      //  make sure it's followed by optional space, plus a colon
      line += strlen( name );
      while( (line < (text+size)) && (*line == ' ') ) {
        ++line;
      }
      if( *line != ':' ) {
        //  not entire key name
        goto again;
      }
      //  extract the command name, which lives until the end of the line
      char const * colon = line;
      while( line < (text+size) ) {
        if( *line == '\n' ) {
          break;
        }
        else if ( (*colon == ':') || (*colon == ' ') ) {
          ++colon;
        }
        ++line;
      }
      //  compare to the commands
      for( int i = 0; i < numCommands_; ++i ) {
        if( (strlen( commands_[i]->name_ ) == (line-colon)) && 
            !strnicmp( colon, commands_[i]->name_, line-colon ) ) {
          if( keys_[j] != 0 ) {
            //  more than one command bound to the same key!
            ret |= DOUBLE_MAPPED_KEY;
          }
          keys_[j] = commands_[i];
        }
      }
    }
  }
  for( int i = 0; i < numCommands_; ++i ) {
    if( !commands_[i]->key_ ) {
      ret |= UNMAPPED_COMMAND;
    }
  }
  return ret;
}

size_t CommandBase::saveKeysText( char * text, size_t maxSize )
{
  size_t offset = 0;
  size_t realLen = 0;
  for( unsigned int j = 0; j < 256; ++j ) {
    if( !keys_[j] ) {
      continue;
    }
    char name[32];
    if( !GetKeyNameText( MapVirtualKey( j, 0 )<<16, name, 32 ) ) {
      //  unfortunately, I couldn't get a key name for this mapped key...
      assert( !"why is this?" );
      continue;
    }
    char line[128];
    sprintf( line, "%s: %s\n", name, keys_[j]->name_ );
    size_t l = strlen( line );
    if( l+offset <= maxSize ) {
      memcpy( text+offset, line, l );
      offset += l;
    }
    realLen += l;
  }
  return realLen;
}

