

//  networking has to go first, because of WinSock2.h
#include "network.h"
#include "message.h"
#include "clock.h"
#include "sort_addr.h"

#include "glut-start.h"
#include "glut-content.h"
#include "gl-ase.h"
#include "mymath.h"
#include "gl-text.h"
#include "mytweakables.h"
#include "soundmanager.h"
#include "terrain.h"
#include "mysystem.h"
#include "mywheeledcar.h"
#include "mytank.h"
#include "image.h"
#include "misc.h"
#include "mytweakables.h"
#include "myfilesystem.h"
#include "mycamera.h"
#include "myexplodable.h"
#include "mycrate.h"
#include "mygrenade.h"
#include "myparticles.h"
#include "myrocket.h"
#include "mygui.h"
#include "myweapon.h"
#include "myprop.h"
#include "glshaderstate.h"
#include "commands.h"
#include "mymine.h"
#include "mymachinegun.h"
#include "log.h"
#include "scene.h"

#include "ode/ode.h"
#include <gl/glu.h>

#include <stdarg.h> 
#include <vector>
#include <list>
#include <algorithm>
#include <string>


//  set to 1 to drive the car, else you get the tank
#define WHEELED_CAR 0

//  If you want to see things rendered from above (in meters)
#define CAMERA_FROM_ABOVE 0

//  for quick tests of various kinds
#define INVESTIGATE 0

//  Mmm, it's so nice and toasty inside this steaming pile of hacks :-)
//  DONE 1) create a class per Body for callbacks
//  DONE 2) create a class that manages bodies, meshes, joints and geoms
//  DONE 13) separate car into its own class!
//    DONE - turn crates into MySystem
//    DONE - turn camera into MySystem
//    DONE - turn mines into MySystem
//    DONE - turn grenades into MySystem
//    DONE - separate mine from grenade explosion strength
//           by using void* for parameters; make Explodable<>
//    DONE - check out usage of "front" and "right" vs "up"; should
//           the matrix be RightUpBackwards or LeftUpFront?
//    DONE -  Just remove it for now?
//            - turn GrabbedBody into a MySystem deal
//              Possibly turn GrabbedBody into a joint instead of force?
//    DONE - turn car into MySystem -- hard!
//    DONE - what to do about static environment?
//    DONE - use generic MySystem handling
//    DONE - clean up BodyIsTire, IsGrenade, etc (should be done at this point)
//    DONE - remove camera reliance on gPlayer
//    DONE - re-introduce SINK_AMT (rendering offset from body)
//  DONE 24) non-GLUT application driver
//  DONE 29) weapon aim
//    DONE - crosshairs
//    DONE - needs user input for (up/down) aim
//      DONE - needs Win32 based app driver
//  DONE 22) Sky Dome
//    DONE - better texturing
//    - better clouds (and cheaper!)
//  DONE 6) Sound effects
//    DONE - bug where sounds sometimes cut off
//    DONE - engine pitch from wheel angular velocity
//    - tank engine sounds
//    - bug where sound thread appears to eat all CPU
//  26) terrain
//    DONE - LOD
//    DONE - texturing (requires render state updates)
//    - bounds at the edges (requires level definition)
//    - auto-flora
//    DONE - more flexible shader
//    - better art
//    - use description rather than hard-coded (roads texture, etc)
//  27) more cars
//    DONE - ray-wheels tank
//    - helicopter?
//  DONE 34) Select car on start-up
//  DONE 23) pull pieces into separate files
//    DONE - MySystem
//    DONE - wheeled car
//    DONE - camera
//    DONE - grenades
//    DONE - mines
//    DONE - crates
//    DONE - explosion effects
//  DONE 28) data driven asset references
//  30) another kind of weapon
//    DONE - machine gun
//    DONE - rockets
//    - homing missiles
//  39) Provide NETWORKED compile-time switch to remove dependency 
//        on unreleased libintroduce.
//    - make new source release!
//    - make it run on plain GL again!
//  16) octree (use scene.cpp); culling
//  38) Properly factor MySystem updates (no separate lists)
//    - render (octree?)
//    - preStep
//    - postStep
//    - others
//  18) networking
//    - auto-serve
//    - auto-join
//    - spawn points
//    - teams
//    - team display
//    - vehicle position and speeds
//    - interpolation/extrapolation
//    - command forwarding (?)
//    - firing
//    - ammunition
//    - entity sync
//    - tie-breaking
//    - scoring
//  9) something to shoot at
//  8) damage model
//  35) Select terrain on start-up
//    - level representation
//  7) effects when firing
//    DONE - smoke
//    - muzzle flash
//  DONE 36) update everything to GLShader from GWState
//    DONE - update MySystem to take shader instances
//  14) multi-sized rubble
//  DONE 5) tweakables from data (and at runtime?)
//    - don't duplicate into all includers
//    - read vehicle params from XML
//  20) better art
//  37) "rolling" cube reflections for movement
//  31) factor tweakables better
//    - class and object properties?
//  32) particle effects
//    DONE - exhaust
//    DONE - XML description of particle effects
//    DONE - smoke when firing
//    DONE - jeep wheel dust
//    - better explosions
//    - dieOnContact
//    - stopOnContact
//    DONE - speedDampening
//    DONE - speed sensitivity
//  10) vertex arrays in ASE renderer
//    - currently uses display lists -- good enough?
//    - what about animation?
//  11) real sub-materials
//    DONE - ASE sub-material exposure
//    - mesh-with-materials abstraction
//    - cache render states to avoid changes
//  37) Shader support
//    DONE - GLSL loading and linking
//    DONE - some example shaders
//    DONE - built-in variables (g_Time, etc)
//    - shared program merging
//    - state sorting
//    - transparency sorting
//  DONE 12) DDS textures
//  33) Better sound
//    - attenuate with distance
//    - attenuate with occlusion (?)
//    - filter when resampling
//    - filter with distance
//    - add/improve sound effects
//      - launch grenade
//      - engine
//      - block-on-ground
//      - skid
//  17) more camera control
//  21) object layout
//  25) filter when re-sampling sound
//  3) abstract rendering a bit (?)
//  4) separate out text overlay
//  15) Linux port
//  19) make it a game


//todo: move this
MyVehicle * gPlayer;

float gCurWeaponOpacity;
float g_aim;
float g_elevation;

class Blank {
  public:
    void operator()() {}
};

Command< Blank > cmdGas( "Gas", "W" );
Command< Blank > cmdBrake( "Brake", "S" );
Command< Blank > cmdLeft( "Left", "A" );
Command< Blank > cmdRight( "Right", "D" );
Command< Blank > cmdFire( "Fire", "Space" );
Command< Blank > cmdNextWeapon( "Next Weapon", "]" );
Command< Blank > cmdPrevWeapon( "Prev Weapon", "[" );
Command< Blank > cmdRecover( "Recover", "R" );
Command< Blank > cmdGotoCenter( "Goto Center", "=" );
Command< Blank > cmdJump( "Jump", "J" );
Command< Blank > cmdInitCrates( "Init Crates", "I" );
Command< Blank > cmdTeleport( "Teleport", "T" );
Command< Blank > cmdChaseCam( "ChaseCam", "8" );
Command< Blank > cmdFPSCam( "FPSCam", "9" );
Command< Blank > cmdSideCam( "SideCam", "0" );

SoundManager * gSoundManager;
bool gWireframe;





struct PacketHdr {
  unsigned short timeMs;
  unsigned char node;
  unsigned char numPlayers;
};
struct PacketPayloadHdr {
  PacketPayloadHdr() { what = 0; size = 0; }
  unsigned char what;
  unsigned char size;
};
struct PacketServerOffer : public PacketPayloadHdr {
  unsigned char numPlayers;
  char name[15];
};
struct PacketPlayerJoined : public PacketPayloadHdr {
  unsigned short yourTime;
  unsigned short myTime;
  char name[15];
  unsigned char playerNode;
};
struct PacketAckReliable : public PacketPayloadHdr {
  unsigned short yourTime;
};

enum PacketWhat {
  ServerOffer = 128,
  //  there is no RequestToJoin -- that's handled by ServerOffer
  AckReliable = 129,

  //  "reliable" packet types follow below; these are repeated until acked
  PlayerJoined = 192,
};


#define PEER_TIMEOUT 6000
#define SEND_INTERVAL 20

class MyNetwork : public libintroduce::I_SocketUser, public libintroduce::I_SocketPeerUser {
  public:
    MyNetwork()
      {
        clock_ = 0;
        network_ = 0;
        socket_ = 0;
        server_ = 0;
        lastServerTime_ = 0;
        initialTime_ = 0;
        now_ = 0;
        myNode_ = 0;
      }
    static void warnFunc( int level, char const * str, va_list list );
    bool open();
    void step( unsigned long deltaTime );

    virtual void disposing( libintroduce::I_Socket * socket ) { assert(socket == socket_); socket_ = 0; }
    virtual void onConnection( libintroduce::I_Socket * socket, sockaddr_in const & addr, 
        char const * buf, int bSize );

    virtual void disposing( libintroduce::I_SocketPeer * peer ) { assert(peer == server_); server_ = 0; }
    virtual void onMessage( libintroduce::I_SocketPeer *, char const * buf, int size );

    void addBroadcast( PacketPayloadHdr const & pph );
    void maybeBroadcast();
    void addToServer( PacketPayloadHdr const & pph );
    void maybeSendToServer();

    void disconnected();
    void addPlayer( libintroduce::I_Socket * socket, sockaddr_in const & addr, PacketServerOffer const & pso, PacketHdr const & ph );

    void processPlayerJoined( sockaddr_in const & addr, PacketHdr const & ph, libintroduce::Message const & m, PacketPayloadHdr const & pph );
    void processServerOffer( sockaddr_in const & addr, PacketHdr const & ph, libintroduce::Message const & m, PacketPayloadHdr const & pph );

    libintroduce::I_Clock * clock_;
    libintroduce::I_Network * network_;
    libintroduce::I_Socket * socket_;
    libintroduce::I_SocketPeer * server_;

    unsigned long lastServerTime_;
    unsigned long initialTime_;
    unsigned long now_;
    libintroduce::Message pending_;
    unsigned char myNode_;
    sockaddr_in sendAddr_;

    struct ConnectedPeer : public libintroduce::I_SocketPeerUser {

      ConnectedPeer() {
        lastReceive_ = 0;
        lastSend_ = 0;
        peer_ = 0;
        network_ = 0;
      }
      virtual void disposing( libintroduce::I_SocketPeer * peer ) {}
      virtual void onMessage( libintroduce::I_SocketPeer * peer, char const * buf, int bSize );

      void addReliable( PacketPayloadHdr const & pph );
      void addPacket( PacketPayloadHdr const & pph );
      void maybeSend();

      void processAckReliable( PacketHdr const & ph, libintroduce::Message const & m, PacketPayloadHdr const & pph );

      unsigned char id_;
      unsigned long lastReceive_;
      unsigned long lastSend_;
      libintroduce::I_SocketPeer * peer_;
      MyNetwork * network_;
      char name_[16];

      libintroduce::Message pending_;
      struct ReliablePacket {
        libintroduce::Message * msg;
        unsigned short sentAtTimes[8];
      };
      std::list< ReliablePacket > reliableQueue_;
      //  I need a list, so that the address doesn't change
      std::list< libintroduce::Message > reliableMessages_;
    };
    std::list< ConnectedPeer > peers_;
    struct HandlePeer {
      HandlePeer( unsigned long now ) : now_( now ), killPeer_( 0 ) {}
      void operator()( ConnectedPeer & cp ) {
        if( cp.lastReceive_ + PEER_TIMEOUT < now_ ) {
          killPeer_ = cp.id_;
        }
        else if( cp.lastSend_ + SEND_INTERVAL <= now_ ) {
          cp.maybeSend();
        }
      }
      unsigned long now_;
      unsigned char killPeer_;
    };
};




class MyContent : public IGlutContent {
  public:
    MyContent() {
      selectedVehicle_ = 0xffffffff;
      stepTimer_ = 3;
      clickedButtons_ = 0;
      networked_ = false;
      network_ = 0;
      scene_ = 0;
    }
    virtual void createScene();
    virtual void createScene2( size_t vehicle );
    virtual void step( float deltaTime );
    //  key() means some text was typed
    virtual void key( int x, int y, unsigned char key );
    //  special() gets key codes (VK_...) which can be both down and up
    virtual void special( int x, int y, int key, bool down );

    virtual void beginDrag( int x, int y, unsigned int buttons );
    virtual void dragMove( int x, int y );
    virtual void endDrag( int x, int y );
    virtual void mouseMove( int x, int y );

    void setSelectedVehicle( size_t s ) { 
      selectedVehicle_ = s;
      stepTimer_ = 3;
    }
    void setNetworked( bool n ) {
      networked_ = n;
    }
    void setupNetwork();
    size_t selectedVehicle_;
    size_t stepTimer_;
    unsigned int clickedButtons_;
    bool networked_;

    MyNetwork * network_;
    I_Scene * scene_;
    Vector4 frustumPlanes_[6];   //  left, right, bottom, top, near, far
};
MyContent * myContent;


IGlutContent * MakeGlutContent()
{
  return (myContent = new MyContent());
}



void MyNetwork::warnFunc( int level, char const * str, va_list list )
{
  char buf[2048];
  _vsnprintf( buf, 2048, str, list );
  buf[2047] = 0;
  switch( level ) {
  case WL_WARNING:
    warning( "%s", buf );
    break;
  case WL_LOG:
    info( "%s", buf );
    break;
  default:
    debug( "%s", buf );
    break;
  }
}

#define MY_PORT 56785
#define SERVER_BROADCAST_INTERVAL 3000

bool MyNetwork::open()
{
  clock_ = libintroduce::CreateClock();
  char error[128];
  error[0] = 0;
  network_ = libintroduce::CreateNetwork( error, 128, warnFunc );
  if( error[0] ) {
    ::error( "Could not open network: %s\n", error );
  }
  if( !network_ ) {
    return false;
  }
  socket_ = network_->createSocket( MY_PORT, this );
  if( !socket_ ) {
    ::error( "Network socket creation failed.\n" );
    return false;
  }
  socket_->setSocketUser( this );
  memset( &sendAddr_, 0, sizeof( sendAddr_ ) );

  server_ = 0;
  lastServerTime_ = 0;
  initialTime_ = lastServerTime_;
  return true;
}

void MyNetwork::step( unsigned long deltaTime )
{
  now_ += deltaTime;

  network_->step(0);

  if( (now_ > lastServerTime_ + SERVER_BROADCAST_INTERVAL) && !server_ ) {
    //  say that I can be a server
    PacketServerOffer pso;
    pso.what = ServerOffer;
    pso.size = sizeof( pso );
    _snprintf( pso.name, sizeof( pso.name ), UserName() );
    pso.name[sizeof(pso.name)-1] = 0;
    pso.numPlayers = (unsigned char)(peers_.size() & 0xff);
    addBroadcast( pso );
    lastServerTime_ = now_;
  }

  HandlePeer hp( now_ );
  std::for_each( peers_.begin(), peers_.end(), hp );
  if( hp.killPeer_ != 0 ) {
    std::list< ConnectedPeer >::iterator i = peers_.begin();
    while( i != peers_.end() ) {
      if( (*i).id_ == hp.killPeer_ ) {
        info( "Timeout for peer %d (%s) @ %d\n", (*i).id_, (*i).name_, now_ );
        peers_.erase( i );
        break;
      }
      ++i;
    }
  }
  maybeBroadcast();
  maybeSendToServer();

  if( server_ && (lastServerTime_ + PEER_TIMEOUT < now_) ) {
    disconnected();
  }
}

void MyNetwork::disconnected()
{
  info( "Disconnected from server @ %d.\n", now_ );
  server_->dispose();
  myNode_ = 0;
}

void MyNetwork::addToServer( PacketPayloadHdr const & pph )
{
  assert( server_ );
  assert( pph.what < 192 );     //  can't send-to-server reliable packets
  assert( pph.size > 0 );
  if( (size_t)pending_.size_ > MAX_MESSAGE_SIZE-pph.size-sizeof(PacketHdr) ) {
    //  Flush the data we already have pending.
    maybeSendToServer();
    //  IF I did reliable messaging, then it would be a problem that 
    //  I could send more than one packet with the same timestamp...
    assert( pending_.size_ == 0 );
  }
  pending_.putn( pph.size, &pph );
}

void MyNetwork::maybeSendToServer()
{
  if( pending_.size_ > 0 ) {
    PacketHdr ph;
    ph.timeMs = unsigned short( ((unsigned long) (now_-initialTime_)) & 0xffff );
    ph.node = myNode_;
    ph.numPlayers = (unsigned char)peers_.size();
    pending_.prepend( ph );
    server_->sendMessage( pending_.buf_, pending_.size_ );
    pending_.clear();
  }
  assert( pending_.size_ == 0 );
}

void MyNetwork::addBroadcast( PacketPayloadHdr const & pph )
{
  assert( !server_ );
  assert( pph.what < 192 );   //  can't broadcast reliable messages
  assert( pph.size > 0 );
  if( (size_t)pending_.size_ >= MAX_MESSAGE_SIZE-pph.size-sizeof(PacketHdr) ) {
    maybeBroadcast();
  }
  pending_.putn( pph.size, &pph );
}

void MyNetwork::maybeBroadcast()
{
  if( pending_.size_ > 0 ) {
    sockaddr_in sin = sendAddr_;
    if( !(SIN_ADDR_UINT(sin.sin_addr)) ) {
      sin = socket_->address();
      unsigned long laddr = htonl( SIN_ADDR_UINT( sin.sin_addr ) );
      if( laddr == 0 ) {
        laddr = 0xffffffff;
      }
      else if( laddr < 128 ) {
        //  class A broadcast
        laddr |= 0xffffff;
      }
      else if( laddr < 192 ) {
        //  class B broadcast
        laddr |= 0xffff;
      }
      else {
        //  class C broadcast
        laddr |= 0xff;
      }
      SIN_ADDR_UINT(sin.sin_addr) = laddr;
      sendAddr_ = sin;
    }

    PacketHdr ph;
    ph.timeMs = unsigned short( ((unsigned long) (now_-initialTime_)) & 0xffff );
    ph.node = myNode_;
    ph.numPlayers = (unsigned char)peers_.size();
    pending_.prepend( ph );
    socket_->sendMessage( sin, pending_.buf_, pending_.size_ );
    pending_.clear();
  }
  assert( pending_.size_ == 0 );
}

void MyNetwork::addPlayer( libintroduce::I_Socket * socket, sockaddr_in const & addr, PacketServerOffer const & pso, PacketHdr const & ph )
{
  //  find an un-used node id
  unsigned int id = 1;
  std::list< ConnectedPeer >::iterator i = peers_.begin();
  while( i != peers_.end() ) {
    if( (*i).id_ > id ) {
      //  this is an un-used id!
      break;
    }
    ++id;
    ++i;
  }
  if( id >= 255 ) {
    //  no space!
    info( "Wanted to add player, but no space! @ %d\n", now_ );
  }
  else {

    PacketPlayerJoined ppj;
    ppj.what = PlayerJoined;
    ppj.size = sizeof( ppj );
    ppj.yourTime = ph.timeMs;
    ppj.myTime = (unsigned short)(now_ & 0xffff);
    strncpy( ppj.name, pso.name, sizeof(ppj.name) );
    ppj.name[sizeof(ppj.name)-1] = 0;
    ppj.playerNode = id;

    peers_.insert( i, ConnectedPeer() );
    --i;
    (*i).id_ = id;
    (*i).lastReceive_ = now_;
    (*i).network_ = this;
    (*i).peer_ = socket->createPeer( addr, &(*i) );
    strncpy( (*i).name_, ppj.name, sizeof((*i).name_) );
    (*i).name_[sizeof((*i).name_)-1] = 0;

    (*i).addReliable( ppj );

    info( "Player %s added from %s @ %d\n", ppj.name, 
        libintroduce::sort_sockaddr_in(addr).staticText(), now_ );
  }
}

void MyNetwork::processPlayerJoined( sockaddr_in const & addr, PacketHdr const & ph, libintroduce::Message const & m, PacketPayloadHdr const & pph )
{
  //  now, this means I'm not the server...
  if( peers_.size() ) {
    error( "Received PlayerJoined message from %s (node %d), but I'm a server!\n",
        libintroduce::sort_sockaddr_in( addr ).staticText(), ph.node );
    return;
  }
  PacketPlayerJoined ppj;
  if( pph.size != sizeof( ppj ) ) {
    //  wrong version
    warning( "PlayerJoined packet has wrong version! (size %d but wanted %d)\n", pph.size, sizeof(ppj) );
  }
  else {
    memcpy( &ppj, m.buf_+m.base_-sizeof(pph), sizeof(ppj) );
    ppj.name[sizeof(ppj.name)-1] = 0;
    info( "Player '%s' joined (node %d, addr %s)\n",
        ppj.name, ppj.playerNode, libintroduce::sort_sockaddr_in( addr ).staticText() );
    //  Only join if there's 256 ms or less of latency!
    //  If there are two players with the same name, their clocks will typically be 
    //  not in sync, and thus we'll have a 255/256 chance of breaking a collision 
    //  by looking at the clock like this.
    if( !strcmp( ppj.name, UserName() ) && (((unsigned short)now_ - ppj.yourTime) < 256) ) {
      //  This is me!
      info( "This is me (now %d, yourTime %d)\n", (unsigned short)now_, ppj.yourTime );

      if( server_ && (libintroduce::sort_sockaddr_in(addr) == server_->peerAddress()) ) {
        //  This is a second server that tries to join me
        char srv1[64]; strcpy( srv1, libintroduce::sort_sockaddr_in(addr).staticText() );
        char srv2[64]; strcpy( srv2, libintroduce::sort_sockaddr_in(server_->peerAddress()).staticText() );
        info( "There are two servers on this net! (%s and %s)\n", srv1, srv2 );
      }
      else {
        if( server_ == 0 ) {
          server_ = socket_->createPeer( addr, this );
          myNode_ = ppj.playerNode;
        }
      }
    }
  }
}

void MyNetwork::processServerOffer( sockaddr_in const & addr, PacketHdr const & ph, libintroduce::Message const & m, PacketPayloadHdr const & pph )
{
  PacketServerOffer pso;
  if( pph.size != sizeof( pso ) ) {
    warning( "ServerOffer packet has wrong version! (size %d wanted %d)\n", pph.size, sizeof(pso) );
    //  wrong version
  }
  else {
    memcpy( &pso, m.buf_+m.base_-sizeof(pph), sizeof(pso) );
    pso.name[sizeof(pso.name)-1] = 0;
    //  seven cases:
    //  0) I'm joined somewhere
    //      -> no change
    //  1) I have players, and he has no players
    //      -> he joins me
    //  2) I have no players, and he has players
    //      -> I join him
    //  3) We both have no players
    //      if my name < his name
    //        -> he joins me
    //      if my name > his name
    //        -> I join him
    //      if names equal
    //        -> no change
    //  4) We both have players
    //      -> no change
    debug( "ServerOffer: %s (%d players) (I'm %s %d players)\n",
        pso.name, pso.numPlayers, UserName(), peers_.size() );
    bool imJoined = (server_ != 0);
    bool iHavePlayers = (peers_.size() != 0);
    bool heHasPlayers = (pso.numPlayers != 0);
    int myNameCompare = stricmp( UserName(), pso.name );
    if( imJoined ) {
      //  no change
    }
    else if( iHavePlayers && !heHasPlayers ) {
      //  he joins me
      addPlayer( socket_, addr, pso, ph );
    }
    else if( !iHavePlayers && heHasPlayers ) {
      //  I join him -- this will happen automatically
    }
    else if( !iHavePlayers && !heHasPlayers ) {
      if( myNameCompare < 0 ) {
        //  he joins me
        addPlayer( socket_, addr, pso, ph );
      }
      else if( myNameCompare > 0 ) {
        //  I join him -- this will happen automatically
      }
      else {
        //  no change -- same name!
      }
    }
    else {
      assert( iHavePlayers && heHasPlayers );
      //  no change
    }
  }
}

void MyNetwork::onConnection( libintroduce::I_Socket * socket, sockaddr_in const & addr, char const * buf, int bSize )
{
  //  The fact that the data ends up in onConnection() means I don't (yet) have 
  //  a peer from this guy. This may be becuase I'm a server, or because I'm not 
  //  yet joined to a game, or because there's a second game going on the same 
  //  network segment.

  if( server_ ) {
    //  This is from some other server on the net than 
    //  the guy I picked!
    char srv1[64]; strcpy( srv1, libintroduce::sort_sockaddr_in(addr).staticText() );
    char srv2[64]; strcpy( srv2, libintroduce::sort_sockaddr_in(server_->peerAddress()).staticText() );
    debug( "onConnection() from %s; I'm already connected to %s.\n", srv1, srv2 );
    return;
  }

  debug( "onConnection() from %s.\n", libintroduce::sort_sockaddr_in( addr ).staticText() );

  libintroduce::Message m( buf, bSize );
  PacketHdr ph;
  if( !m.get( ph) ) {
    return;
  }
  if( (ph.node == 0) && (peers_.size() != 0) && (ph.numPlayers != 0) ) {
    //  is this from another server with players? (or myself!)
    return;
  }
  PacketPayloadHdr pph;
  while( m.get( pph ) ) {
    switch( pph.what ) {
    case ServerOffer: 
      processServerOffer( addr, ph, m, pph );
      break;
    case PlayerJoined: 
      processPlayerJoined( addr, ph, m, pph );
      break;
    default:
      //  Typically, I'll end up here if I drop a client, who then comes back.
      //  I'll just let that client time out and re-connect.
      break;
    }
    m.base_ += pph.size - sizeof(pph);
  }
}

void MyNetwork::onMessage( libintroduce::I_SocketPeer * peer, char const * buf, int bSize )
{
  //  I'm a slave
  debug( "slave onMessage() from %s.\n", libintroduce::sort_sockaddr_in( peer->peerAddress() ).staticText() );
  assert( server_ );
  assert( peers_.size() == 0 );

  lastServerTime_ = now_;

  libintroduce::Message m( buf, bSize );
  PacketHdr ph;
  if( !m.get( ph) ) {
    return;
  }
  PacketPayloadHdr pph;
  unsigned char mask = 0;
  while( m.get( pph ) ) {
    mask |= pph.what;
    switch( pph.what ) {
    case PlayerJoined:
      //  the server may have missed our previous ack
      processPlayerJoined( peer->peerAddress(), ph, m, pph );
      break;
    default:
      break;
    }
  }
  if( (mask & 192) == 192 ) {
    //  ack whatever reliable messages were in there
    PacketAckReliable par;
    par.what = AckReliable;
    par.size = sizeof(par);
    par.yourTime = ph.timeMs;
    addToServer( par );
  }
}

void MyNetwork::ConnectedPeer::addReliable( PacketPayloadHdr const & pph )
{
  debug( "addReliable( %d:%d ): queue size was %d\n", pph.what, pph.size, reliableQueue_.size() );
  assert( pph.what >= 192 );   //  can only make reliable messages reliable
  assert( pph.size > 0 );
  reliableMessages_.push_back( libintroduce::Message( &pph, pph.size ) );
  ReliablePacket rp;
  memset( &rp.sentAtTimes, 0, sizeof( rp.sentAtTimes ) );
  rp.msg = &reliableMessages_.back();
  reliableQueue_.push_back( rp );
}

void MyNetwork::ConnectedPeer::addPacket( PacketPayloadHdr const & pph )
{
  debug( "addPacket( %d:%d )\n", pph.what, pph.size );
  assert( pph.what < 192 );   //  can't send reliable messages unreliably
  assert( pph.size > 0 );
  if( (size_t)pending_.size_ >= MAX_MESSAGE_SIZE-pph.size-sizeof(PacketHdr) ) {
    //  flush the queue of unreliable messages
    maybeSend();
    assert( pending_.size_ == 0 );
  }
  pending_.putn( pph.size, &pph );
}

void MyNetwork::ConnectedPeer::maybeSend()
{
  if( pending_.size_ > 0 || reliableQueue_.size() ) {
    PacketHdr ph;
    ph.node = id_;
    ph.numPlayers = 0;
    ph.timeMs = unsigned short(network_->now_ & 0xffff);
    pending_.prepend( ph );

    //  fill up pending message with reliable data waiting
    size_t n = reliableQueue_.size();
    if( n > 0 ) {
      while( n > 0 ) {
        if( !pending_.putn( reliableQueue_.front().msg->size_, reliableQueue_.front().msg->buf_ ) ) {
          //  couldn't fit more now -- let this possibly big boy be first 
          //  next time.
          break;
        }
        --n;
        ReliablePacket & p = reliableQueue_.front();
        assert( sizeof(p.sentAtTimes[0]) == sizeof(unsigned short) );
        assert( sizeof(p.sentAtTimes) == sizeof(unsigned short)*8 );
        if( p.sentAtTimes[7] != 0 ) {
          //  Replace some random instance of the already-sent timestamps. 
          //  If RTT is longer than 8 send intervals, I could never get a 
          //  message out, and an ack back, otherwise. Now, it just becomes 
          //  more likely that unnecessary re-transmits happen, the longer 
          //  the latency is. Ideally, RTT < 8 * send interval, so with 
          //  send interval 50 ms (broadband), RTT < 400 ms; with send 
          //  interval 20 ms (LAN), RTT < 160 ms; with send interval 100 ms 
          //  (modems), RTT < 800 ms.
          memmove( &p.sentAtTimes[1], p.sentAtTimes, sizeof(unsigned short)*(rand()&7) );
        }
        else {
          memmove( &p.sentAtTimes[1], p.sentAtTimes, sizeof(unsigned short)*7 );
          debug( "sent reliable packet %d:%d to %s @ %d\n", p.msg->buf_[0], p.msg->buf_[1], 
              name_, network_->now_ );
        }
        p.sentAtTimes[0] = ph.timeMs;
        reliableQueue_.push_back( p );
        reliableQueue_.pop_front();
      }
      //  If this isn't true, we're sending enough unreliable stuff that 
      //  there's no space for even a single reliable packet -- that's 
      //  bad, as the reliable packet would never get sent!
      assert( n < reliableQueue_.size() );
    }

    peer_->sendMessage( pending_.buf_, pending_.size_ );
    pending_.clear();

    lastSend_ = ph.timeMs;
  }
}

void MyNetwork::ConnectedPeer::processAckReliable( PacketHdr const & ph, libintroduce::Message const & m, PacketPayloadHdr const & pph )
{
  PacketAckReliable par;
  if( pph.size != sizeof( par ) ) {
    //  wrong version
    warning( "AckReliable packet has wrong version! (size %d but wanted %d)\n", pph.size, sizeof(par) );
  }
  else {
    memcpy( &par, m.buf_+m.base_-sizeof(pph), sizeof(par) );
    debug( "AckReliable (yourTime %d)\n", par.yourTime );
    bool removed = false;
    if( par.yourTime != 0 ) {
      std::list< ReliablePacket >::iterator ptr = reliableQueue_.begin();
      while( ptr != reliableQueue_.end() ) {
        std::list< ReliablePacket >::iterator next = ptr;
        ++next;
        ReliablePacket const & rp = (*ptr);
        for( int i = 0; i < 8; ++i ) {
          if( rp.sentAtTimes[i] == par.yourTime ) {
            libintroduce::Message * m = rp.msg;
            debug( "Removed reliable packet %d:%d for %s @ %d (tag %d)\n", m->buf_[0], m->buf_[1], name_,
                network_->now_, par.yourTime );
            removed = true;
            for( std::list< libintroduce::Message >::iterator i = reliableMessages_.begin(); i != reliableMessages_.end(); ++i ) {
              if( &(*i) == m ) {
                reliableMessages_.erase( i );
                break;
              }
            }
            reliableQueue_.erase( ptr );
            break;
          }
        }
        ptr = next;
      }
    }
    if( !removed ) {
      removed = true;
    }
  }
}

void MyNetwork::ConnectedPeer::onMessage( libintroduce::I_SocketPeer * peer, char const * buf, int bSize )
{
  //  I'm a server
  debug( "server onMessage() from %s.\n", libintroduce::sort_sockaddr_in( peer->peerAddress() ).staticText() );
  assert( !network_->server_ );
  assert( network_->peers_.size() );

  //  remember that we saw something from this guy
  lastReceive_ = network_->now_;

  libintroduce::Message m( buf, bSize );
  PacketHdr ph;
  if( !m.get( ph) ) {
    return;
  }
  PacketPayloadHdr pph;
  while( m.get( pph ) ) {
    switch( pph.what ) {
    case AckReliable:
      //  the server may have missed our previous ack
      processAckReliable( ph, m, pph );
      break;
    default:
      debug( "Peer message %d:%d from %s ignored\n", pph.what, pph.size, 
          libintroduce::sort_sockaddr_in(peer->peerAddress()).staticText() );
      break;
    }
    m.base_ += pph.size-sizeof(pph);
  }
}


static I_GL_ASE * chassisModel;
static I_GL_ASE * tankModel;
static I_GL_ASE * tireModel;
//static I_GL_ASE * terrain;
static I_Terrain * terrain;
static GlText * text;
static char debugText[ 16 ][ 64 ];
static bool debugDirty;
static bool overlayVisible;

static GLShader * groundShader;
static GLShader * crateShader;
static GLShader * carShader;
static GLShader * wheelShader;
static GLShader * explosionShader;
static GLShader * textShader;
static GLShader * bgShader;

static I_GL_ASE * skyMesh;
static GLShader * skyShader;


GLShaderFactory * gShaderFactory;

namespace {
  std::list< GuiRoot * > gGuis;
  void RenderGui( GuiRoot * r ) {
    r->render();
  }
  //  start out grabbed!
  bool wantMouseGrabbed = true;
}


using namespace FS;


char const * imagePath( char const * handle )
{
  static char path[ 256 ];
  if( !translateHandle( handle, path, sizeof( path ) ) ) {
    warning( "Missing image handle value in filemap: '%s'\n", handle );
    return "media/missing.dds";
  }
  assert( (strlen( path ) > 4) && (!strcmp( &path[strlen(path)-4], ".tga" ) || !strcmp( &path[strlen(path)-4], ".dds" )) );
  return path;
}

char const * asePath( char const * handle )
{
  static char path[ 256 ];
  if( !translateHandle( handle, path, sizeof( path ) ) ) {
    warning( "Missing ASE handle value in filemap: '%s'\n", handle );
    return "media/missing.ase";
  }
  assert( (strlen( path ) > 4) && !strcmp( &path[strlen(path)-4], ".ase" ) );
  return path;
}

char const * wavPath( char const * handle )
{
  static char path[ 256 ];
  if( !translateHandle( handle, path, sizeof( path ) ) ) {
    warning( "Missing WAV handle value in filemap: '%s'\n", handle );
    return "media/missing.wav";
  }
  assert( (strlen( path ) > 4) && !strcmp( &path[strlen(path)-4], ".wav" ) );
  return path;
}






namespace {

  GuiRoot * selectGui;
  GuiRoot * pleaseWait;

  class Selected : public MyButtonCallback {
    public:
      size_t selected_;
      GuiWidget * gw_;
      GuiWidget * networked_;
      Selected() : selected_( 0 ), gw_( 0 ) {}
      void onButtonClicked( GuiWidget * w, uint ) {
        gw_->getData( &selected_, sizeof( selected_ ) );
        selectGui->dispose();
        std::list< GuiRoot * >::iterator ptr = std::find( gGuis.begin(), gGuis.end(), selectGui );
        gGuis.erase( ptr );
        selectGui = 0;
        myContent->setSelectedVehicle( selected_ );
        bool nw = false;
        networked_->getData( &nw, sizeof( nw ) );
        myContent->setNetworked( nw );
      }
  };

  Selected gSelected;

  void ShowSelectGui()
  {
    if( selectGui ) {
      return;
    }
    WidgetSize ws = { uint(WINWIDTH/2-200), uint(WINHEIGHT/3), uint(WINWIDTH/2+200), uint(WINHEIGHT/3+240) };
    selectGui = NewGui( ws );

    Image jeep;
    loadTgaImage( imagePath( "jeep-image" ), &jeep );

    Image tank;
    loadTgaImage( imagePath( "tank-image" ), &tank );

    WidgetSize choice = { 36, 10, 200+36+128, 10+128+16+6 };
    Image * images[2] = { &jeep, &tank };
    GuiWidget * w = NewImageSelectorWidget( "Select a vehicle:", 
        (Image const **)images, 2 );
    selectGui->add( w, choice );
    gSelected.gw_ = w;

    WidgetSize checkbox = { 20, 213, 130, 227 };
    GuiWidget * network = NewCheckboxWidget( "Networked" );
    selectGui->add( network, checkbox );
    gSelected.networked_ = network;

    WidgetSize button = { 300, 210, 380, 230 };
    GuiWidget * ok = NewButton( "OK", &gSelected );
    selectGui->add( ok, button );

    gGuis.push_back( selectGui );
  }

  void ShowPleaseWait()
  {
    if( pleaseWait ) {
      return;
    }
    WidgetSize ws = { uint(WINWIDTH/2-100), uint(WINHEIGHT/3), uint(WINWIDTH/2+100), uint(WINHEIGHT/3+140) };
    pleaseWait = NewGui( ws );

    GuiWidget * w = NewStaticTextWidget( "Please Wait...", 1, GWColor( 0xffffffff ) );
    WidgetSize pwRect = { 20, 20, ws.right-ws.left-40, ws.bottom-ws.top-40 };
    pleaseWait->add( w, pwRect );

    gGuis.push_back( pleaseWait );
  }

  void HidePleaseWait()
  {
    pleaseWait->dispose();
    std::list< GuiRoot * >::iterator ptr = std::find( gGuis.begin(), gGuis.end(), pleaseWait );
    gGuis.erase( ptr );
    pleaseWait = 0;
  }
}







void debugVar( int x, int y, char const * fmt, ... )
{
  if( FPS_OVERLAY > 0 ) {
    assert( x < 64 );
    assert( y < 16 );
    char buf[ 256 ];
    va_list vl;
    va_start( vl, fmt );
    _vsnprintf( buf, 256, fmt, vl );
    va_end( vl );
    buf[255] = 0;
    size_t l = strlen( buf );
    if( l + x > 64 ) {
      l = 64 - x;
    }
    if( memcmp( &debugText[y][x], buf, l ) ) {
      memcpy( &debugText[y][x], buf, l );
      debugDirty = true;
    }
  }
}

I_GL_ASE * ASEFromFile( char const * path )
{
  //todo: mesh sharing w/materials
  FILE * f = fopen( path, "rb" );
  assert( f );
  fseek( f, 0, 2 );
  long l = ftell( f );
  fseek( f, 0, 0 );
  char * p = new char[ l ];
  fread( p, 1, l, f );
  fclose( f );
  char const * sl = strrchr( path, '\\' );
  if( !sl ) sl = strrchr( path, '/' );
  if( !sl ) sl = path; else ++sl;
  I_GL_ASE * ret = MakeGL_ASE( sl, p, l, 0.01f );
  assert( ret );
  delete[] p;
  return ret;
}


static double curTime;
static dWorldID gWorld;
static dJointGroupID gTempGroup;
static dSpaceID gWorldSpace;
static MyCamera * camera;
static float g_turn;
static float g_gas;
//static dTriMeshDataID terrainData;
static dGeomID terrainGeom;
static dGeomID utilRay;
static dGeomID utilSphere;
std::vector< MySystem * > crates_;
std::vector< MyProp > trees_;

//todo: fold this into MySystem support
std::vector< MySystem * > deferredDelete_;


void AddDeferredDelete( MySystem * sys )
{
  assert( std::find( deferredDelete_.begin(), deferredDelete_.end(), sys ) == deferredDelete_.end() );
  deferredDelete_.push_back( sys );
}


void ClearForceAndTorque( dBodyID body )
{
  dBodySetTorque( body, 0, 0, 0 );
  dBodySetForce( body, 0, 0, 0 );
  dBodySetLinearVel( body, 0, 0, 0 );
  dBodySetAngularVel( body, 0, 0, 0 );
}






static void MoveCarBy( Vector3 const & dp )
{
  gPlayer->moveBy( dp.x, dp.y, dp.z );
}

static void MoveCarBy( float dx, float dy, float dz )
{
  gPlayer->moveBy( dx, dy, dz );
}

static void MoveAndOrientCar( Vector3 const & pos, Matrix4 const & ori )
{
  Vector3 delta = pos - gPlayer->P();
  for( size_t i = 0; i < gPlayer->bodyCount(); ++i ) {
    Vector3 nu = gPlayer->P(i) + delta;
    dBodyID b = gPlayer->bodyAt(i);
    dBodySetPosition( b, nu.x, nu.y, nu.z );
    dBodySetRotation( b, (float const *)&ori );
    dBodySetAngularVel( b, 0, 0, 0 );
    dBodySetLinearVel( b, 0, 0, 0 );
  }
}

Vector3 RayCastAgainstWorld( Vector3 const & from )
{
  dGeomRaySet( utilRay, from.x, from.y, from.z, 0, -1, 0 );
  dGeomRaySetLength( utilRay, 500 );
  dContactGeom c;
  if( !dCollide( utilRay, terrainGeom, 1, &c, sizeof( c ) ) ) {
    return from;
  }
  return Vector3( from.x, from.y - c.depth, from.z );
}


namespace {
  class InitCrate {
    public:
      virtual void operator()( MySystem::BodyInfo const & bi ) = 0;
  };
  class InitCrateStack : public InitCrate {
    public:
      int zpos;
      int ypos;
      int ix;
      int cdown;
      InitCrateStack() {
        ix = 0; cdown = 0; zpos = 0; ypos = 0;
      }
      void operator()( MySystem::BodyInfo const & bi )
        {
          dBodyID body = bi.body;
          dBodyEnable( body );
          dBodySetPosition( body, 30, ypos*1.01f+0.505f, -zpos*1.01f-ypos*0.505f );
          dBodySetRotation( body, &Matrix3::Identity.xx );
          ClearForceAndTorque( body );
          if( zpos == 0 ) {
            ++cdown;
            ypos = 0;
            zpos = cdown;
          }
          else {
            --zpos;
            ++ypos;
          }
          ++ix;
        }
  };
  class InitCrateRandom : public InitCrate {
    public:
      enum {
        cWidth = 8,
        cDepth = 5,
      };
      bool occupied[ cWidth ][ cDepth ];
      float curHeight;
      int numPlaced;
      static int seed;
      InitCrateRandom( int s = 0 )
        {
          if( s ) {
            seed = s;
          }
          else {
            ++seed;
          }
          srand( seed );
          Vector3 v = RayCastAgainstWorld( Vector3( 20, 100, -10 ) );
          initLevel( 0.5f + v.y + 2 );
        }
      void operator()( MySystem::BodyInfo const & bi )
        {
          if( numPlaced == int( cWidth * cDepth * 0.4 ) ) {
            debug( "numPlaced is %d; moving up from %.1f\n", numPlaced, curHeight );
            initLevel( curHeight + 1.01f );
          }
          int nTries = 0;
        again:
          int x = (rand() & 0x7fff) * (cWidth - 1) / 0x8000;
          int y = (rand() & 0x7fff) * (cDepth - 1) / 0x8000;
          int dx = 0;
          int dy = 1;
          if( rand() & 8 ) {
            dy = 0; dx = 1;
          }
          if( occupied[x][y] || occupied[x+dx][y+dy] ) {
            if( ++nTries < 20 ) {
              goto again;
            }
            debug( "Could not place block with 20 tries; moving up from %.1f\n", curHeight );
            initLevel( curHeight + 1.01f );
            nTries = 0;
            goto again;
          }
          occupied[x][y] = true;
          occupied[x+dx][y+dy] = true;
          dBodyID body = bi.body;
          dBodyEnable( body );
          dBodySetPosition( body, x*1.01f + 0.5f*dx + 20, curHeight, y*1.01f + 0.5f*dy - 10 );
          if( dy == 1 ) {
            static Matrix3 rotY( Vector3( 0, 0, -1 ), Vector3( 0, 1, 0 ), Vector3( 1, 0, 0 ) );
            dBodySetRotation( body, &rotY.xx );
          }
          else {
            dBodySetRotation( body, &Matrix3::Identity.xx );
          }
          ClearForceAndTorque( body );
          ++numPlaced;
        }
      void initLevel( float ch )
        {
          curHeight = ch;
          memset( occupied, 0, sizeof( occupied ) );
          numPlaced = 0;
        }
  };
  int InitCrateRandom::seed;
  class InitCrate2 {
      InitCrate * ic_;
    public:
      InitCrate2( InitCrate * ic ) : ic_( ic ) {}
      void operator()( MySystem::BodyInfo const & bi )
        {
          (*ic_)( bi );
        }
  };
}
void InitCratePositions()
{
  //-InitCrateStack ic;
  //InitCrateRandom ic;
  InitCrateRandom ic( 3 );
  InitCrate2 ic2( &ic );
  for( size_t ix = 0; ix < crates_.size(); ++ix ) {
    crates_[ix]->each_body( ic2 );
  }
}


void CreateForest( I_Scene * scene )
{
  I_GL_ASE * tree = ASEFromFile( asePath( "appletree_ase" ) );
  GLShader * treeShader = gShaderFactory->newShader( "treeShader" );

  unsigned int nToPlace = 0;
  Vector3 placePos;
  unsigned int nTrees = (unsigned int)TERRAIN_TREES;
  trees_.resize( 120 /*nTrees*/ );

  uint numVertices = tree->numVertices();
  uint numTriangles = tree->numTriangles();
  Vector3 * verts = new Vector3[ numVertices ];
  int * inds = new int[ numTriangles*3 ];
  tree->getMesh( &verts[0].x, numVertices, inds, numTriangles );
  dTriMeshDataID tmdid = dGeomTriMeshDataCreate();
  dGeomTriMeshDataBuildSingle( tmdid, verts, sizeof( Vector3 ), numVertices, 
      inds, numTriangles*3, sizeof(uint)*3 );

  uint i = 0;
  for( float x = -50; x < 55; x += 10 ) {
    for( float y = -50; y < 55; y += 10 ) {
      if( fabsf(x) < 1 && fabsf(y) < 1 ) {
        continue;
      }
      Vector3 actualPos = RayCastAgainstWorld( Vector3( x, 200, y ) );
      Matrix4 m4 = Matrix4::Identity;
      m4.pos() = actualPos;
      dGeomID geom = dCreateTriMesh( 0, tmdid, 0, 0, 0 );
      //  trees don't collide with other trees, or ground
      dGeomSetCollideBits( geom, ~1 );
      dGeomSetCategoryBits( geom, 1 );
      dGeomSetPosition( geom, actualPos.x, actualPos.y, actualPos.z );
      dSpaceAdd( gWorldSpace, geom );
      trees_[i].init( tree, treeShader, m4, scene );
      ++i;
    }
  }  
#if 0
  for( uint i = 0; i < nTrees; i++ ) {
    if( !nToPlace ) {
      nToPlace = 3 + (rand() & 0x7fff) * 20 / 0x8000;
nu_rand:
      placePos = Vector3( (rand() & 0x7fff) * 1001/float(0x8000) - 500, 0, (rand() & 0x7fff) * 1001/float(0x8000) - 500 );
    }
    if( placePos.lengthSquared() < 400 ) {
      goto nu_rand;
    }
    --nToPlace;
    Vector3 actualPos = RayCastAgainstWorld( placePos + Vector3( 0, 200, 0 ) );
    Matrix4 m4 = Matrix4::Identity;
    dRFromAxisAndAngle( &m4.xx, 0, 1, 0, (rand() & 0x7fff) * 6.28f / 0x8000 );
    Matrix3 m3 = m4;
    m4 = m3.transpose();
    m4.pos() = actualPos;
    dGeomID geom = dCreateTriMesh( 0, tmdid, 0, 0, 0 );
    //  trees don't collide with other trees, or ground
    dGeomSetCollideBits( geom, ~1 );
    dGeomSetCategoryBits( geom, 1 );
    dGeomSetRotation( geom, &m3.xx );
    dGeomSetPosition( geom, actualPos.x, actualPos.y, actualPos.z );
    dSpaceAdd( gWorldSpace, geom );
    trees_[i].init( tree, treeShader, m4, scene );
    if( nToPlace ) {
      placePos = placePos + Vector3( float(rand() & 0x7fff) * 16.0f/0x8000 - 8.0f, 0, float(rand() & 0x7fff) * 16.0f/0x8000 - 8.0f );
      while( placePos.x < -500 ) placePos.x += 30;
      while( placePos.x > 500 ) placePos.x -= 30;
      while( placePos.z < -500 ) placePos.z += 30;
      while( placePos.z > 500 ) placePos.z -= 30;
    }
  }
#endif
}


#if INVESTIGATE

#include "mygui.h"


GuiRoot * testGui;

namespace {
  class FooCallback : public MyButtonCallback {
    public:
      FooCallback() {}
      void onButtonClicked( GuiWidget * w, uint mbMask ) {
        assert( mbMask != 0 );
        size_t s;
        val_->getData( &s, sizeof( s ) );
        char buf[ 40 ];
        sprintf( buf, "You chose %d", (int)s );
        txt_->setData( buf, strlen( buf ) );
      }
      GuiWidget * val_;
      GuiWidget * txt_;
  };
  FooCallback callback;
}


#endif

void MyContent::createScene()
{
  Data d;
  if( !FS::readFileByPath( "keymap.twk", &d ) ) {
    CommandBase::initKeysDefault();
    d = Data( 4096 );
    size_t s = CommandBase::saveKeysText( d.writeBase(), 4096 );
    d.trim( s );
    FS::writeFileByPath( "keymap.twk", d );
  }
  else {
    CommandBase::initKeysText( d.readBase(), d.size() );
  }

  gShaderFactory = NewShaderFactory();

  FS::clearHandles();
  FS::addHandles( "media/filemap.xml" );

  setDefaultKeyHandling( false );
  setDefaultMouseCamera( false );

  ShowSelectGui();
  bgShader = gShaderFactory->newShader( "guiBackground" );
}

void MyContent::createScene2( size_t vehicle )
{
  scene_ = MakeScene();
  groundShader = gShaderFactory->newShader( "groundShader" );

#if INVESTIGATE

  WidgetSize size = { 100, 50, 600, 450 };
  testGui = NewGui( size );
  gGuis.push_back( testGui );

  GuiWidget * w = NewStaticTextWidget( "Hello, world!", 0 );
  WidgetSize pos = { 0, 0, 500, 15 };
  testGui->add( w, pos );
  callback.txt_ = w;

  WidgetSize p2 = { 400, 205, 480, 230 };
  w = NewButton( "Done", &callback );
  testGui->add( w, p2 );

  WidgetSize p3 = { 100, 40, 400, 140 };
  char * sels[] = {
    "Two",
    "III",
    "4",
  };
  w = NewSelector( "Please choose one:", sels, 3, 0 );
  testGui->add( w, p3 );
  callback.val_ = w;

  WidgetSize img = { 100, 250, 200, 310 };
  Image i;
  loadTgaImage( imagePath( "jeep-image" ), &i );
  w = NewImageWidget( i );
  testGui->add( w, img );

#else

  gSoundManager = NewSoundManager();
#if !defined( NDEBUG )
  //  quieter in debug mode
  gSoundManager->setVolume( 0.1f );
#endif
  ExplodableSoundBase::setGunshotSound( gSoundManager->loadSound( wavPath( "gunshot" ) ) );

  explosionShader = gShaderFactory->newShader( "explosionShader" );
  textShader = gShaderFactory->newShader( "textShader" );
  skyShader = gShaderFactory->newShader( "skyShader" );
  skyMesh = ASEFromFile( asePath( "skyMesh" ) );

  chassisModel = ASEFromFile( asePath( "chassis" ) );
  tankModel = ASEFromFile( asePath( "tank" ) );
  tireModel = ASEFromFile( asePath( "wheel" ) );
  Image tga;
  char const * terrainName = imagePath( "terrain" );
  if( !loadTgaImage( terrainName, &tga ) ) {
    error( "Could not find terrain image '%s'.\n", terrainName );
    exit( 1 );
  }

  terrain = NewTerrainFromImage( tga, TERRAIN_SCALE, TERRAIN_OFFSET );
  assert( terrain );

  text = MakeGlText( 512, 256 );
  text->clear();
  memset( debugText, ' ', sizeof( debugText ) );
  textShader->setTexture( "diffuseMap", text->texture() );

  dMass m;
  dGeomID tempGeom;

  gWorld = dWorldCreate();
  MySystem::setWorld( gWorld );
  dWorldSetAutoDisableFlag( gWorld, AUTO_DISABLE > 0 );
  dWorldSetAutoDisableLinearThreshold( gWorld, AUTO_DISABLE_LINEAR );
  dWorldSetAutoDisableAngularThreshold( gWorld, AUTO_DISABLE_ANGULAR );
  dWorldSetAutoDisableSteps( gWorld, int( AUTO_DISABLE_STEPS ) );
  dWorldSetGravity( gWorld, 0, -9, 0 );
  gTempGroup = dJointGroupCreate( 100 );
  MySystem::setTempJointGroup( gTempGroup );
  gWorldSpace = dHashSpaceCreate( 0 );
  dHashSpaceSetLevels( gWorldSpace, 2, 5 );
  MySystem::setWorldSpace( gWorldSpace );
  camera = new MyCamera();
  utilRay = dCreateRay( 0, 0 );
  MySystem::setUtilRay( utilRay );
  utilSphere = dCreateSphere( 0, 1 );
  MySystem::setUtilSphere( utilSphere );

  //  a bottoming-out plane
  tempGeom = dCreatePlane( gWorldSpace, 0, 1, 0, -PLANEBOTTOM );
  //  and the actual terrain
  terrainGeom = terrain->terrainODE()->makeGeom( gWorldSpace );
  //  terrain doesn't collide with other static terrain
  dGeomSetCollideBits( terrainGeom, ~1 );
  dGeomSetCategoryBits( terrainGeom, 1 );

  if( vehicle == 0 ) {
    carShader = gShaderFactory->newShader( "carBody" );
    wheelShader = gShaderFactory->newShader( "carWheel" );
    MyWheeledCar::Params carParams;
    carParams.carMesh = chassisModel;
    carParams.carShader = carShader;
    carParams.wheelMesh = tireModel;
    carParams.wheelShader = wheelShader;
    carParams.terrain = terrain;
    gPlayer = new MyWheeledCar( carParams );
  }
  else {
    assert( vehicle == 1 );
    MyTank::Params tankParams;
    tankParams.body = tankModel;
    tankParams.terrain = terrain;
    gPlayer = new MyTank( tankParams );
  }
  camera->setFocus( gPlayer );

  {
    GLShader * crateShader = gShaderFactory->newShader( "vertexlit" );
    crateShader->setTexture( "diffuseMap", TextureRef( imagePath( "noise" ) ) );
    I_GL_ASE * crateMesh = ASEFromFile( asePath( "crate" ) );
    for( int ix = 0; ix < NUM_CRATES; ++ix ) {
      MyCrate * aCrate = new MyCrate( crateShader, crateMesh, 2, 1, 1 );
      crates_.push_back( aCrate );
    }
  }

  InitCratePositions();
  CreateForest( scene_ );

  //  Car starting position
  Vector3 pos = RayCastAgainstWorld( Vector3( 0, 100, 0 ) );
  MoveCarBy( pos.x, pos.y+1, pos.z );

  debugVar( 1, 5, "space shoot; 'g' aim; ']' weapon; '=' reset" );
  debugVar( 1, 6, "WASD drive; 'r' un-flip; 'j' jump; 'i' init crates" );
  debugVar( 1, 7, "8/9/0 camera" );

  camera->cutCamera();

  if( networked_ ) {
    setupNetwork();
  }
#endif
}

void MyContent::setupNetwork()
{
  network_ = new MyNetwork;
  if( !network_->open() ) {
    delete network_;
    network_ = 0;
  }
}



extern "C" {
  static void NearCallback( void *, dGeomID g1, dGeomID g2 )
  {
    if( dGeomIsSpace( g1 ) || dGeomIsSpace( g2 ) ) {
      dSpaceCollide2( g1, g2, 0, &NearCallback );
    }
    else {
      dContact contact;
      contact.fdir1[0] = 0;
      contact.fdir1[1] = 0;
      contact.fdir1[2] = -1;
      contact.surface.mode = 0;
      contact.surface.mu = SURFACE_MU;

      MySystem * s1 = (MySystem *)dGeomGetData( g1 );
      MySystem * s2 = (MySystem *)dGeomGetData( g2 );
      dBodyID b1 = dGeomGetBody( g1 );
      dBodyID b2 = dGeomGetBody( g2 );
      float inv = 1;
      dContactGeom contacts[8];
      int nc = dCollide( g1, g2, 8, contacts, sizeof( dContactGeom ) );
      if( b1 || b2 ) {
        for( int ix = 0; ix < nc; ++ix ) {
          contact.geom = contacts[ix];
          if( (!s1 || s1->wantsContact( contact, s2, g1, g2, true )) && 
              (!s2 || s2->wantsContact( contact, s1, g2, g1, true )) ) {
            dJointID tempJoint = dJointCreateContact( gWorld, gTempGroup, &contact );
            dJointAttach( tempJoint, b1, b2 );
          }
        }
      }
    }
  }
}

static void AccumulateForces()
{
  camera->preStep();
  MyMine::preStepAll();
  MyGrenade::preStepAll();
  MyRocket::preStepAll();

  dReal const * cP = dBodyGetPosition( gPlayer->chassisBody_ );
  dReal const * cR = dBodyGetRotation( gPlayer->chassisBody_ );

  if( cmdRecover.isset() ) {
    gPlayer->recoverStep();
    cmdRecover.clear();
  }
  if( cmdJump.isset() ) {
    gPlayer->jumpImpulse();
  }
  if( cmdInitCrates.isset() ) {
    InitCratePositions();
    cmdInitCrates.clear();
  }

  //  apply forces
  bool zero = true;
  if( cmdRight.isset() ) {
    if( g_turn < TURN_MAX ) {
      g_turn += TURN_INCREMENT;
    }
    zero = false;
  }
  if( cmdLeft.isset() ) {
    if( g_turn > -TURN_MAX ) {
      g_turn -= TURN_INCREMENT;
    }
    zero = false;
  }
  if( zero ) {
    if( g_turn > TURN_COAST ) {
      g_turn -= TURN_COAST;
    }
    else if( g_turn < -TURN_COAST ) {
      g_turn += TURN_COAST;
    }
    else {
      g_turn = 0;
    }
  }
  debugVar( 14, 0, "turn %6.3f", g_turn );
  gPlayer->setTurn( g_turn );

  zero = true;
  float gasTarget = 0;

//  if( g_keys['b'] ) {
//    gPlayer->setSwayBars( !gPlayer->swayBars() );
//    g_keys['b'] = false;
//  }
//  debugVar( 1, 3, "sway bars %s", gPlayer->swayBars() ? "ON " : "OFF" );

  if( cmdGas.isset() ) {
    if( g_gas < MAX_FORWARD ) {
      g_gas += GAS_INCREMENT;
    }
  }
  else if( cmdBrake.isset() ) {
    if( g_gas > -MAX_FORWARD*BACK_POWER/POWER ) {
      g_gas -= GAS_DECREMENT;
    }
    else if( g_gas < -MAX_FORWARD*BACK_POWER/POWER ) {
      g_gas += GAS_DECREMENT;
    }
  }
  else {
    if( g_gas > GAS_COAST ) {
      g_gas -= GAS_COAST;
    }
    else if( g_gas < -GAS_COAST ) {
      g_gas += GAS_COAST;
    }
    else {
      g_gas = 0;
    }
  }
  float p = (g_gas > 0) ? POWER * g_gas / MAX_FORWARD : - BACK_POWER * g_gas / MAX_FORWARD * BACK_POWER / POWER;
  gPlayer->setGas( IDLE_POWER + p, g_gas );

  debugVar( 27, 0, "gas %6.3f", g_gas );
  gPlayer->preStep();

  std::for_each( crates_.begin(), crates_.end(), MySystem::preStepOne );
}

static void drawSkybox()
{
  skyShader->setFloat( "wind", float( fmod( curTime, 600.0 ) / 600.0 ) );
  skyShader->apply();

  glPushMatrix();
  float mvm[ 16 ];
  glGetFloatv( GL_MODELVIEW_MATRIX, mvm );
  mvm[12] = mvm[13] = mvm[14] = 0;
  glLoadMatrixf( mvm );

  skyMesh->issueAll();

  glPopMatrix();
}

static void SetupCamera( Vector4 * frustumPlanesOut )
{
  glMatrixMode( GL_PROJECTION );
  glLoadIdentity();
  extern float ASPECT;
  glFrustum( -0.5f, 0.5f, -ASPECT*0.5f, ASPECT*0.5f, 0.7f, 5000 );
  Vector3 lookAt, lookFrom, lookUp;
  camera->setAim( wantMouseGrabbed ? g_aim : 0, wantMouseGrabbed ? g_elevation : 0 );
  camera->parameters( &lookFrom, &lookAt, &lookUp );
  glMatrixMode( GL_MODELVIEW );
  glLoadIdentity();
#if CAMERA_FROM_ABOVE
  glTranslatef( 0, 0, -CAMERA_FROM_ABOVE );
  glRotatef( 45, 1, 0, 0 );
#endif
  gluLookAt( lookFrom.x, lookFrom.y, lookFrom.z, lookAt.x, lookAt.y, lookAt.z, lookUp.x, lookUp.y, lookUp.z );
  assert( !glGetError() );

  float mvm[16];
  glGetFloatv( GL_MODELVIEW_MATRIX, mvm );
  gShaderFactory->setSharedMatrix( "s_camera4", mvm, 4, 1 );
  float mvmi[9] = {
    mvm[0], mvm[4], mvm[8],
    mvm[1], mvm[5], mvm[9],
    mvm[2], mvm[6], mvm[10],
  };
  gShaderFactory->setSharedMatrix( "s_invCamera3", mvmi, 3, 1 );

  float pm[16];
  glGetFloatv( GL_PROJECTION_MATRIX, pm );
  ExtractFrustumPlanes( *(Matrix4 const *)mvm, *(Matrix4 const *)pm, frustumPlanesOut );

#if 1
  //  render camera back and up a bit, so that culling can be seen
  glLoadIdentity();
  glTranslatef( 0, 0, -75 );
  glRotatef( 15, 1, 0, 0 );
  glMultMatrixf( mvm );
#endif
}

static void SetupScene()
{
  assert( !glGetError() );
  float a[4] = { 0.4f, 0.5f, 0.6f, 1 };
  float d[4] = { 0.6f, 0.5f, 0.4f, 0 };
  glLightfv( GL_LIGHT0, GL_AMBIENT, a );
  glLightfv( GL_LIGHT0, GL_DIFFUSE, d );
  glEnable( GL_COLOR_MATERIAL );
  glColorMaterial( GL_FRONT, GL_AMBIENT_AND_DIFFUSE );
  glColor4f( 1, 1, 1, 1 );
  float r = 1.0f/sqrtf( 2*2+5*5 );
  float dir[4] = { 2, 5, 0, 0 };
  dir[0] *= r; dir[1] *= r;
  glLightfv( GL_LIGHT0, GL_POSITION, dir );
  float zero[4] = { 0, 0, 0, 0 };
  glLightModelfv( GL_LIGHT_MODEL_AMBIENT, zero );
  glMaterialfv( GL_FRONT, GL_SPECULAR, zero );
  glMaterialfv( GL_FRONT, GL_EMISSION, zero );

  glFogi( GL_FOG_MODE, GL_LINEAR );
  float f[4] = { 0.5f, 0.55f, 1.0f, 0 };
  glFogfv( GL_FOG_COLOR, f );
  glFogf( GL_FOG_START, 30 );
  glFogf( GL_FOG_END, 500 );

  assert( !glGetError() );
}

static void regenDebug()
{
  text->clear();
  for( int ix = 0; ix < 16; ++ix ) {
    text->drawText( GWColor( 255, 255, 192, 128 ), 0, ix*16, debugText[ix], 64 );
  }
  text->texture();    //  side-effect is to update texture
  debugDirty = false;
}


void NextWeapon()
{
  gPlayer->nextWeapon( 1 );
  gCurWeaponOpacity = 3;
}

void PrevWeapon()
{
  gPlayer->nextWeapon( -1 );
  gCurWeaponOpacity = 3;
}

void FireWeapon()
{
  I_Weapon * w = gPlayer->weapon();
  if( w ) {
    w->fire( curTime );
  }
}

void MyContent::beginDrag( int x, int y, unsigned int buttons )
{
  if( !gGuis.size() ) {
    clickedButtons_ |= buttons;
    if( (buttons & 1) != 0 ) {
      cmdFire.set();
    }
    if( buttons & (1<<GW_WHEELUP) ) {
      NextWeapon();
    }
    if( buttons & (1<<GW_WHEELDOWN) ) {
      PrevWeapon();
    }
    return;
  }
  GuiRoot * r = gGuis.back();
  r->beginClick( x, y, buttons );
}

void MyContent::dragMove( int x, int y )
{
  if( !gGuis.size() ) {
    if( wantMouseGrabbed ) {
      g_aim -= x / 100.0f;
      g_elevation -= y / 100.0f;
    }
    return;
  }
  GuiRoot * r = gGuis.back();
  r->mouseMove( x, y );
}

void MyContent::endDrag( int x, int y )
{
  clickedButtons_ = 0;
  cmdFire.clear();    //  for auto-fire weapons
  if( !gGuis.size() ) return;
  GuiRoot * r = gGuis.back();
  r->endClick( x, y );
}

void MyContent::mouseMove( int x, int y )
{
  if( !gGuis.size() ) {
    if( wantMouseGrabbed ) {
      g_aim -= x / 100.0f;
      if( g_aim < -1.5f ) {
        g_aim = -1.5f;
      }
      else if( g_aim > 1.5f ) {
        g_aim = 1.5f;
      }
      g_elevation -= y / 100.0f;
      if( g_elevation < -1.0f ) {
        g_elevation = -1.0f;
      }
      else if( g_elevation > 1.0f ) {
        g_elevation = 1.0f;
      }
    }
    return;
  }
  GuiRoot * r = gGuis.back();
  r->mouseMove( x, y );
}

static void ResetCar()
{
  //  right the car up
  gPlayer->recoverStep();

  //  teleport
  Vector3 pos = RayCastAgainstWorld( Vector3( 0, 200, 0 ) );
  pos.y += 2;
  MoveAndOrientCar( pos, Matrix4::Identity );
  for( size_t i = 0; i < gPlayer->bodyCount(); ++i ) {
    dBodyEnable( gPlayer->bodyAt( i ) );
  }
  camera->cutCamera();
  g_aim = 0;
  g_elevation = 0;
}

namespace {
  class RenderPropVisitor : public I_SceneVisitor {
    public:
      void onSceneVisit( I_Scene * scene, SceneRef ref, I_SceneNode * node, Matrix4 const & posOri ) {
        static_cast< MyProp * >( node )->render();
      }
  };
};

void MyContent::step( float deltaTime )
{
  static long cntFrames = 0;
  static double frameTime = 0;

  static double lastTime = 0;
  if( deltaTime > 0.1f ) {
    deltaTime = 0.1f;
  }
  curTime += deltaTime;
  MySystem::setCurTime( curTime );

  gShaderFactory->setSharedFloat( "s_time_1", float( fmod( curTime, 1 ) ) );
  gShaderFactory->setSharedFloat( "s_time_2", float( fmod( curTime, 2 )/2 ) );
  gShaderFactory->setSharedFloat( "s_time_3", float( fmod( curTime, 3 )/3 ) );
  gShaderFactory->setSharedFloat( "s_time_4", float( fmod( curTime, 4 )/4 ) );
  gShaderFactory->setSharedFloat( "s_time_5", float( fmod( curTime, 5 )/5 ) );
  gShaderFactory->setSharedFloat( "s_time_10", float( fmod( curTime, 5 )/5 ) );
  gShaderFactory->setSharedFloat( "s_time_60", float( fmod( curTime, 60 )/60 ) );
  gShaderFactory->setSharedFloat( "s_time_600", float( fmod( curTime, 600 )/600 ) );
  gShaderFactory->setSharedFloat( "s_deltaTime", deltaTime );

  grabMouse( wantMouseGrabbed && (gGuis.size() == 0) );
  static double fps = 0;
  if( !gPlayer ) {
    glClearColor( 0, 0, 0, 0 );
    glClear( GL_COLOR_BUFFER_BIT );
  }
  if( gPlayer ) {
    if( gWireframe ) {
      glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );
    }
    else {
      glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
    }

    //  update controls
    //  update physics
    //  traverse scene graph
    //  render

    static double timeStepDuration = TIME_STEP;

    while( lastTime < curTime ) {

      if( cmdNextWeapon.isset() ) {
        NextWeapon();
        cmdNextWeapon.clear();
      }
      if( cmdPrevWeapon.isset() ) {
        PrevWeapon();
        cmdPrevWeapon.clear();
      }

      if( cmdFire.isset() ) {
        dReal const * R = dBodyGetRotation( gPlayer->chassisBody_ );
        //  am I facing mostly up?
        if( R[5] > 0 ) {
          FireWeapon();
        }
        else {
          //  upside down
          R = 0;
        }
        if( !gPlayer->weapon() || !gPlayer->weapon()->autoFire() ) {
          cmdFire.clear();
        }
      }

      if( cmdTeleport.isset() ) { //  teleport
        Vector3 left, up, front;
        gPlayer->basis( &left, &up, &front );
        Vector3 pos = gPlayer->P() + front * 100;
        pos = RayCastAgainstWorld( pos + Vector3( 0, 200, 0 ) );
        MoveCarBy( pos - gPlayer->P() + Vector3( 0, 2, 0 ) );
        camera->preventCutForTime( 3.0f );
        for( uint i = 0; i < gPlayer->bodyCount(); ++i ) {
          ClearForceAndTorque( gPlayer->bodyAt( i ) );
        }
        cmdTeleport.clear();
      }

      if( cmdGotoCenter.isset() ) {
        ResetCar();
        cmdGotoCenter.clear();
      }

      gPlayer->setAim( &g_aim, &g_elevation );

      lastTime += TIME_STEP;

      //  todo: dSpaceCollide, and the ray sensors in the tank, 
      //  sometimes take FOREVER to complete -- figure out why!
      extern double now();
      double a = now();
      dSpaceCollide( gWorldSpace, 0, &NearCallback );

      AccumulateForces();

      dWorldQuickStep( gWorld, TIME_STEP );
  //    dWorldStep( world, TIME_STEP );
      dJointGroupEmpty( gTempGroup );

      MyGrenade::postStepAll();
      MyMine::postStepAll();
      MyRocket::postStepAll();
      gPlayer->postStep();
      std::for_each( crates_.begin(), crates_.end(), MySystem::postStepOne );
      camera->postStep();

      std::for_each( deferredDelete_.begin(), deferredDelete_.end(), MySystem::destroy );
      deferredDelete_.clear();

#if 0 && !defined( NDEBUG )
      double b = now();
      timeStepDuration = timeStepDuration * 0.9f + (b-a) * 0.1f;
      if( timeStepDuration > TIME_STEP * 2 ) {
        error( "Time step is taking %.3f seconds; re-setting!\n", b-a );
        timeStepDuration = TIME_STEP;
        lastTime = curTime;
        ResetCar();
        goto done;
      }
#endif
    }

    MyParticles::stepAll( deltaTime );
    camera->step( deltaTime );

    ++cntFrames;
    if( curTime - frameTime > 1 ) {
      fps = cntFrames / (curTime - frameTime);
      frameTime = curTime;
      cntFrames = 0;
      static int logInterval = 0;
      if( ++logInterval == 5 ) {
        info( "fps: %.1f\n", fps );
        logInterval = 0;
      }
    }

    SetupCamera( frustumPlanes_ );
    SetupScene();

    drawSkybox();

    assert( !glGetError() );

    groundShader->apply();
    terrain->issueAll( gPlayer->P() );

    std::for_each( crates_.begin(), crates_.end(), MySystem::renderOne );  
    RenderPropVisitor rpv;
    scene_->visitNodesPlanes( frustumPlanes_, 6, &rpv );

  // draw car
    gPlayer->render();

    MyGrenade::renderAll();
    MyMine::renderAll();
    MyRocket::renderAll();
    RenderMachinegunBullets();

    glBegin( GL_QUADS );
    glNormal3f( 0, 1, 0 );
    glTexCoord2f( -10, -10 ); glVertex3f( -300, -PLANEBOTTOM, -300 );
    glTexCoord2f( -10, 10 ); glVertex3f( -300, -PLANEBOTTOM, 300 );
    glTexCoord2f( 10, 10 ); glVertex3f( 300, -PLANEBOTTOM, 300 );
    glTexCoord2f( 10, -10 ); glVertex3f( 300, -PLANEBOTTOM, -300 );
    glEnd();

    assert( !glGetError() );

    //  transparent things go here

  #if 0
  extern void DebugRenderTankLines();
  DebugRenderTankLines();
  #endif

    DrawExplosionEffects( deltaTime, explosionShader );
    MyParticles::renderAll();
  }

  Matrix4 camMtx;
  glGetFloatv( GL_MODELVIEW_MATRIX, &camMtx.xx );
  Matrix4 projMtx;
  glGetFloatv( GL_PROJECTION_MATRIX, &projMtx.xx );

  glMatrixMode( GL_PROJECTION );
  glLoadIdentity();
  extern float WINWIDTH, WINHEIGHT;
  glOrtho( 0, WINWIDTH, 0, WINHEIGHT, 0, 1 );
  glMatrixMode( GL_MODELVIEW );
  glLoadIdentity();

  extern float WINWIDTH, WINHEIGHT;

  if( gPlayer ) {

    I_Weapon * w = gPlayer->weapon();
    if( w ) {
      if( gCurWeaponOpacity > 0 ) {
        glTranslatef( WINWIDTH-96, 48, 0 );
        glScalef( 128, 128, 0 );
        if( camera->desiredShot() == MyCamera::ChaseCam || 
            camera->desiredShot() == MyCamera::FirstPersonIsh ) {
          w->renderIndicators( gCurWeaponOpacity, gCurWeaponOpacity > 1 ? 1 : gCurWeaponOpacity );
        }
        glLoadIdentity();
        gCurWeaponOpacity -= deltaTime;
      }
      w->renderSight( WINWIDTH, WINHEIGHT, camMtx, projMtx );
    }

    if( FPS_OVERLAY > 0 ) {
      debugVar( 50, 0, "fps: %5.1f", fps );
      if( debugDirty ) {
        regenDebug();
      }
      textShader->apply();
      glBegin( GL_QUADS );
      glTexCoord2f( 0, 0.5 ); glVertex2f( 0, 0 );
      glTexCoord2f( 1, 0.5 ); glVertex2f( 512, 0 );
      glTexCoord2f( 1, 1 ); glVertex2f( 512, 128 );
      glTexCoord2f( 0, 1 ); glVertex2f( 0, 128 );
      glEnd();
    }
  }

  if( gGuis.size() ) {
    //  add a fiery inferno
    bgShader->apply();
    glBegin( GL_QUADS );
    glTexCoord2f( 0, 0 ); glVertex2f( 0, 0 );
    glTexCoord2f( 1, 0 ); glVertex2f( WINWIDTH, 0 );
    glTexCoord2f( 1, 1 ); glVertex2f( WINWIDTH, WINHEIGHT );
    glTexCoord2f( 0, 1 ); glVertex2f( 0, WINHEIGHT );
    glEnd();

    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();
    glOrtho( 0, WINWIDTH, WINHEIGHT, 0, 0, 1 );
    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();
    std::for_each( gGuis.begin(), gGuis.end(), RenderGui );
  }

  assert( !glGetError() );

  if( overlayVisible ) {
    gwSetOverlayVisibility( 2 );
  }
  if( !gPlayer && selectedVehicle_ != 0xffffffff ) {
    if( stepTimer_ == 0 ) {
      createScene2( selectedVehicle_ );
      HidePleaseWait();
    }
    else {
      ShowPleaseWait();
      --stepTimer_;
    }
  }

  if( network_ ) {
    static double oldTime;
    oldTime += deltaTime;
    unsigned long toSend = (unsigned long)(oldTime * 1000);
    oldTime -= toSend * 0.001;
    network_->step( toSend );
  }
}

extern int gLod; //===
extern bool gWireframe;

namespace {

  GuiRoot * quitGui;

  void ToggleQuitGui();

  class QuitYes : public MyButtonCallback {
    public:
      QuitYes() {}
      void onButtonClicked( GuiWidget *, uint ) {
        exit( 1 );
      }
  };
  QuitYes gQuitCallbackYes;

  class QuitNo : public MyButtonCallback {
    public:
      QuitNo() {}
      void onButtonClicked( GuiWidget *, uint ) {
        ToggleQuitGui();
      }
  };
  QuitNo gQuitCallbackNo;
  
  class QuitFilter : public MyGuiCallback {
    public:
      virtual bool captureKey( GuiRoot * root, unsigned char key ) {
        if( key == 27 ) {
          ToggleQuitGui();
        }
        else if( key == 'y' || key == 'Y' ) {
          exit( 1 );
        }
        else if( key == 'n' || key == 'N' ) {
          ToggleQuitGui();
        }
        return true;
      }
      virtual void guiClosing( GuiRoot * root ) {
        // do nothing
      }
      virtual void clickDone( GuiRoot * root, GuiWidget * target ) {
        //  do nothing
      }
  };
  QuitFilter gQuitCallback;

  bool QuitGuiShowing()
  {
    return !(gGuis.size() == 0 || *gGuis.rbegin() != quitGui);
  }

  void ToggleQuitGui()
  {
    if( !quitGui ) {
      WidgetSize quitSize = { uint(WINWIDTH/2-120),uint(WINHEIGHT/3),uint(WINWIDTH/2+120),uint(WINHEIGHT/3+120) };
      quitGui = NewGui( quitSize );
      quitGui->setCallback( &gQuitCallback );

      WidgetSize quitText = { 20,30,220,45 };
      GuiWidget * w = NewStaticTextWidget( "Really quit?", 0, 0xff8080 );
      quitGui->add( w, quitText );

      WidgetSize quitNo = { 20,70,100,100 };
      w = NewButton( "No", &gQuitCallbackNo, 0x80ff80 );
      quitGui->add( w, quitNo );

      WidgetSize quitYes = { 140,70,220,100 };
      w = NewButton( "Yes", &gQuitCallbackYes, 0xff0000 );
      quitGui->add( w, quitYes );

    }
    if( !QuitGuiShowing() ) {
      gGuis.push_back( quitGui );
    }
    else {
      gGuis.pop_back();
    }
  }
}


void MyContent::key( int x, int y, unsigned char key )
{
  if( gGuis.size() && (*gGuis.rbegin())->captureKey( x, y, key ) ) {
    return;
  }
  //todo: turn these characters into Command thingies
  switch( key ) {
    case 'c': {
      camera->cutCamera();
      break;
    }
    case 'g':
      wantMouseGrabbed = !wantMouseGrabbed;
      g_aim = 0;
      g_elevation = 0;
      break;
    case 'G':
      wantMouseGrabbed = false;
      g_aim = 0;
      g_elevation = 0;
      break;
    case 9: //  tab
      overlayVisible = !overlayVisible;
      break;
    case '`': {
      ShowSelectGui();
      break;
    }
    case 27: {
      ToggleQuitGui();
      break;
    }
    case '8': {
      camera->setDesiredShot( MyCamera::ChaseCam );
      break;
    }
    case '9': {
      camera->setDesiredShot( MyCamera::FirstPersonIsh );
      break;
    }
    case '0': {
      camera->setDesiredShot( MyCamera::SideCam );
      break;
    }
    case '1': {
      if( gLod > 0 ) --gLod;
      break;
    }
    case '2': {
      if( gLod < 10 ) ++gLod;
      break;
    }
    case '4': {
      gWireframe = !gWireframe;
      break;
    }
  }
}

void MyContent::special( int x, int y, int k, bool down )
{
  if( gGuis.size() == 0 ) {
    CommandBase * cb = CommandBase::keyCommand( k );
    if( cb ) {
      cb->setTo( down );
    }
  }
}
