
//  get MK_XBUTTON1 and 2 defined
#define _WIN32_WINNT 0x500

#include "glwrap.h"
#include "internaltexture.h"
#include "extsetup.h"
#include "gl-text.h"
#include "glshaderstate.h"

#include <stdarg.h>

#include <vector>
#include <algorithm>


bool g_glsl = false;
bool g_textureCompression = false;


static GlText * gLogText;
static float gOverlayTransparency;

static HDC curDC;
static HGLRC curGLRC;
static bool running = true;
static HINSTANCE hInstance;
static bool mouseGrabbed = false;

std::vector< C_GWWindow * > windows;
std::vector< C_GWWindow * > toPaint;


void 
gwInitGL()
{
  info( "Vendor: %s  Renderer: %s  Version: %s\n",
      glGetString( GL_VENDOR ), glGetString( GL_RENDERER ), glGetString( GL_VERSION ) );

  assert( !glGetError() );
  bool ok = (SetupGL1_3() != 0);
  ok = ok && SetupExtensionsFromString( "GL_ARB_vertex_buffer_object" );
  if( !ok ) {
    warning( "OpenGL 1.3 is not supported; version = %s; renderer = %s\n", 
        glGetString( GL_VERSION ), glGetString( GL_RENDERER ) );
    MessageBox( 0, "You need at least OpenGL 1.3 support\nin your graphics card.", 
        "Bad OpenGL Version", MB_OK );
    exit( 1 );
  }
  ok = SetupExtensionsFromString( "GL_ARB_texture_compression "
      "GL_EXT_texture_compression_s3tc" ) != 0;
  if( ok ) {
    g_textureCompression = true;
    info( "texture compression support detected.\n" );
  }
  else {
    info( "texture compression support not detected.\n" );
  }
  ok = SetupExtensionsFromString(
      "GL_ARB_shader_objects GL_ARB_vertex_shader GL_ARB_fragment_shader" ) != 0;
  if( ok ) {
    g_glsl = true;
    if( SetupExtensionsFromString( "GL_ARB_shading_language_100" ) ) {
      info( "GLSL support detected: version %s\n", 
          glGetString( GL_SHADING_LANGUAGE_VERSION_ARB ) );
    }
    else {
      info( "GLSL support detected: version 100 (no string)\n" );
    }
  }
  else {
    info( "GLSL support not detected.\n" );
  }
  assert( !glGetError() );
}

void gwSetOverlayVisibility( float v )
{
  gOverlayTransparency = v;
}

void 
gwSwapBuffers()
{
  assert( !glGetError() );
  if( gOverlayTransparency > 0 && gLogText != 0 ) {

    glUseProgramObjectARB( 0 );
    glDisable( GL_LIGHTING );
    glColor4ub( 255, 255, 255, (gOverlayTransparency > 1) ? 255 : unsigned char( 255 * gOverlayTransparency ) );
    glActiveTexture( GL_TEXTURE0_ARB );
    glBindTexture( GL_TEXTURE_2D, gLogText->texture().textureIdHack() );
    glEnable( GL_TEXTURE_2D );
    glDisable( GL_TEXTURE_CUBE_MAP_ARB );
    glDisable( GL_DEPTH_TEST );
    glEnable( GL_BLEND );
    glBlendFunc( GL_ONE, GL_ONE_MINUS_SRC_ALPHA );
    glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
    assert( !glGetError() );

    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();
    int vp[4];
    glGetIntegerv( GL_VIEWPORT, vp );
    glOrtho( 0, vp[2], 0, vp[3], 0, 1 );
    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();
    assert( !glGetError() );
    glBegin( GL_QUADS );
    glTexCoord2f( 0, 0 ); glVertex2i( vp[2]/2-256, vp[3]-256 );
    glTexCoord2f( 1, 0 ); glVertex2i( vp[2]/2+256, vp[3]-256 );
    glTexCoord2f( 1, 1 ); glVertex2i( vp[2]/2+256, vp[3] );
    glTexCoord2f( 0, 1 ); glVertex2i( vp[2]/2-256, vp[3] );
    glEnd();
    assert( !glGetError() );

    gOverlayTransparency -= 0.01f;
    glDisable( GL_TEXTURE_2D );
  }
  SwapBuffers( curDC );
}

void 
gwAddOverlayText( char const * str, ... )
{
  va_list args;
  va_start( args, str );
  gwAddOverlayTextV( str, args );
  va_end( args );
}

void 
gwAddOverlayTextV( char const * str, va_list args )
{
  char buf[2048];
  int i = _vsnprintf( buf, 2048, str, args );
  if( i < 0 ) {
    i = 2047;
  }
  buf[i] = 0;
  gwAddOverlayTextStr( buf );
}

void gwAddOverlayTextStr( char const * str )
{
  HGLRC cc = wglGetCurrentContext();
  if( !cc || cc == INVALID_HANDLE_VALUE ) {
    return;
  }
  if( !gLogText ) {
    gLogText = MakeGlText( 512, 256 );
    gLogText->fill( 0x404040, 0, 0, 512, 256 );
  }
  char const * end = str + strlen( str );
  while( str < end ) {
    char const * lf = strchr( str, '\n' );
    if( !lf ) {
      lf = end;
    }
    gLogText->scroll( 0x404040, 0, -16 );
    gLogText->fill( 0x404040, 0, 256-16, 512, 256 );
    gLogText->drawText( 0xf0f0f0, 0, 256-16, str, int(lf-str) );
    str = lf+1;
  }
  gOverlayTransparency = 4;
}


void
gwReset()
{
  glMatrixMode( GL_PROJECTION );
  glLoadIdentity();
  glMatrixMode( GL_TEXTURE );
  glLoadIdentity();
  glMatrixMode( GL_MODELVIEW );
  glLoadIdentity();
  glDisable( GL_LIGHTING );
  glDisable( GL_TEXTURE_CUBE_MAP );
  glDisable( GL_TEXTURE_3D );
  glDisable( GL_TEXTURE_2D );
  glDisable( GL_TEXTURE_1D );
  glFrontFace( GL_CCW );
  glCullFace( GL_BACK );
  glEnable( GL_CULL_FACE );
  glDisable( GL_DEPTH_TEST );
  glDisable( GL_BLEND );
  glDisable( GL_ALPHA_TEST );
  glDisable( GL_STENCIL_TEST );

  assert( !glGetError() );

  GLShader::flushState();
}

void gwInit( HINSTANCE hi, int * argcp, char const * const * argv )
{
  hInstance = hi;
}

static void logPixelFormat( HDC dc, int pf )
{
#define A(x) { x, #x }
  struct {
    GLenum what;
    char const * desc;
  } attribs[] = {
     A(WGL_DRAW_TO_WINDOW_ARB),
     A(WGL_DRAW_TO_BITMAP_ARB),
     A(WGL_ACCELERATION_ARB),
     A(WGL_NEED_PALETTE_ARB),
     A(WGL_NEED_SYSTEM_PALETTE_ARB),
     A(WGL_SWAP_LAYER_BUFFERS_ARB),
     A(WGL_SWAP_METHOD_ARB),
     A(WGL_NUMBER_OVERLAYS_ARB),
     A(WGL_NUMBER_UNDERLAYS_ARB),
     A(WGL_SHARE_DEPTH_ARB),
     A(WGL_SHARE_STENCIL_ARB),
     A(WGL_SHARE_ACCUM_ARB),
     A(WGL_SUPPORT_GDI_ARB),
     A(WGL_SUPPORT_OPENGL_ARB),
     A(WGL_DOUBLE_BUFFER_ARB),
     A(WGL_STEREO_ARB),
     A(WGL_PIXEL_TYPE_ARB),
     A(WGL_COLOR_BITS_ARB),
     A(WGL_RED_BITS_ARB),
     A(WGL_GREEN_BITS_ARB),
     A(WGL_BLUE_BITS_ARB),
     A(WGL_ALPHA_BITS_ARB),
     A(WGL_ACCUM_BITS_ARB),
     A(WGL_ACCUM_RED_BITS_ARB),
     A(WGL_ACCUM_GREEN_BITS_ARB),
     A(WGL_ACCUM_BLUE_BITS_ARB),
     A(WGL_ACCUM_ALPHA_BITS_ARB),
     A(WGL_DEPTH_BITS_ARB),
     A(WGL_STENCIL_BITS_ARB),
     A(WGL_AUX_BUFFERS_ARB),
     A(WGL_DRAW_TO_PBUFFER_ARB),
  };
#undef A
  char buf[4096];
  buf[0] = 0;
  bool first = true;
  char * line = buf;
  int linelen = 0;
  for( int i = 0; i < sizeof(attribs)/sizeof(attribs[0]); ++i ) {
    int attrib = attribs[i].what;
    int v = 0;
    wglGetPixelFormatAttribivARB( dc, pf, 0, 1, &attrib, &v );
    char a[128];
again:
    sprintf( a, "%s%s: %d", first ? "" : ", ", attribs[i].desc, v );
    int xlen = int( strlen( a ) );
    assert( xlen < 120 );
    if( linelen + xlen >= 120 ) {
      *line++ = '\n';
      linelen = 0;
      first = true;
      goto again;
    }
    first = false;
    assert( line + xlen < &buf[sizeof(buf)] );
    strcpy( line, a );
    linelen += xlen;
    line += xlen;
  }
  info( "---------- Pixel format  %d  ----------\n%s\n", pf, buf );
}

class C_GWWindow {
  public:
    I_GWWindow * iwin_;
    HWND wnd_;
    int x_;
    int y_;
    int width_;
    int height_;
    HDC dc_;
    bool active_;
    bool wantCursor_;
    bool clicking_;
    WPARAM btnState_;
    GLShaderInstanceState shaderState_;

    static bool hasCursor_;
    static bool appActive_;
    static int pixelFormat_;
    static HGLRC glrc_;
    static ATOM class_;
    static HCURSOR cursor_;

    C_GWWindow( I_GWWindow * win ) {
      win->impl_ = this;
      iwin_ = win;
      wnd_ = 0;
      dc_ = 0;
      x_ = 0;
      y_ = 0;
      width_ = 0;
      height_ = 0;
      active_ = false;
      wantCursor_ = true;
      clicking_ = false;
      btnState_ = 0;
      shaderState_ = 0;
    }

    static LRESULT CALLBACK WindowProc( HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam ) {
#pragma warning( push )
//  pointer truncation warning
#pragma warning( disable: 4311 )
//  pointer cast warning
#pragma warning( disable: 4312 )
      C_GWWindow * win = reinterpret_cast< C_GWWindow * >( GetWindowLong( hwnd, GWL_USERDATA ) );
      LRESULT res = 0;
      bool handled = false;
      if( win ) {
        handled = true;
        switch( msg ) {
        case WM_SIZE:
          handled = false;
          win->width_ = LOWORD( lParam );
          win->height_ = HIWORD( lParam );
          break;
        case WM_MOVE:
          handled = false;
          win->x_ = LOWORD( lParam );
          win->y_ = HIWORD( lParam );
          break;
        case WM_CLOSE:
          exit( 1 );
          break;
        case WM_PAINT: {
            PAINTSTRUCT ps;
            BeginPaint( hwnd, &ps );
            EndPaint( hwnd, &ps );
            toPaint.push_back( win );
          }
          break;
        case WM_CHAR:
          win->character( wParam, lParam );
          break;
        case WM_MOUSEWHEEL:
          win->mouseWheel( wParam, lParam );
          break;
        case WM_SYSKEYDOWN:
        case WM_KEYDOWN:
          if( !(lParam & 0x40000000) ) {  //  ignore key repeats
            win->keyDown( wParam, lParam );
          }
          break;
        case WM_SYSKEYUP:
        case WM_KEYUP:
          win->keyUp( wParam, lParam );
          break;
        case WM_TIMER:
          if( appActive_ ) {
            gwRedrawAll();
            gwStepAll();
          }
          break;
        case WM_LBUTTONDOWN:
        case WM_RBUTTONDOWN:
        case WM_MBUTTONDOWN:
        case WM_XBUTTONDOWN:
        case WM_LBUTTONUP:
        case WM_RBUTTONUP:
        case WM_MBUTTONUP:
        case WM_XBUTTONUP:
          win->click( wParam, (short)LOWORD( lParam ), (short)HIWORD( lParam ) );
          break;
        case WM_ACTIVATEAPP:
          appActive_ = wParam != 0;
          win->setActive( wParam != 0 );
          break;
        case WM_ACTIVATE:
          win->setActive( HIWORD(wParam) != 0 );
          break;
        case WM_MOUSEMOVE:
          win->updateCursor();
          win->moveCursor( (short)LOWORD( lParam ), (short)HIWORD( lParam ) );
          break;
        case WM_SETCURSOR:
          SetCursor( cursor_ );
          break;
        default:
          handled = false;
          break;
        }
      }
      else if( msg == WM_NCCREATE ) {
        CREATESTRUCT * cs = reinterpret_cast< CREATESTRUCT * >( lParam );
        SetWindowLong( hwnd, GWL_USERDATA, reinterpret_cast< LONG >( cs->lpCreateParams ) );
      }
      if( !handled ) {
        res = DefWindowProc( hwnd, msg, wParam, lParam );
      }
#pragma warning( pop )
      return res;
    }

    void setActive( bool a ) {
      if( !a && mouseGrabbed ) {
        //  Let go of the capture, even if we'll want it
        //  again (because we may be switching to other window).
        ReleaseCapture();
        mouseGrabbed = false;
      }
      if( active_ != a ) {
        active_ = a;
        clicking_ = false;
        updateCursor();
      }
    }

    void character( WPARAM wParam, LPARAM lParam ) {
      POINT pt;
      GetCursorPos( &pt );
      ScreenToClient( wnd_, &pt );
      iwin_->keyboard( unsigned char( wParam & 0xff ), pt.x, pt.y );
    }

    void keyDown( WPARAM wParam, LPARAM lParam ) {
      POINT pt;
      GetCursorPos( &pt );
      ScreenToClient( wnd_, &pt );
      iwin_->special( int( (wParam & 0xffff) | (lParam & 0xffff0000) ), pt.x, pt.y );
    }

    void keyUp( WPARAM wParam, LPARAM lParam ) {
      POINT pt;
      GetCursorPos( &pt );
      ScreenToClient( wnd_, &pt );
      iwin_->specialUp( int( (wParam & 0xffff) | (lParam & 0xffff0000) ), pt.x, pt.y );
    }

    void updateCursor() {
      bool wantCursor = (!appActive_) || wantCursor_;
      if( wantCursor != hasCursor_ ) {
        ShowCursor( wantCursor );
        hasCursor_ = wantCursor;
      }
      bool grabsCursor = (clicking_ || !wantCursor) && appActive_;
      if( grabsCursor != mouseGrabbed ) {
        if( grabsCursor ) {
          if( !wantCursor_ ) {
            //  center the cursor on capture
            int cx = width_/2;
            int cy = height_/2;
            POINT p; p.x = cx; p.y = cy;
            ClientToScreen( wnd_, &p );
            SetCursorPos( p.x, p.y );
          }
          //  capture the mouse
          SetCapture( wnd_ );
        }
        else {
          ReleaseCapture();
        }
        mouseGrabbed = grabsCursor;
      }
    }
    
    void moveCursor( int x, int y ) {
      if( !wantCursor_ && mouseGrabbed ) {
        int cx = width_/2;
        int cy = height_/2;
        x -= cx;
        y -= cy;
        if( x != 0 || y != 0 ) {
          static bool reEntry;
          assert( !reEntry );
          reEntry = true;
          POINT p; p.x = cx; p.y = cy;
          ClientToScreen( wnd_, &p );
          SetCursorPos( p.x, p.y );
          reEntry = false;
        }
      }
      if( clicking_ ) {
        iwin_->activeMotion( x, y );
      }
      else {
        iwin_->passiveMotion( x, y );
      }
    }

    void mouseWheel( WPARAM wParam, LPARAM lParam ) {
      int x = LOWORD( lParam );
      int y = HIWORD( lParam );
      if( HIWORD(wParam) & 0x8000 ) {
        iwin_->mouse( 6, GW_DOWN, x, y );
        iwin_->mouse( 6, GW_UP, x, y );
      }
      else {
        iwin_->mouse( 5, GW_DOWN, x, y );
        iwin_->mouse( 5, GW_UP, x, y );
      }
    }

    void click( WPARAM wParam, int x, int y ) {
      WPARAM changed = btnState_ ^ wParam;
      btnState_ = wParam;
      bool nuClicking =
          (btnState_ & (MK_LBUTTON | MK_MBUTTON | MK_RBUTTON | MK_XBUTTON1 | MK_XBUTTON2)) != 0;
      if( nuClicking != clicking_ ) {
        clicking_ = nuClicking;
        updateCursor();
      }
      if( changed & MK_LBUTTON ) {
        iwin_->mouse( 0, (wParam & MK_LBUTTON) ? GW_DOWN : GW_UP, x, y );
      }
      if( changed & MK_RBUTTON ) {
        iwin_->mouse( 1, (wParam & MK_RBUTTON) ? GW_DOWN : GW_UP, x, y );
      }
      if( changed & MK_MBUTTON ) {
        iwin_->mouse( 2, (wParam & MK_MBUTTON) ? GW_DOWN : GW_UP, x, y );
      }
      if( changed & MK_XBUTTON1 ) {
        iwin_->mouse( 3, (wParam & MK_XBUTTON1) ? GW_DOWN : GW_UP, x, y );
      }
      if( changed & MK_XBUTTON2 ) {
        iwin_->mouse( 4, (wParam & MK_XBUTTON2) ? GW_DOWN : GW_UP, x, y );
      }
    }

    void create( char const * title, int x, int y, int w, int h ) {

      PIXELFORMATDESCRIPTOR pfd;
      memset( &pfd, 0, sizeof( pfd ) );
      pfd.nSize = sizeof( pfd );
      pfd.nVersion = 1;
      pfd.dwFlags = PFD_DRAW_TO_WINDOW |
                    PFD_SUPPORT_OPENGL |
                    PFD_DOUBLEBUFFER;
      pfd.cAlphaBits = 8;
      pfd.iPixelType = PFD_TYPE_RGBA;
      pfd.cColorBits = 24;
      pfd.cDepthBits = 24;
      pfd.cStencilBits = 8;
      pfd.iLayerType = PFD_MAIN_PLANE;

      if( class_ == NULL ) {
        WNDCLASSEX wclex;
        memset( &wclex, 0, sizeof( wclex ) );
        wclex.cbSize = sizeof( wclex );
        wclex.style = CS_OWNDC | CS_HREDRAW | CS_VREDRAW;
        wclex.lpfnWndProc = WindowProc;
        wclex.hInstance = hInstance;
        wclex.lpszClassName = "GL Window Class";
        class_ = RegisterClassEx( &wclex );
      }
      if( !wglChoosePixelFormatARB ) {
        //  don't pollute the class DC of the main window class
        WNDCLASSEX wclex;
        memset( &wclex, 0, sizeof( wclex ) );
        wclex.cbSize = sizeof( wclex );
        wclex.style = CS_OWNDC | CS_HREDRAW | CS_VREDRAW;
        wclex.lpfnWndProc = WindowProc;
        wclex.hInstance = hInstance;
        wclex.lpszClassName = "Temporary Class";
        ATOM tempClass = RegisterClassEx( &wclex );
        cursor_ = LoadCursor( NULL, IDC_ARROW );

        //  create a temporary window to use for extracting pixel formats
        HWND w = CreateWindowEx(
            WS_EX_OVERLAPPEDWINDOW, 
            (LPCSTR)tempClass, 
            "", 
            WS_POPUPWINDOW, 
            0, 
            0, 
            100, 
            100, 
            0, 
            0, 
            hInstance, 
            0 );
        HDC dc = GetDC( w );
        int i = ChoosePixelFormat( dc, &pfd );
        SetPixelFormat( dc, i, &pfd );
        HGLRC glrc = wglCreateContext( dc );
        wglMakeCurrent( dc, glrc );
        //  set up the extensions
        if( !SetupWGLExtensions() ) {
          error( "SetupWGLExtensions() failed. Renderer: %s\n", glGetString( GL_RENDERER ) );
          MessageBox( 0, "Could not find WGL extensions. This program needs modern OpenGL drivers.", 
              "Missing Drivers", MB_OK );
          exit( 1 );
        }
        wglMakeCurrent( 0, 0 );
        wglDeleteContext( glrc );
        DestroyWindow( w );
      }
      RECT r;
      r.left = x; r.top = y; r.right = x+w; r.bottom = y+h;
      AdjustWindowRectEx( 
          &r, 
          WS_CAPTION | WS_POPUPWINDOW | WS_VISIBLE, 
          false, 
          WS_EX_APPWINDOW | WS_EX_OVERLAPPEDWINDOW );
      wnd_ = CreateWindowEx(
          WS_EX_APPWINDOW | WS_EX_OVERLAPPEDWINDOW,
          (LPCSTR)class_,
          title,
          WS_CAPTION | WS_POPUPWINDOW | WS_VISIBLE,
          r.left,
          r.top,
          r.right-r.left,
          r.bottom-r.top,
          0,
          0,
          hInstance,
          this );
      HDC hdc = GetDC( wnd_ );    //  they all should share this!
      dc_ = hdc;
      if( !glrc_ ) {
        int iFormats[100];
        int iAttribs[40];
        float fAttribs[10];
        UINT numFormats;

        //  find a pbuffer format
        memset( iFormats, 0, sizeof( iFormats ) );
        memset( iAttribs, 0, sizeof( iAttribs ) );
        memset( fAttribs, 0, sizeof( fAttribs ) );
        int i = 0;
        iAttribs[i++] = WGL_DRAW_TO_PBUFFER_ARB;  iAttribs[i++] = 1;
        iAttribs[i++] = WGL_SUPPORT_OPENGL_ARB;   iAttribs[i++] = 1;
        iAttribs[i++] = WGL_DOUBLE_BUFFER_ARB;    iAttribs[i++] = 0;
        iAttribs[i++] = WGL_PIXEL_TYPE_ARB;       iAttribs[i++] = WGL_TYPE_RGBA_FLOAT_ATI;
        iAttribs[i++] = WGL_COLOR_BITS_ARB;       iAttribs[i++] = 48;
        iAttribs[i++] = WGL_ALPHA_BITS_ARB;       iAttribs[i++] = 16;
        iAttribs[i++] = WGL_DEPTH_BITS_ARB;       iAttribs[i++] = 24;
        iAttribs[i++] = WGL_STENCIL_BITS_ARB;     iAttribs[i++] = 8;
        BOOL b = wglChoosePixelFormatARB( dc_, iAttribs, fAttribs, 100, iFormats, &numFormats );
        if( !b || numFormats == 0 ) {
          info( "no float pbuffer support\n" );
        }
        else {
          info( "There are %d float pbuffer formats\n", numFormats );
#if !defined( NDEBUG )
          for( size_t j = 0; j < numFormats; ++j ) {
            logPixelFormat( dc_, iFormats[j] );
          }
#else
          logPixelFormat( dc_, iFormats[0] );
#endif
        }

//  todo: render to pbuffer, copy to screen for display

        //  find a screen format
        memset( iFormats, 0, sizeof( iFormats ) );
        memset( iAttribs, 0, sizeof( iAttribs ) );
        memset( fAttribs, 0, sizeof( fAttribs ) );
        i = 0;
        iAttribs[i++] = WGL_DRAW_TO_WINDOW_ARB;   iAttribs[i++] = 1;
        iAttribs[i++] = WGL_DRAW_TO_PBUFFER_ARB;  iAttribs[i++] = 1;
        iAttribs[i++] = WGL_SUPPORT_OPENGL_ARB;   iAttribs[i++] = 1;
        iAttribs[i++] = WGL_DOUBLE_BUFFER_ARB;    iAttribs[i++] = 1;
        iAttribs[i++] = WGL_PIXEL_TYPE_ARB;       iAttribs[i++] = WGL_TYPE_RGBA_ARB;
        iAttribs[i++] = WGL_COLOR_BITS_ARB;       iAttribs[i++] = 24;
        iAttribs[i++] = WGL_ALPHA_BITS_ARB;       iAttribs[i++] = 8;
        iAttribs[i++] = WGL_DEPTH_BITS_ARB;       iAttribs[i++] = 24;
        iAttribs[i++] = WGL_STENCIL_BITS_ARB;     iAttribs[i++] = 8;
        b = wglChoosePixelFormatARB( dc_, iAttribs, fAttribs, 100, iFormats, &numFormats );
        if( !b ) {
          LPVOID str = 0;
          DWORD err = GetLastError();
          FormatMessage(
            FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, 
            0,
            err,
            MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
            (LPSTR)&str,
            0,
            0 );
          error( "wglChoosePixelFormatARB() failed: 0x%x (%s)\n", err, str );
          char * str2;
          FormatMessage( 
            FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_STRING | FORMAT_MESSAGE_ARGUMENT_ARRAY, 
            "The system could not choose a pixel format. \nThis usually means the graphics card drivers \n"
            "are not properly installed, or that the graphics \ncard is too old. \nError: %1",
            0,
            0,
            (LPSTR)&str2,
            0,
            (va_list *)&str );
          LocalFree( str );
          MessageBox( 0, str2, "Pixel Format Failed", MB_OK );
          LocalFree( str2 );
          exit( 1 );
        }
        else {
          info( "There are %d suitable pixel formats.\n", numFormats );
        }
        if( numFormats > 100 ) {
          numFormats = 100;
        }
#if !defined( NDEBUG )
        for( size_t j = 0; j < numFormats; ++j ) {
          logPixelFormat( dc_, iFormats[j] );
        }
#else
        logPixelFormat( dc_, iFormats[0] );
#endif
        info( "Using format %d\n", iFormats[0] );
        SetPixelFormat( dc_, iFormats[0], &pfd );
        glrc_ = wglCreateContext( dc_ );
        pixelFormat_ = iFormats[0];
      }
      else {
        //  I'm going to assume that I can re-use the same glrc!
        SetPixelFormat( dc_, pixelFormat_, &pfd );
      }
      wglMakeCurrent( dc_, glrc_ );
      gwInitGL();
      GLSwitchToShaderInstanceState( shaderState_ );
      curDC = dc_;
      curGLRC = glrc_;
    }

    static void stepOne( C_GWWindow * it ) {
      it->iwin_->timer();
    }
    static void redrawOne( C_GWWindow * it ) {
      wglMakeCurrent( it->dc_, it->glrc_ );
      GLSwitchToShaderInstanceState( it->shaderState_ );
      curDC = it->dc_;
      curGLRC = it->glrc_;
      RECT r;
      GetClientRect( it->wnd_, &r );
      glViewport( r.left, r.top, r.right-r.left, r.bottom-r.top );
      it->iwin_->display();
    }

    static void position( I_GWWindow * win, int x, int y ) {
      MoveWindow( win->impl_->wnd_, x, y, win->impl_->width_, win->impl_->height_, FALSE );
    }

    static void reshape( I_GWWindow * win, int w, int h ) {
      MoveWindow( win->impl_->wnd_, win->impl_->x_, win->impl_->y_, w, h, FALSE );
    }

    static void redisplay( I_GWWindow * win ) {
      RECT r;
      GetClientRect( win->impl_->wnd_, &r );
      InvalidateRect( win->impl_->wnd_, &r, FALSE );
    }

    static void grabMouse( I_GWWindow * win, bool grabbed ) {
      win->impl_->wantCursor_ = !grabbed;
      if( grabbed && win->impl_->active_ ) {
        POINT p; p.x = win->impl_->width_/2; p.y = win->impl_->height_/2;
        ClientToScreen( win->impl_->wnd_, &p );
        SetCursorPos( p.x, p.y );
      }
      win->impl_->updateCursor();
    }
};

bool C_GWWindow::hasCursor_ = true;
bool C_GWWindow::appActive_;
int C_GWWindow::pixelFormat_;
HGLRC C_GWWindow::glrc_;
ATOM C_GWWindow::class_;
HCURSOR C_GWWindow::cursor_;

void gwRedrawAll()
{
  std::for_each( toPaint.begin(), toPaint.end(), C_GWWindow::redrawOne );
  toPaint.clear();
}

void gwStepAll()
{
  std::for_each( windows.begin(), windows.end(), C_GWWindow::stepOne );
}

void gwMainLoop()
{
  assert( !glGetError() );
  SetTimer( 0, 0, 50, 0 );
  while( running ) {
    MSG msg;
    while( PeekMessage( &msg, 0, 0, 0, PM_REMOVE ) ) {
      TranslateMessage( &msg );
      DispatchMessage( &msg );
    }
    // only do this if the application has focus
    if( C_GWWindow::appActive_ ) {
      gwStepAll();
      // todo: only do this for windows that has requested full frame rate
      toPaint.clear();
      toPaint.insert( toPaint.begin(), windows.begin(), windows.end() );
    }
    gwRedrawAll();
  }
}

void gwCreateWindow( char const * title, int x, int y, int w, int h, I_GWWindow * win )
{
  C_GWWindow * cgww = new C_GWWindow( win );
  cgww->create( title, x, y, w, h );
  windows.push_back( cgww );
}

void gwPositionWindow( I_GWWindow * win, int x, int y )
{
  C_GWWindow::position( win, x, y );
}

void gwReshapeWindow( I_GWWindow * win, int w, int h )
{
  C_GWWindow::reshape( win, w, h );
}

void gwPostRedisplay( I_GWWindow * win )
{
  C_GWWindow::redisplay( win );
}

void gwGrabMouse( I_GWWindow * win, bool grabbed )
{
  C_GWWindow::grabMouse( win, grabbed );
}
