
#if !defined( glwrap_h )
#define glwrap_h

#include "glut-start.h"
#include "textures.h"

#include <GL/GL.h>
#include "myext.h"
#include "extsetup.h"



union GWColor {
  GWColor( ) { }
  GWColor( unsigned int c ) : color( c ) { }
  GWColor( unsigned char a, unsigned char r, unsigned char g, unsigned char b ) : blue( b ), green( g ), red( r ), alpha( a ) { }
//fixme:  this is only true for little-endian platform
  struct {
    unsigned char   blue;
    unsigned char   green;
    unsigned char   red;
    unsigned char   alpha;
  };
  unsigned int      color;

  bool operator==( GWColor const & o ) const
            {
              return color == o.color;
            }
  bool operator!=( GWColor const & o ) const
            {
              return color == o.color;
            }
};


void gwInitGL();
void gwSwapBuffers();
void gwAddOverlayText( char const * str, ... );
void gwAddOverlayTextV( char const * str, va_list args );
void gwAddOverlayTextStr( char const * str );
void gwReset();
void gwSetOverlayVisibility( float v );

//  GLUT replacement
class I_GWWindowBase {
  private:
    friend class C_GWWindow;
    C_GWWindow * impl_;
};
class I_GWWindow : public I_GWWindowBase {
  public:
    virtual void display() = 0;
    //  keyboard() means the user typed text
    virtual void keyboard( unsigned char key, int x, int y ) = 0;
    //  special() gets all key codes
    virtual void special( int key, int x, int y ) = 0;
    virtual void specialUp( int key, int x, int y ) = 0;
    //  timer() gets called every so often
    virtual void timer() = 0;
    //  mouse means a button was clicked, or un-clicked
    virtual void mouse( int button, int state, int x, int y ) = 0;
    //  active motion means mouse was moved while clicked
    virtual void activeMotion( int x, int y ) = 0;
    //  active motion means mouse was moved while not clicked
    virtual void passiveMotion( int x, int y ) = 0;
};

void gwInit( HINSTANCE hi, int * argcp, char const * const * argv );
void gwMainLoop();
void gwStepAll();
void gwRedrawAll();
void gwCreateWindow( char const * title, int x, int y, int w, int h, I_GWWindow * );
void gwPositionWindow( I_GWWindow *, int x, int y );
void gwReshapeWindow( I_GWWindow *, int w, int h );
void gwPostRedisplay( I_GWWindow * );
//  When mouse is grabbed, mouse movement is relative
void gwGrabMouse( I_GWWindow * w, bool grabbed );

enum {
  GW_UP = 0,
  GW_DOWN = 1,

  GW_LEFTMB = 0,
  GW_RIGHTMB = 1,
  GW_MIDMB = 2,
  GW_X1MB = 3,
  GW_X2MB = 4,
  GW_WHEELUP = 5,
  GW_WHEELDOWN = 6,
};



#endif  //  glwrap_h
