
#include "glut-start.h"

#include "glwrap.h"
#include "myext.h"
#include "log.h"
#include "glshaderstate.h"
#include "misc.h"

#include "glut-content.h"

#include <string.h>


IGlutContent * g_content;
bool g_defaultKeyHandling = true;
bool g_defaultMouseCamera = true;
bool g_rendering = false;
bool g_mouseGrabbed = false;
float g_cameraXyz[ 3 ];
float g_cameraLookAt[ 3 ];
float g_cameraHeading;  //  degrees
float g_cameraPitch;    //  degrees
float g_cameraSpeed = 1;
float WINWIDTH = 1280;
float WINHEIGHT = 768;
float ASPECT = float(WINHEIGHT)/float(WINWIDTH);

__int64 g_lastTime;
double g_frequency;


void initCamera()
{
  g_cameraXyz[0] = 0;
  g_cameraXyz[1] = 2;
  g_cameraXyz[2] = 10;
  g_cameraLookAt[0] = 0;
  g_cameraLookAt[1] = 2;
  g_cameraLookAt[2] = 9;
  g_cameraHeading = 0;
  g_cameraPitch = 0;
}

void updateCameraHeading()
{
  g_cameraLookAt[0] = -float( sin( g_cameraHeading * 3.14159265358979 / 180 ) )+g_cameraXyz[0];
  g_cameraLookAt[1] = g_cameraXyz[1];
  g_cameraLookAt[2] = -float( cos( g_cameraHeading * 3.14159265358979 / 180 ) )+g_cameraXyz[2];
}

void updateMatrices()
{
  glMatrixMode( GL_MODELVIEW );
  glLoadIdentity();
  glRotatef( -g_cameraPitch, 1, 0, 0 );
  glRotatef( -g_cameraHeading, 0, 1, 0 );
  glTranslatef( -g_cameraXyz[0], -g_cameraXyz[1], -g_cameraXyz[2] );
  assert( !glGetError() );
}

struct IntOptionValue {
  char const * name;
  int val;
};
static IntOptionValue intValues[] = {
  { "winSizeX", int(WINWIDTH) },
  { "winSizeY", int(WINHEIGHT) },
};

struct CharOptionValue {
  char const * name;
  char const * val;
};
static CharOptionValue charValues[] = {
  { "userName", 0 },
};


int
intOption( char const * o )
{
  for( int ix = 0; ix < sizeof(intValues)/sizeof(intValues[0]); ++ix ) {
    if( !stricmp( intValues[ix].name, o ) ) {
      return intValues[ix].val;
    }
  }
  assert( 0 );
  return -1;
}

char const * charOption( char const * o )
{
  for( int ix = 0; ix < sizeof(charValues)/sizeof(charValues[0]); ++ix ) {
    if( !stricmp( charValues[ix].name, o ) ) {
      return charValues[ix].val;
    }
  }
  assert( 0 );
  return 0;
}



double now()
{
  __int64 tnow;
  QueryPerformanceCounter( (LARGE_INTEGER *)&tnow );
  return g_frequency * tnow;
}

class C_MyWindow : public I_GWWindow
{
  public:

    int g_dragX;
    int g_dragY;
    bool g_inDrag;

    C_MyWindow()
    {
      g_dragX = 0;
      g_dragY = 0;
      g_inDrag = false;
    }

    void display()
    {
      (void)glGetError();
      glEnable( GL_DEPTH_TEST );
      glDepthMask( GL_TRUE );
      glClearDepth( 1 );
      glClearColor( 0, 1, 1, 0 );
    #if !defined( NDEBUG )
      glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
    #else
      glClear( GL_DEPTH_BUFFER_BIT );
    #endif

      assert( !glGetError() );
      updateCameraHeading();
      updateMatrices();

      __int64 tnow;
      QueryPerformanceCounter( (LARGE_INTEGER *)&tnow );
      tnow -= g_lastTime;
      g_lastTime += tnow;
      double deltaTime = g_frequency * (double)tnow;

      GLShader::flushState();
      g_rendering = true;
      g_content->step( float( deltaTime ) );
      g_rendering = false;

      assert( !glGetError() );
      gwSwapBuffers();
      assert( !glGetError() );
    }

    void keyboard( unsigned char key, int x, int y )
    {
      g_content->key( x, y, key );
      if( g_defaultKeyHandling ) {
        switch( key ) {
          case 27:
            exit( 1 );
            break;
          case 'r':
            initCamera();
            break;
        }
      }
    }

    void special( int key, int x, int y )
    {
      g_content->special( x, y, key, true );
      if( g_defaultKeyHandling ) {
        if( key == VK_UP ) {
          float x = g_cameraXyz[0], y = g_cameraXyz[1], z = g_cameraXyz[2];
          g_cameraXyz[0] += (g_cameraLookAt[0]-x) * g_cameraSpeed;
          g_cameraXyz[1] += (g_cameraLookAt[1]-y) * g_cameraSpeed;
          g_cameraXyz[2] += (g_cameraLookAt[2]-z) * g_cameraSpeed;
          updateCameraHeading();
        }
        else if( key == VK_DOWN ) {
          float x = g_cameraXyz[0], y = g_cameraXyz[1], z = g_cameraXyz[2];
          g_cameraXyz[0] -= (g_cameraLookAt[0]-x) * g_cameraSpeed;
          g_cameraXyz[1] -= (g_cameraLookAt[1]-y) * g_cameraSpeed;
          g_cameraXyz[2] -= (g_cameraLookAt[2]-z) * g_cameraSpeed;
          updateCameraHeading();
        }
        else if( key == VK_RIGHT ) {
          //  assume we're using Y = up
          float x = g_cameraXyz[0], y = g_cameraXyz[1], z = g_cameraXyz[2];
          g_cameraXyz[0] -= (g_cameraLookAt[2]-z) * g_cameraSpeed;
          g_cameraXyz[2] += (g_cameraLookAt[0]-x) * g_cameraSpeed;
          updateCameraHeading();
        }
        else if( key == VK_LEFT ) {
          //  assume we're using Y = up
          float x = g_cameraXyz[0], y = g_cameraXyz[1], z = g_cameraXyz[2];
          g_cameraXyz[0] += (g_cameraLookAt[2]-z) * g_cameraSpeed;
          g_cameraXyz[2] -= (g_cameraLookAt[0]-x) * g_cameraSpeed;
          updateCameraHeading();
        }
        else if( key == VK_PRIOR ) {
          //  assume we're using Y = up
          g_cameraXyz[1] += 2 * g_cameraSpeed;
          updateCameraHeading();
        }
        else if( key == VK_NEXT ) {
          //  assume we're using Y = up
          g_cameraXyz[1] -= 2 * g_cameraSpeed;
          updateCameraHeading();
        }
      }
    }

    void specialUp( int key, int x, int y )
    {
      g_content->special( x, y, key, false );
    }

    void timer()
    {
    }

    void mouse( int button, int state, int x, int y )
    {
      if( g_defaultMouseCamera ) {
        g_dragX = x;
        g_dragY = y;
      }
      else {
        if( state == GW_DOWN ) {
          if( !g_inDrag ) {
            g_inDrag = true;
            g_content->beginDrag( x, y, 1U<<button );
          }
        }
        else if( g_inDrag ) {
          g_inDrag = false;
          g_content->endDrag( x, y );
        }
      }
    }

    void activeMotion( int x, int y )
    {
      if( g_defaultMouseCamera ) {
        if( x != g_dragX ) {
          g_cameraHeading += (g_dragX - x);
          if( g_cameraHeading > 180 ) {
            g_cameraHeading -= 360;
          }
          else if( g_cameraHeading < -180 ) {
            g_cameraHeading += 360;
          }
          g_dragX = x;
        }
        if( y != g_dragY ) {
          g_cameraPitch += (g_dragY - y);
          if( g_cameraPitch > 75 ) {
            g_cameraPitch = 75;
          }
          else if( g_cameraPitch < -75 ) {
            g_cameraPitch = -75;
          }
          g_dragY = y;
        }
      }
      else if( g_inDrag ) {
        g_content->dragMove( x, y );
      }
      else {
        g_content->mouseMove( x, y );
      }
    }

    void passiveMotion( int x, int y )
    {
      if( g_inDrag ) {
        g_inDrag = false;
        g_content->endDrag( x, y );
      }
      else {
        g_content->mouseMove( x, y );
      }
    }
};

static   C_MyWindow w;






void IGlutContent::setDefaultKeyHandling( bool def )
{
  g_defaultKeyHandling = def;
}

void IGlutContent::setDefaultMouseCamera( bool def )
{
  g_defaultMouseCamera = def;
}

void IGlutContent::mouseMove( int, int )
{
}
void IGlutContent::endDrag( int, int )
{
}
void IGlutContent::dragMove( int, int )
{
}
void IGlutContent::beginDrag( int, int, unsigned int )
{
}
void IGlutContent::key( int, int, unsigned char )
{
}

void IGlutContent::special( int, int, int, bool )
{
}

void IGlutContent::getCamera( float * outXyz, float * outLookAt, float * outHeading, float * outPitch )
{
  memcpy( outXyz, g_cameraXyz, sizeof( g_cameraXyz ) );
  memcpy( outLookAt, g_cameraLookAt, sizeof( g_cameraLookAt ) );
  *outHeading = g_cameraHeading;
  *outPitch = g_cameraPitch;
}

void IGlutContent::setCamera( float const * xyz, float const heading, float const pitch )
{
  memcpy( g_cameraXyz, xyz, sizeof( g_cameraXyz ) );
  g_cameraHeading = heading;
  g_cameraPitch = pitch;
  updateCameraHeading();
  if( g_rendering ) {
    updateMatrices();
  }
}

void IGlutContent::setDefaultCameraSpeed( float s )
{
  g_cameraSpeed = s;
}

void IGlutContent::grabMouse( bool grabbed )
{
  if( grabbed != g_mouseGrabbed ) {
    gwGrabMouse( &w, grabbed );
    g_mouseGrabbed = grabbed;
  }
}






bool options[ 128 ];

static void
putOptions( int argc, char ** argv )
{
  for( int ix = 0; ix < argc; ++ix ) {
    if( argv[ix][0] == '-' && argv[ix][1] != '-' ) {
      char * p = &argv[ix][1];
      while( *p ) {
        options[*p & 0x7f] = true;
        ++p;
      }
    }
    else {
      char * eq = strchr( argv[ ix ], '=' );
      if( eq ) {
        *eq = 0;
        ++eq;
        for( int o = 0; o < sizeof(intValues)/sizeof(intValues[0]); ++o ) {
          if( !stricmp( intValues[o].name, argv[ix] ) ) {
            intValues[o].val = atoi(eq);
            printf( "setting %s to %d\n", intValues[o].name, intValues[o].val );
            goto foundit;
          }
        }
        for( int o = 0; o < sizeof(charValues)/sizeof(charValues[0]); ++o ) {
          if( !stricmp( charValues[o].name, argv[ix] ) ) {
            charValues[o].val = eq;
            printf( "setting %s to %s\n", charValues[o].name, eq );
            goto foundit;
          }
        }
      }
      printf( "unknown option: %s\n", argv[ix] );
foundit:
      ;
    }
  }
}

bool
hasOption( char ch )
{
  return options[ ch & 0x7f ];
}


void
goBackToNormalScreen()
{
  ChangeDisplaySettings( 0, 0 );
}

bool ChdirToSentinel( char const * sentinel )
{
  assert( strlen( sentinel ) < 64 );
  HMODULE cur = GetModuleHandle( NULL );
  char path[ 256+64 ] = "";
  GetModuleFileName( cur, path, sizeof( path ) );
  while( true ) {
    char * slash = strrchr( path, '\\' );
    if( !slash ) {
      path[0] = 0;
      break;
    }
    strcpy( slash+1, sentinel );
    if( GetFileAttributes( path ) != INVALID_FILE_ATTRIBUTES ) {
      slash[1] = 0;
      info( "Found sentinel '%s' at path '%s'\n", sentinel, path );
      break;
    }
    *slash = 0;
  }
  if( !path[0] ) {
    info( "Could not find sentinel '%s'; leaving CWD un-changed.\n", sentinel );
    return false;
  }
  return SetCurrentDirectory( path ) != 0;
}


extern void UnittestScene();
extern void UnittestGL_ASE();
static HINSTANCE hInstance;

int main(int argc, char** argv)
{
#if !defined( NDEBUG )
  initLog( "log.txt", LogFilterNone, gwAddOverlayTextStr );
#else
  initLog( "log.txt", LogFilterDebug, gwAddOverlayTextStr );
#endif

  ChdirToSentinel( "media" );
#if !defined( NDEBUG )
  UnittestScene();
  UnittestGL_ASE();
#endif

  RECT r = { 0, 0, 0, 0 };
  AdjustWindowRectEx( &r, WS_OVERLAPPEDWINDOW, false, WS_EX_OVERLAPPEDWINDOW );
  
  putOptions( argc, argv );
  gwInit( hInstance, &argc, argv);

  if( charOption( "userName" ) ) {
    SetUserName( charOption( "userName" ) );
  }

  WINWIDTH = float( intOption( "winSizeX" ) );
  WINHEIGHT = float( intOption( "winSizeY" ) );
  ASPECT = WINHEIGHT/WINWIDTH;
  if( options['w'] ) {
// this doesn't seem to work -- can't find documentation anywhere
//    glutGameModeString( "800x600@60" );
//    glutEnterGameMode();
    gwCreateWindow( "Carworld", 0, 0, int(WINWIDTH), int(WINHEIGHT), &w );
  }
  else {
    gwCreateWindow( "Carworld", 96, 32, int(WINWIDTH), int(WINHEIGHT), &w );
  }
  assert( !glGetError() );
  //  -w for full-screen
  BOOL b = TRUE;
  if( options[ 'w' ] ) {
    DEVMODE dm;
    memset( &dm, 0, sizeof( dm ) );
    dm.dmSize = sizeof( DEVMODE );
    dm.dmFields = DM_PELSWIDTH | DM_PELSHEIGHT;
    dm.dmBitsPerPel = 32;
    dm.dmPelsWidth = int(WINWIDTH);
    dm.dmPelsHeight = int(WINHEIGHT);
    dm.dmDisplayFlags = 0;
    dm.dmDisplayFrequency = 60;
    b = ChangeDisplaySettings( &dm, CDS_FULLSCREEN );
    if( b ) {
      atexit( goBackToNormalScreen );
      gwPositionWindow( &w, 0, 0 );
      gwReshapeWindow( &w, int(WINWIDTH), int(WINHEIGHT) );
    }
  }
  initCamera();
  assert( !glGetError() );
  g_content = MakeGlutContent();
  assert( !glGetError() );
  g_content->createScene();
  assert( !glGetError() );

  QueryPerformanceFrequency( (LARGE_INTEGER *)&g_lastTime );
  g_frequency = 1.0 / (double)g_lastTime;
  QueryPerformanceCounter( (LARGE_INTEGER *)&g_lastTime );

  assert( !glGetError() );

  if( !b ) {
    info( "ChangeDisplaySettings(): cannot go fullscreen?\n" );
  }

  gwPostRedisplay( &w );
  gwMainLoop();

  return 0;
}


static int
splitCmdLine( char const * cmdLine, char ** argv, int n )
{
  int args = 0;
  while( n > 1 ) {
    while( *cmdLine && isspace( *cmdLine ) ) {
      ++cmdLine;
    }
    if( !*cmdLine ) break;
    char const * lStart;
    char const * lEnd;
    if( *cmdLine == '\"' ) {
      lStart = cmdLine+1;
      lEnd = strchr( lStart, '\"' );
    }
    else {
      lStart = cmdLine;
      lEnd = strchr( lStart, ' ' );
    }
    if( !lEnd ) {
      lEnd = lStart + strlen( lStart );
    }
    *argv = new char[ lEnd-lStart+1 ];
    memcpy( *argv, lStart, lEnd-lStart );
    (*argv)[ lEnd-lStart ] = 0;
    --n;
    ++args;
    ++argv;
    if( *lEnd ) {
      cmdLine = lEnd+1;
    }
    else {
      cmdLine = lEnd;
    }
  }
  *argv = 0;
  return args;
}

int __stdcall
WinMain(
  HINSTANCE cur,
  HINSTANCE prev,
  LPSTR     cmdLine,
  int       cmdShow )
{
  hInstance = cur;
  char * argv[ 100 ];
  char progName[ 32 ] = "carworld";
  argv[ 0 ] = progName;
  int argc = 1 + splitCmdLine( cmdLine, &argv[ 1 ], 98 );
  return main( argc, argv );
}


