
//#version 110

// varying goes here
varying vec4 g_FragPos;
varying vec4 g_VertPos;

#section context


#section defaults

diffuseMap "media/ground.dds"
roadMap "media/volcano-roads.dds"
redMap "media/lavacracks.dds"
greenMap "media/speckledrock.dds"
blueMap "media/bluerock.dds"
coverMap "media/volcano-cover.dds"

roadS 0.0009765625 0 0 0.5
roadT 0 0 -0.0009765625 -0.5

#section vertex-shader

void main()
{
  gl_Position = ftransform();
  g_VertPos = gl_Vertex;
  g_FragPos = gl_Position;
  gl_TexCoord[0] = gl_MultiTexCoord0;
  gl_FrontColor = gl_Color * (gl_LightSource[0].ambient + 
      gl_LightSource[0].diffuse * vec4( clamp( 
          dot( normalize( gl_NormalMatrix * gl_Normal ), 
               gl_LightSource[0].position.xyz ), 
          0.0, 
          1.0 ) ) );
}



#section fragment-shader

uniform sampler2D diffuseMap;
uniform sampler2D roadMap;
uniform sampler2D redMap;
uniform sampler2D greenMap;
uniform sampler2D blueMap;
uniform sampler2D coverMap;
uniform vec4 roadS;
uniform vec4 roadT;

void main()
{
  vec4 groundColor = texture2D(
      diffuseMap,
      gl_TexCoord[0].xy );
  vec4 redColor = texture2D(
      redMap,
      gl_TexCoord[0].xy );
  vec4 greenColor = texture2D(
      greenMap,
      gl_TexCoord[0].xy );
  vec4 blueColor = texture2D(
      blueMap,
      gl_TexCoord[0].xy );
  vec4 roadColor = texture2D( 
      roadMap,
      gl_TexCoord[0].xy );


  //  road decal
  vec2 roadCoord;
  roadCoord.x = dot( g_VertPos, roadS );
  roadCoord.y = dot( g_VertPos, roadT );

  vec4 coverColor = texture2D( 
      coverMap,
      roadCoord );
  float roadAlpha = texture2D(
      roadMap,
      roadCoord ).a;

  vec4 degree = vec4( 1.0, 1.0, 1.0, 1.0 );
  vec4 tint = coverColor.r * redColor;
  degree *= (1.0-coverColor.r);
  tint += degree * coverColor.g * greenColor;
  degree *= (1.0-coverColor.g);
  tint += degree * coverColor.b * blueColor;
  degree *= (1.0-coverColor.b);
  tint += degree * groundColor;

  vec4 color = mix(
      tint,
      roadColor,
      roadAlpha );

  //  linear fog
  float fogDepth = length( g_FragPos.z );
  fogDepth = clamp( 
      gl_Fog.scale * (fogDepth - gl_Fog.start),
      0.0,
      1.0 );

  //  calculate final color
  gl_FragColor = mix(
      gl_Color * color,
      gl_Fog.color,
      fogDepth );
}
