
varying vec4 g_FragPos;

#section context

blend disable
depthTest disable
alphaTest disable
depthWrite disable
color 1 1 1 1

#section defaults

noiseMap "media/multinoise.tga"
wind 0
components 1 0.5 0.25

#section vertex-shader

uniform float wind;

void main()
{
  gl_Position = ftransform();
  g_FragPos = gl_Vertex;
  vec4 w = wind + gl_MultiTexCoord0;
  gl_TexCoord[0] = w * vec4( 1.0 );
  gl_TexCoord[1] = w * vec4( 3.0, 3.0, 3.0, 1.0 );
  gl_TexCoord[2] = w * vec4( 9.0, 9.0, 9.0, 1.0 );
}


#section fragment-shader

uniform sampler2D noiseMap;
uniform vec3 components;

// light blue for sunwise
const vec4 blue1 = vec4( 0.8, 0.9, 1.0, 1.0 );
// dark blue for anti-sunwise
const vec4 blue2 = vec4( 0.5, 0.55, 1.0, 1.0 );
// white for fluff
const vec4 white = vec4( 1.0, 1.0, 1.0, 1.0 );

void main()
{
  vec3 tex1;

  tex1.x = float( texture2D(
      noiseMap,
      gl_TexCoord[0].xy ).x );

  tex1.y = float( texture2D(
      noiseMap,
      gl_TexCoord[1].xy ).y );

  tex1.z = float( texture2D(
      noiseMap,
      gl_TexCoord[2].xy ).x );

  float fluff = dot(
      tex1,
      components )
      * 2.0 
      - 2.2;

  float whiteness = clamp( 
      fluff,
      0.0,
      1.0 );

  float blueScale = dot(
      normalize(
          g_FragPos.xyz ),
      gl_LightSource[0].position.xyz )
      * 0.5
      + 0.5;
  blueScale *= blueScale;

  vec4 blue = mix(
      blue2,
      blue1,
      blueScale );

  gl_FragColor = mix(
      blue,
      white,
      whiteness );
}


