

//  The fact that I have this header is pretty sad...
//  At least some of the extern declarations should go elsewhere.

#if !defined( misc_h )
#define misc_h

#include "mymath.h"


//  RDTSC reads the CPU time stamp counter, which increments by one 
//  for each cycle executed by the CPU clock. This is usually a very 
//  stable clock, but on laptops with SpeedStep(tm), it may vary over 
//  time.
__declspec(naked) inline __int64 rdtsc() {
  __asm {
    rdtsc
    ret
  }
}



//  Hermite4 -- cubic hermite interpolation using 4 input values and 
//  a parameter in the range [0,1], which will interpolate between 
//  input[1] and input[2]. The interpolation is not convex; it may 
//  extend beyond the values of the inputs.

//  Use type traits to define the intermediate type as "float" for most types.
template< typename T > class hermite4_traits {
public:  typedef float float_t;
};
//  Partially specialize to get doubles for some types.
template<> class hermite4_traits<double> {
public:  typedef double float_t;
};
//  Might want to specialize for int and long (and unsigned ditto), too,
//  to avoid possible bit loss...

template< typename T >
inline float hermite4( typename hermite4_traits<T>::float_t frac, T * ptr )
{
  //  promote all expressions to float
  const typename hermite4_traits<T>::float_t    c     = ((typename hermite4_traits<T>::float_t)ptr[2] - ptr[0]) * 0.5f;
  const typename hermite4_traits<T>::float_t    v     = (typename hermite4_traits<T>::float_t)ptr[1] - ptr[2];
  const typename hermite4_traits<T>::float_t    w     = (typename hermite4_traits<T>::float_t)c + v;
  const typename hermite4_traits<T>::float_t    a     = (typename hermite4_traits<T>::float_t)w + v + (ptr[3] - ptr[1]) * 0.5f;
  const typename hermite4_traits<T>::float_t    b     = (typename hermite4_traits<T>::float_t)w + a;
  return ((((a * frac) - b) * frac + c) * frac + ptr[1]);
}

template< typename T >
typename hermite4_traits<T>::float_t hermite4_2D( T const * data, uint pitch, 
    uint x, uint z, float xp, float zp )
{
  assert( xp < 1 && zp < 1 );
  assert( xp >= 0 && zp >= 0 );
  assert( x > 0 && z > 0 );
  hermite4_traits<T>::float_t xx[ 4 ];
  hermite4_traits<T>::float_t yy[ 4 ];
  for( int iz = 0; iz < 4; ++iz ) {
    T const * data2 = data + x - 1 + (iz + z - 1) * pitch;
    for( int ix = 0; ix < 4; ++ix ) {
      xx[ ix ] = typename hermite4_traits<T>::float_t( data2[ ix ] );
    }
    yy[ iz ] = hermite4( xp, xx );
  }
  typename hermite4_traits<T>::float_t sv = hermite4( zp, yy );
  return sv;
}

template< typename T >
T clamp( T a, T min, T max ) {
  return (a < min) ? min : (a > max) ? max : a;
}


//  OK, it's lame to put THIS in here, but ...
extern void debugVar( int x, int y, char const * fmt, ... );
Vector3 RayCastAgainstWorld( Vector3 const & from );

#define strcasecmp stricmp 
inline int stricmp( char const * a, char const * b )
{
  while( *a && *b && (tolower(*a) == tolower(*b)) ) {
    ++a;
    ++b;
  }
  return *a - *b;
}

inline int strincmp( char const * a, char const * b, size_t l )
{
  while( l > 0 ) {
    int d = tolower(*a)-tolower(*b);
    if( d != 0 ) {
      return d;
    }
    ++a;
    ++b;
    --l;
  }
  return 0;
}

inline char const * strnistr( char const * haystack, char const * needle, size_t n )
{
  size_t nl = strlen( needle );
  char tln = tolower( *needle );
  while( n >= nl ) {
    if( (tolower(*haystack) == tln) && 
        !strincmp( haystack, needle, nl ) ) {
      return haystack;
    }
    ++haystack;
    --n;
  }
  return 0;
}

class SoundManager;
extern SoundManager * gSoundManager;

class I_GL_ASE;
I_GL_ASE * ASEFromFile( char const * path );

class GLShaderFactory;
extern GLShaderFactory * gShaderFactory;

char const * UserName();
void SetUserName( char const * name );

#endif
