

#include "glshaderstate.h"
#include "soundmanager.h"
#include "misc.h"

#include <assert.h>
#include <set>
#include <GL/GL.h>

#include "myexplodable.h"


namespace {

  struct ExplosionEffect {
    bool active;
    Vector3 pos;
    float age;
  };

  enum {
    MAX_EXPLOSIONS = 8,
  };
  ExplosionEffect explosionEffects[MAX_EXPLOSIONS];
}


void ExplodableSoundBase::setGunshotSound( SoundHandle gss )
{
  assert( gss != 0 );
  gunshotSound_ = gss;
}

SoundHandle ExplodableSoundBase::gunshotSound_;


extern "C" {
  void ExplodeCallback( void * param, dGeomID g1, dGeomID g2 )
  {
    assert( !dGeomGetBody( g2 ) );
    assert( dGeomGetClass( g2 ) == dSphereClass );
    if( dGeomIsSpace( g1 ) ) {
      dSpaceCollide2( g1, g2, param, ExplodeCallback );
    }
    else {
      dContactGeom cg;
      int nc = dCollide( g1, g2, 1, &cg, sizeof( cg ) );
      if( nc ) {
        MySystem * ms = (MySystem *)dGeomGetData( g1 );
        if( ms ) {
          ((std::set< MySystem * > *)param)->insert( ms );
        }
      }
    }
  }
}


void MakeExplosionEffect( Vector3 const & pos )
{
  float foundAge = 0;
  int foundIx;
  for( int ix = 0; ix < MAX_EXPLOSIONS; ++ix ) {
    if( !explosionEffects[ix].active ) {
      foundIx = ix;
      goto do_explosion;
    }
    if( explosionEffects[ix].age > foundAge ) {
      foundIx = ix;
      foundAge = explosionEffects[ix].age;
    }
  }
do_explosion:
  explosionEffects[foundIx].active = true;
  explosionEffects[foundIx].pos = pos;
  explosionEffects[foundIx].age = 0;
}


void DrawExplosionEffects( float deltaTime, GLShader * explosionShader )
{
  float SIZE = 6;
  glDepthMask( GL_FALSE );
  float mx[16];
  glGetFloatv( GL_MODELVIEW_MATRIX, mx );
  Vector3 x( mx[0], mx[4], mx[8] );
  Vector3 y( mx[1], mx[5], mx[9] );
  Vector3 z( mx[2], mx[6], mx[10] );
  ExplosionEffect * xf = explosionEffects;
  while( xf < &explosionEffects[MAX_EXPLOSIONS] ) {
    if( xf->active ) {
      if( xf->age > 0.7 ) {
        xf->active = false;
      }
      else {
        unsigned char ageColor = unsigned char( 255 - 300 * xf->age );
        explosionShader->context().color = GLColor( ageColor, ageColor, ageColor, ageColor );
        explosionShader->apply();
        glBegin( GL_QUADS );
        glTexCoord2f( 0, 0 );
        Vector3 v1( xf->pos.x + (xf->age*SIZE+1)*(-x.x-y.x),
            xf->pos.y + (xf->age*SIZE+1)*(-x.y-y.y),
            xf->pos.z + (xf->age*SIZE+1)*(-x.z-y.z) );
        glVertex3fv( &v1.x );
        glTexCoord2f( 1, 0 );
        Vector3 v2( xf->pos.x + (xf->age*SIZE+1)*(x.x-y.x),
            xf->pos.y + (xf->age*SIZE+1)*(x.y-y.y),
            xf->pos.z + (xf->age*SIZE+1)*(x.z-y.z) );
        glVertex3fv( &v2.x );
        glTexCoord2f( 1, 1 );
        Vector3 v3( xf->pos.x + (xf->age*SIZE+1)*(x.x+y.x),
            xf->pos.y + (xf->age*SIZE+1)*(x.y+y.y),
            xf->pos.z + (xf->age*SIZE+1)*(x.z+y.z) );
        glVertex3fv( &v3.x );
        glTexCoord2f( 0, 1 );
        Vector3 v4( xf->pos.x + (xf->age*SIZE+1)*(-x.x+y.x),
            xf->pos.y + (xf->age*SIZE+1)*(-x.y+y.y),
            xf->pos.z + (xf->age*SIZE+1)*(-x.z+y.z) );
        glVertex3fv( &v4.x );
        glEnd();
        assert( !glGetError() );
        xf->age += deltaTime;
      }
    }
    ++xf;
  }
  glDepthMask( GL_TRUE );
}

