
#if !defined( myexplodable_h )
#define myexplodable_h

#include <set>
#include <assert.h>

#include "mysystem.h"
#include "soundmanager.h"
#include "misc.h"


void MakeExplosionEffect( Vector3 const & pos );
void DrawExplosionEffects( float deltaTime, GLShader * explosionShader );

class ExplodableSoundBase {
  public:
    static void setGunshotSound( SoundHandle gss );
  protected:
    static SoundHandle gunshotSound_;
};

extern "C" void ExplodeCallback( void *, dGeomID, dGeomID );

template< class S >
class Explodable : public S, public ExplodableSoundBase {
  protected:

    struct ExplodeParams {
      float radius_;
      float force_;
      Vector3 pos_;
    };

    Explodable() {
      explodeParams_.force_ = 0;
      explodeParams_.radius_ = 0;
      explodeParams_.pos_ = Vector3::Null;
    }

    void explode() {
      //  this assert may fail if you're making an explosion at exactly the origin.
      //todo: create Vector3::Invalid instead
      assert( !(explodeParams_.pos_ == Vector3::Null) );
      assert( explodeParams_.force_ > 0 );
      assert( explodeParams_.radius_ > 0 );
      dGeomSetBody( geoms_[0], 0 );
      explodeParams_.pos_ = P();
      dGeomSetPosition( MySystem::utilSphere(), explodeParams_.pos_.x, explodeParams_.pos_.y, explodeParams_.pos_.z );
      dGeomSphereSetRadius( MySystem::utilSphere(), explodeParams_.radius_ );
      std::set< MySystem * > toExplode;
      dSpaceCollide2( (dGeomID)MySystem::worldSpace(), MySystem::utilSphere(), &toExplode, ExplodeCallback );
      std::for_each( toExplode.begin(), toExplode.end(), AffectByExplosion( explodeParams_ ) );
      MakeExplosionEffect( explodeParams_.pos_ );
      PlayParameters pp;
      pp.autoStop = true;
      pp.looping = false;
      pp.pitch = ((rand() & 0xff) * 0.001f + 0.5f) * 0.8f;
      pp.volL = 0.5f;
      pp.volR = 0.5f;
      gSoundManager->playInstance( gunshotSound_, &pp );
      AddDeferredDelete( this );
    }

    class AffectByExplosion {
      public:
        AffectByExplosion( ExplodeParams const & ep ) : ep_( ep ) {}
        void operator()( MySystem * ms )
          {
            ms->affectByExplosion( ep_.pos_, ep_.radius_, ep_.force_ );
          }
        ExplodeParams const & ep_;
    };

    ExplodeParams explodeParams_;
};



#endif  //  myexplodable_h
