
#if !defined( myfilesystem_h )
#define myfilesystem_h

#include "xmlscanner.h"

class Data {
  public:
    Data( void * ptr, size_t size );
    Data( void const * ptr, size_t size );
    Data( void const * ptr, size_t size, bool copy );
    Data( size_t size );
    Data();
    ~Data();
    //  Data implements data COPY semantics (which is kind-of hard 
    //  if I ever re-target Data to point at a file...)
    Data( Data const & other );
    Data & operator=( Data const & other );

    size_t size() const;
    char * writeBase();
    char const * readBase() const;
    size_t pos() const;
    bool setPos( size_t pos );
    size_t read( void * buf, size_t amount );
    size_t write( void const * buf, size_t amount );
    bool trim( size_t newSmallerSize );

    bool ownsPtr() const;
    bool canWrite() const;
    void * detach();
    Data & allocate( size_t size );
    Data & attach( void * ptr, size_t size );
    Data const & attach( void const * ptr, size_t size );

  private:
    void * ptr_;
    size_t size_;
    size_t pos_;
    bool ownsPtr_;
    bool canWrite_;
};


namespace FS {

  //  Handles is an XML document which can map virtual file names 
  //  to physical paths. All handles loaded are merged, until such 
  //  time as you call clearHandles(). Everything is lowercased 
  //  internally (handles and paths).
  void clearHandles();
  bool addHandles( char const * path );
  bool addHandles( Data const & iData );
  bool addHandles( XMLNode * node );
  bool translateHandle( char const * handle, char * buf, size_t bufSize );
  bool readFileByHandle( char const * token, Data * oData );
  bool readFileByPath( char const * file, Data * oData );
  //  writeFileByPath() will first create a .bak file.
  bool writeFileByPath( char const * file, Data const & iData );

};

#endif  //  myfilesystem_h
