
#if !defined( myparticles_h )
#define myparticles_h

#include "mymath.h"
#include "glshaderstate.h"

#include <ode/ode.h>
#include <vector>

//  This is a dead stupid particle system, to get some smoke and 
//  dust into the simulation. I could add an arbitrary number of 
//  features (like texture rotation, flipbook, color animation, 
//  various random distribution parameters, affects on the world 
//  from particles, particle chaining, ...) but I won't.

struct ParticleOp {
  ParticleOp() {
    alphaDelta = -0.001f;
    alphaTarget = 0;
    sizeDelta = 0.001f;
    sizeTarget = 2;
    deltaVperSec = Vector3( 0, -3, 0 );
    maxDuration = 10;
    speedDampening = 0;
    sizeDeltaSpeedWeight = 0;
    spin = 0;
  }
  float alphaDelta;
  float alphaTarget;
  float sizeDelta;
  float sizeTarget;
  Vector3 deltaVperSec;
  float speedDampening;
  float sizeDeltaSpeedWeight;
  float maxDuration;
  float spin;

  //todo: move this out of public struct
  float temp;   //  used by particle system internally
  float spinS;
  float spinC;
};

struct ParticleSettings {
  ParticleSettings() {
    maxParticlesToEmit = 20;
    particleCount = 0;
    material = 0;
    emitterRate = 10;
    emitterVectors[0] = Vector3::X;
    emitterVectors[1] = Vector3::Y;
    emitterVectors[2] = Vector3::Z;
    emitterMinMax[0][0] = -0.5f;
    emitterMinMax[0][1] = 0.5f;
    emitterMinMax[1][0] = 0.5f;
    emitterMinMax[1][1] = 1.0f;
    emitterMinMax[2][0] = -0.5f;
    emitterMinMax[2][1] = 0.5f;
    emitterSize = 0.1f;
    emitterAlpha = 1.0f;
    emitterAlphaVariation = 0;
    emitterSpinMin = -3.14f;
    emitterSpinMax = 3.14f;
    particleOps = 0;
    numParticleOps = 0;
    dieOnContact = false;
    recycleParticles = true;
    autoDeath = true;
  }
  size_t maxParticlesToEmit;
  GLShader * material;
  float emitterRate;
  Vector3 emitterVectors[3];
  float emitterMinMax[3][2];
  float emitterSize;
  float emitterAlpha;
  float emitterAlphaVariation;
  float emitterSpinMin;
  float emitterSpinMax;
  ParticleOp const * particleOps;
  size_t numParticleOps;
  bool dieOnContact;        //  dieOnContact is EXPENSIVE (uses one ray per particle)
  bool recycleParticles;
  bool autoDeath;           //  if TRUE, dies when decays to 0 particles
  size_t particleCount;     //  if recycleParticles is not true, must be 0
};

class MyParticles : private ParticleSettings {

  public:

    MyParticles( Vector3 const & pos, ParticleSettings const * ps, dSpaceID space );
    ~MyParticles();

    static void stepAll( float delta );
    static void renderAll();

    bool emit( Vector3 const & pos, Vector3 const & vel );
    bool step( float delta );     //  returns false when there's nothing to render
    void render();

    void setEmitterPos( Vector3 const & pos, Vector3 const (& basis)[3] );
    void setEmitterAlpha( float nuAlpha );
    void setEmitterRate( float nuRate );
    void setEmitting( bool emitting ) { emitting_ = emitting; }

  private:

    MyParticles( MyParticles const & o );
    MyParticles & operator=( MyParticles const & o );

    struct Particle {
      ParticleOp const * curOp;
      float curOpDuration;
      Vector3 pos;
      Vector3 vel;
      float alpha;
      float size;
      float xPhase;
      float yPhase;
    };
    Vector3 basePos_;
    size_t baseParticle_;
    size_t liveParticles_;
    size_t particlesEmitted_;
    size_t particleCount_;
    Particle * particles_;
    GLuint particleBuffer_;
    float particleFlux_;
    bool emitting_;

    void killParticle( Particle & p );
    void initParticle( Particle & p, Vector3 const & pos, Vector3 const & vel );

    static dGeomID particleRay_;
    static std::vector< MyParticles * > allParticleSystems_;
    static std::vector< MyParticles * > toRemove_;
};


MyParticles * MakeParticleSystem( char const * handleName, dSpaceID space );


#endif  //  myparticles_h
