

#include "mytank.h"
#include "mytweakables.h"
#include "misc.h"
#include "myparticles.h"
#include "terrain.h"
#include "gl-ase.h"
#include "myfilesystem.h"
#include "glshaderstate.h"
#include "mygrenade.h"
#include "myrocket.h"
#include "mymine.h"


#define RENDER_TANK_CONTACTS 0

//  todo: animate treads based on contact
//  todo: animate tower based on mouse input
//  done: animate texture based on movement

TW_FLOAT(tank,TANK_GAS_FORCE,3.3f);
TW_FLOAT(tank,TANK_TURN_FORCE,50);
TW_FLOAT(tank,TANK_TURN_DAMPENING,10);
TW_FLOAT(tank,TANK_SHAKE_DAMPENING,3);
TW_FLOAT(tank,TANK_STABILIZING_FORCE,30);
TW_FLOAT(tank,TANK_SPRING_CONST,40);
TW_FLOAT(tank,TANK_MU,0.4f);
TW_FLOAT(tank,TANK_MU2,0.8f);
TW_FLOAT(tank,TANK_TREAD_MU,0.1f);
TW_FLOAT(tank,TANK_TREAD_MU2,0.6f);
TW_FLOAT(tank,TANK_TREAD_SLIP,0.3f);
TW_FLOAT(tank,TANK_V_DAMPENING,2.0f);
TW_FLOAT(tank,TANK_SPRING_VEL_SCALE,0.03f);
TW_FLOAT(tank,TANK_STAY_FORCE,70);
TW_FLOAT(tank,TANK_JUMP_FORCE,500);
TW_FLOAT(tank,TANK_ANTIJUMP_VEL,10);
TW_FLOAT(tank,TANK_CONTACT_FORCE_THRESHOLD,5);
TW_FLOAT(tank,TANK_RAY_LENGTH,1.4f);
TW_FLOAT(tank,TANK_TREAD_DEPTH,0.35f);
TW_FLOAT(tank,TANK_TREAD_SPEED,0.20f);

TW_FLOAT(tank,MISSILE_TURRET_OFFSET,1.0f);
#define MISSILE_TURRET_FRONT (MISSILE_TURRET_OFFSET+ROCKET_LENGTH/2+ROCKET_RADIUS)
TW_FLOAT(tank,MISSILE_TURRET_UP,0.65f);
TW_FLOAT(tank,MISSILE_TURRET_LEFT,0.6f);

#if RENDER_TANK_CONTACTS
Vector3 lines_[100];
Vector3 normals_[100];

void DebugRenderTankLines()
{
  glDisable( GL_DEPTH_TEST );
  glEnable( GL_BLEND );
  glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
  glDisable( GL_TEXTURE_2D );
  glBegin( GL_LINES );
  for( int i = 0; i < 100; ++i ) {
    float s = (100.0f-i)/100.0f;
    glColor4f( 1.0f*s, 0.0f, 0.5f*s, s );
    glVertex3f( lines_[i].x, lines_[i].y, lines_[i].z );
    glVertex3f( lines_[i].x+normals_[i].x, lines_[i].y+normals_[i].y, lines_[i].z+normals_[i].z );
  }
  glEnd();
  glEnable( GL_DEPTH_TEST );
  assert( !glGetError() );
  gwFlushState();
}
#endif

static dGeomID hackTankTurretGeom;

MyTank::MyTank( Params const & p )
{
  gas_ = turn_ = jumpForce_ = 0;
  still_ = true;
  contact_ = true;
  elevation_ = 0;

  lTread_ = 0;
  lTreadParam_ = -1;
  lTreadOffset_ = 0;
  rTread_ = 0;
  rTreadParam_ = -1;
  rTreadOffset_ = 0;
  gunPart_ = 0;
  turretPart_ = 0;
  launcherPart_ = 0;
  uint mCount = p.body->numMaterials();
  for( uint i = 0; i < mCount; ++i ) {
    char const * mtlName = p.body->materialNameAt( i );
    GLShader * shader;
    if( !stricmp( mtlName, "treadsL" ) ) {
      lTread_ = i;
      shader = gShaderFactory->newShader( "treadsShader" );
      lTreadParam_ = shader->instance()->parameterIndex( "treadOffset" );
    }
    else if( !stricmp( mtlName, "treadsR" ) ) {
      rTread_ = i;
      shader = gShaderFactory->newShader( "treadsShader" );
      rTreadParam_ = shader->instance()->parameterIndex( "treadOffset" );
    }
    else if( !stricmp( mtlName, "gun" ) ) {
      gunPart_ = i;
      shader = gShaderFactory->newShader( "tankShader" );
    }
    else if( !stricmp( mtlName, "tower" ) ) {
      turretPart_ = i;
      shader = gShaderFactory->newShader( "tankShader" );
    }
    else if( !stricmp( mtlName, "launcher" ) ) {
      launcherPart_ = i;
      shader = gShaderFactory->newShader( "tankShader" );
    }
    else {
      shader = gShaderFactory->newShader( "tankShader" );
    }
    materials_.push_back( shader );
    info( "Tank material %d is %s\n", i, mtlName );
    char path[256];
    char const * diffuse = p.body->materialDiffuseMapAt( i );
    if( !diffuse || !diffuse[0] ) {
      continue;
    }
    char handle[132];
    char const * sl = strrchr( diffuse, '\\' );
    if( !sl ) sl = strrchr( diffuse, '/' );
    if( !sl ) sl = diffuse; else ++sl;
    strncpy( handle, sl, 128 );
    handle[127] = 0;
    sl = strrchr( handle, '.' );
    if( sl ) {
      *(char *)sl = 0;
    }
    if( !FS::translateHandle( handle, path, sizeof(path) ) ) {
      warning( "Missing filemap handle for '%s'\n", handle );
      sprintf( path, "media/%s.dds", handle );
    }
    shader->instance()->setParameterValueAt( 
        shader->instance()->parameterIndex( "diffuseMap" ), 
        GLProgramValue( TextureRef( path ) ) );
  }

  //  chassis
  chassisBody_ = dBodyCreate( world() );
  dMass m;
  dMassSetBoxTotal( &m, 10, 2, 1, 3.0f );
  dBodySetMass( chassisBody_, &m );
  dBodySetAutoDisableFlag( chassisBody_, 0 );
  addBody( chassisBody_, 0, p.body );
  dGeomID tempGeom = dCreateBox( worldSpace(), 3, 1.5f, 4.5f );
  addGeom( tempGeom, chassisBody_ );
  dGeomSetCategoryBits( tempGeom, 2 );
  dGeomSetCollideBits( tempGeom, ~2 );

#define TURRET_HEIGHT 0.9f
#define TURRET_FRONT -0.7f
#define GUN_SIZE 4.0f
#define GUN_RADIUS 0.3f
#define GUN_FRONT (GUN_SIZE/2)
#define GUN_HEIGHT (TURRET_HEIGHT+0.3f)
#define GUN_LEFT -0.5f
#define GUN_SWIVEL_BACK 0.9f
#define TURRET_GEOM_FRONT -0.3f

  turretBody_ = dBodyCreate( world() );
  dMassSetBoxTotal( &m, 0.3f, 1, 0.5, 1 );
  dBodySetMass( turretBody_, &m );
  dBodySetAutoDisableFlag( turretBody_, 0 );
  dBodySetPosition( turretBody_, 0, TURRET_HEIGHT, TURRET_FRONT );
  addBody( turretBody_, 0, 0 );
  turretDisplacement_ = Vector3( 0, -TURRET_HEIGHT, -TURRET_FRONT );
  tempGeom = dCreateBox( 0, 2.2f, 1.7f, 1.6f );
  dGeomSetPosition( tempGeom, 0, 0, TURRET_GEOM_FRONT );
  dGeomID xlate = dCreateGeomTransform( worldSpace() );
  dGeomTransformSetCleanup( xlate, 1 );
  dGeomTransformSetGeom( xlate, tempGeom );
  dGeomTransformSetInfo( xlate, 1 );
  addGeom( xlate, turretBody_ );
  dGeomSetCategoryBits( xlate, 2 );
  dGeomSetCollideBits( xlate, ~2 );
  hackTankTurretGeom = xlate;

  dJointID turretSwivel = dJointCreateHinge( world(), 0 );
  addJoint( turretSwivel, chassisBody_, turretBody_ );
  dJointSetHingeAnchor( turretSwivel, 0, TURRET_HEIGHT, TURRET_FRONT );
  dJointSetHingeAxis( turretSwivel, 0, 1, 0 );
  dJointSetHingeParam( turretSwivel, dParamLoStop, 0 );
  dJointSetHingeParam( turretSwivel, dParamHiStop, 0 );
  dJointSetHingeParam( turretSwivel, dParamLoStop, 0 );
  dJointSetHingeParam( turretSwivel, dParamStopCFM, 0.02f );
  dJointSetHingeParam( turretSwivel, dParamStopERP, 0.4f );
  dJointSetHingeParam( turretSwivel, dParamFMax, 3.0f );
  turretJoint_ = turretSwivel;

  gunBody_ = dBodyCreate( world() );
  dMassSetCappedCylinderTotal( &m, 0.3f, 3, GUN_RADIUS, GUN_SIZE );
  dBodySetMass( gunBody_, &m );
  dBodySetAutoDisableFlag( gunBody_, 0 );
  Vector3 gpos( GUN_LEFT, GUN_HEIGHT, GUN_FRONT-GUN_SWIVEL_BACK );
  dBodySetPosition( gunBody_, gpos.x, gpos.y, gpos.z );
  addBody( gunBody_, 0, 0 );
  gunDisplacement_ = Vector3( 0, 0, 0 )-gpos;
  tempGeom = dCreateCCylinder( worldSpace(), GUN_RADIUS, GUN_SIZE );
  addGeom( tempGeom, gunBody_ );
  dGeomSetCategoryBits( tempGeom, 2 );
  dGeomSetCollideBits( tempGeom, ~2 );

  dJointID gunSwivel = dJointCreateHinge( world(), 0 );
  addJoint( gunSwivel, turretBody_, gunBody_ );
  dJointSetHingeAnchor( gunSwivel, GUN_LEFT, GUN_HEIGHT, -GUN_SWIVEL_BACK );
  dJointSetHingeAxis( gunSwivel, 1, 0, 0 );
  dJointSetHingeParam( gunSwivel, dParamLoStop, 0 );
  dJointSetHingeParam( gunSwivel, dParamHiStop, 0 );
  dJointSetHingeParam( gunSwivel, dParamLoStop, 0 );
  dJointSetHingeParam( gunSwivel, dParamStopCFM, 0.02f );
  dJointSetHingeParam( gunSwivel, dParamStopERP, 0.4f );
  dJointSetHingeParam( gunSwivel, dParamFMax, 3.5f );
  gunJoint_ = gunSwivel;

  //  treads
  tempGeom = dCreateCCylinder( 0, 0.6f, 3.5f );
  dGeomSetPosition( tempGeom, -1.45f, -TANK_TREAD_DEPTH, 0.0f );
  dGeomSetCategoryBits( tempGeom, 2 );
  dGeomSetCollideBits( tempGeom, ~2 );
  xlate = dCreateGeomTransform( worldSpace() );
  dGeomTransformSetCleanup( xlate, 1 );
  dGeomTransformSetGeom( xlate, tempGeom );
  dGeomTransformSetInfo( xlate, 1 );
  addGeom( xlate, chassisBody_ );
  dGeomSetCategoryBits( xlate, 2 );
  dGeomSetCollideBits( xlate, ~2 );

  tempGeom = dCreateCCylinder( 0, 0.6f, 3.5f );
  dGeomSetPosition( tempGeom, 1.45f, -TANK_TREAD_DEPTH, 0.0f );
  dGeomSetCategoryBits( tempGeom, 2 );
  dGeomSetCollideBits( tempGeom, ~2 );
  xlate = dCreateGeomTransform( worldSpace() );
  dGeomTransformSetCleanup( xlate, 1 );
  dGeomTransformSetGeom( xlate, tempGeom );
  dGeomTransformSetInfo( xlate, 1 );
  addGeom( xlate, chassisBody_ );
  dGeomSetCategoryBits( xlate, 2 );
  dGeomSetCollideBits( xlate, ~2 );

  // dust trail
  terrain_ = p.terrain;
  dust_[0] = MakeParticleSystem( "dustcloud", worldSpace() );
  dust_[1] = MakeParticleSystem( "dustcloud", worldSpace() );
  dust_[0]->setEmitting( false );
  dust_[1]->setEmitting( false );

  //  weapons
  weapon_ = GrenadeWeapon;
  grenadeWeapon_ = NewGrenadeWeapon();
  rocketWeapon_ = NewRocketWeapon();
  mineWeapon_ = NewMineWeapon();

  snapshot_.capture( *this );
}

MyTank::~MyTank()
{
  delete dust_[0];
  delete dust_[1];
  delete grenadeWeapon_;
  delete rocketWeapon_;
  delete mineWeapon_;
}

void MyTank::render()
{
  I_GL_ASE * mesh = meshes_[0].mesh;
  unsigned int ix = 0;
  if( lTreadParam_ >= 0 ) {
    materials_[lTread_]->instance()->setParameterValueAt( lTreadParam_, GLProgramValue( lTreadOffset_ ) );
  }
  if( rTreadParam_ >= 0 ) {
    materials_[rTread_]->instance()->setParameterValueAt( rTreadParam_, GLProgramValue( rTreadOffset_ ) );
  }
  for( std::vector< GLShader * >::iterator mtl = materials_.begin();
      mtl != materials_.end(); ++mtl ) {
    (*mtl)->apply();
    if( ix == gunPart_ ) {
      Matrix4 m4 = ((Matrix3 const *)dBodyGetRotation( gunBody_ ))->transpose();
      m4.pos() = *(Vector3 const *)dBodyGetPosition( gunBody_ );
      glPushMatrix();
      glMultMatrixf( &m4.xx );
      Vector3 const & d = gunDisplacement_;
      glTranslatef( d.x, d.y, d.z );
    }
    else if( ix == turretPart_ ) {
      Matrix4 m4 = ((Matrix3 const *)dBodyGetRotation( turretBody_ ))->transpose();
      m4.pos() = *(Vector3 const *)dBodyGetPosition( turretBody_ );
      glPushMatrix();
      glMultMatrixf( &m4.xx );
      Vector3 const & d = turretDisplacement_;
      glTranslatef( d.x, d.y, d.z );
    }
    else if( ix == launcherPart_ ) {
      Matrix4 m4 = ((Matrix3 const *)dBodyGetRotation( turretBody_ ))->transpose();
      //  rotate the turret by 20% of actual elevation
      Matrix4 m4b = Matrix4( Vector3::X, elevation_ * -0.2f ) * m4;
      m4b.pos() = *(Vector3 const *)dBodyGetPosition( turretBody_ );
      glPushMatrix();
      glMultMatrixf( &m4b.xx );
      Vector3 const & d = turretDisplacement_;
      glTranslatef( d.x, d.y, d.z );
    }
    else {
      Matrix4 m4 = ((Matrix3 const *)R())->transpose();
      m4.pos() = P();
      glPushMatrix();
      glMultMatrixf( &m4.xx );
    }
    mesh->issue( ix );
    ++ix;
    glPopMatrix();
  }

#if 0
  {
    Vector3 lPos = *(Vector3 const *)dBodyGetPosition( turretBody_ );
    Matrix3 lMat = ((Matrix3 const *)dBodyGetRotation( turretBody_ ))->transpose();
    Vector3 left, up, front;
    lMat.leftUpFront( &left, &up, &front );
    glColor4f( 1, 0, 0, 1 );
    glBegin( GL_LINES );
    Vector3 lPos2 = lPos + left;
    glColor4f( 1, 0, 0, 1 );
    glVertex3fv( &lPos.x ); glVertex3fv( &lPos2.x );
    lPos2 = lPos + up;
    glColor4f( 0, 1, 0, 1 );
    glVertex3fv( &lPos.x ); glVertex3fv( &lPos2.x );
    lPos2 = lPos + front;
    glColor4f( 0, 0, 1, 1 );
    glVertex3fv( &lPos.x ); glVertex3fv( &lPos2.x );
    glEnd();

    lPos += front * MISSILE_TURRET_FRONT + up * MISSILE_TURRET_UP +
            left * MISSILE_TURRET_LEFT;
    lPos2 = lPos + front;
    glColor4f( 1, 1, 1, 1 );
    glBegin( GL_LINES );
    glVertex3fv( &lPos.x );
    glVertex3fv( &lPos2.x );
    glEnd();

    Vector3 const & v3 = *(Vector3 const *)dGeomGetPosition( hackTankTurretGeom );
    Matrix3 m3 = ((Matrix3 const *)dGeomGetRotation( hackTankTurretGeom ))->transpose();
    m3.leftUpFront( &left, &up, &front );
    Vector3 lll = v3 +
        *(Vector3 const *)dGeomGetPosition( dGeomTransformGetGeom( hackTankTurretGeom ) ) * m3;
    lll = ((lll - (left * (2.2f/2))) - (up * (1.7f/2))) - (front * (1.6f/2));
    left *= 2.2f;
    up *= 1.7f;
    front *= 2.0f;
    glColor4f( 0, 0, 0, 1 );
    glBegin( GL_LINES );
    Vector3 l2 = lll+left;
    glVertex3fv( &lll.x ); glVertex3fv( &l2.x );
    l2 = lll+up;
    glVertex3fv( &lll.x ); glVertex3fv( &l2.x );
    l2 = lll+front;
    glVertex3fv( &lll.x ); glVertex3fv( &l2.x );
    l2 = lll+left+up+front;
    glVertex3fv( &lll.x ); glVertex3fv( &l2.x );
    lll = l2;
    l2 = lll-left;
    glVertex3fv( &lll.x ); glVertex3fv( &l2.x );
    l2 = lll-up;
    glVertex3fv( &lll.x ); glVertex3fv( &l2.x );
    l2 = lll-front;
    glVertex3fv( &lll.x ); glVertex3fv( &l2.x );
    glEnd();
  }
#endif

  assert( !glGetError() );
}

namespace {
  struct TankRayInfo {
    MyTank * t;
    float d;
  };
}

extern "C" {
  static void TankRayCallback( void * tri, dGeomID g1, dGeomID g2 )
  {
    assert( dGeomGetClass( g2 ) == dRayClass );
    if( dGeomIsSpace( g1 ) ) {
      dSpaceCollide2( g1, g2, tri, TankRayCallback );
    }
    else {
      dContactGeom cg;
      TankRayInfo * t = (TankRayInfo *)tri;
      if( !t->t->hasGeom( g1 ) && dCollide( g2, g1, 1, &cg, sizeof( cg ) ) ) {
        MySystem * s1 = (MySystem *)dGeomGetData( g1 );
        if( !s1 || s1->physical() ) {
          if( t->d > cg.depth ) {
            t->d = cg.depth;
          }
        }
      }
    }
  }
}

void MyTank::rayCollTest( Vector3 const & v, Vector3 const & up, float * outForce )
{
  TankRayInfo tri;
  tri.t = this;
  tri.d = 1.5f;
  dGeomRaySet( utilRay(), v.x, v.y, v.z, -up.x, -up.y, -up.z );
  dSpaceCollide2( utilRay(), (dGeomID)worldSpace(), &tri, &TankRayCallback );
  if( (tri.d >= 1.5f) || (tri.d <= 0.0f) ) {
    *outForce = 0;
    return;
  }
  Vector4 pv;
  dBodyGetPointVel( chassisBody_, v.x, v.y, v.z, &pv.x );
  float va = up / (Vector3 const &)pv;
  *outForce = clamp( TANK_RAY_LENGTH - tri.d - va * TANK_SPRING_VEL_SCALE, 0.0f, 1.0f ) * TANK_SPRING_CONST;
}

void MyTank::preStep()
{
  Vector3 left, up, front;
  Matrix4 const & m4 = *(Matrix4 const *)R();
  m4.transpose().leftUpFront( &left, &up, &front );

  Vector3 pos = P();
  Vector3 LF = pos + front * 2.25 + left * 1.5;
  Vector3 RF = pos + front * 2.25 - left * 1.5;
  Vector3 LR = pos - front * 2.25 + left * 1.5;
  Vector3 RR = pos - front * 2.25 - left * 1.5;
  dGeomRaySetLength( utilRay(), TANK_RAY_LENGTH );
  dGeomRaySetParams( utilRay(), 0, 1 );
  float force = 0;
  contact_ = false;
  rayCollTest( LF, up, &force );
  if( force > TANK_CONTACT_FORCE_THRESHOLD ) contact_ = true;
  dBodyAddForceAtPos( chassisBody_, up.x * force, up.y * force, up.z * force, LF.x, LF.y, LF.z );
  rayCollTest( RF, up, &force );
  if( force > TANK_CONTACT_FORCE_THRESHOLD ) contact_ = true;
  dBodyAddForceAtPos( chassisBody_, up.x * force, up.y * force, up.z * force, RF.x, RF.y, RF.z );
  rayCollTest( LR, up, &force );
  if( force > TANK_CONTACT_FORCE_THRESHOLD ) contact_ = true;
  dBodyAddForceAtPos( chassisBody_, up.x * force, up.y * force, up.z * force, LR.x, LR.y, LR.z );
  rayCollTest( RR, up, &force );
  if( force > TANK_CONTACT_FORCE_THRESHOLD ) contact_ = true;
  dBodyAddForceAtPos( chassisBody_, up.x * force, up.y * force, up.z * force, RR.x, RR.y, RR.z );

  bool dust = false;
  if( contact_ ) {
    if( terrain_->groundAt( pos.x, pos.z ) == 1 ) {
      dust = true;
    }
  }
  float v2 = L().lengthSquared();
  v2 = clamp( v2/5, 0.0f, 45.0f );
  dust_[0]->setEmitterRate( v2 );
  dust_[1]->setEmitterRate( v2 );
  dReal const * r = R();
  Vector3 ori[3] = { Vector3( r[0], r[4], r[8] ), Vector3( r[1], r[5], r[9] ), Vector3( r[2], r[6], r[10] ) };
  Vector3 rPos = P() + ori[0] * -1.5f + ori[1] * -1.0f;
  dust_[0]->setEmitterPos( rPos, ori );
  dust_[0]->setEmitting( dust );
  rPos = P() - ori[0] * -1.5f + ori[1] * -1.0f;
  dust_[1]->setEmitterPos( rPos, ori );
  dust_[1]->setEmitting( dust );

  if( jumpForce_ > 0 ) {
    dBodyAddForce( chassisBody_, 0, jumpForce_, 0 );
    jumpForce_ = 0;
    contact_ = false;
  }

  if( contact_ ) {
    dBodyAddForce( chassisBody_, front.x * gas_ * TANK_GAS_FORCE, front.y * gas_ * TANK_GAS_FORCE, front.z * gas_ * TANK_GAS_FORCE );
    dBodyAddRelForceAtPos( chassisBody_, 0, 0, turn_ * TANK_TURN_FORCE, left.x, left.y, left.z );
    dBodyAddRelForceAtPos( chassisBody_, 0, 0, -turn_ * TANK_TURN_FORCE, -left.x, -left.y, -left.z );
  }

  Vector3 aVel = *(Vector3 const *)dBodyGetAngularVel( chassisBody_ );
  if( contact_ ) {
    dBodyAddTorque( chassisBody_, -aVel.x * TANK_SHAKE_DAMPENING, -aVel.y * TANK_TURN_DAMPENING, -aVel.z * TANK_SHAKE_DAMPENING );
  }

  Vector3 lVel = *(Vector3 const *)dBodyGetLinearVel( chassisBody_ );
  float v = sqrtf( lVel.lengthSquared() );
  if( contact_ ) {
    dBodyAddForce( chassisBody_, 
        -lVel.x * TANK_V_DAMPENING, -lVel.y * TANK_V_DAMPENING, -lVel.z * TANK_V_DAMPENING );
    if( still_ ) {
      float stay = TANK_STAY_FORCE / (1 + v);
      dBodyAddForce( chassisBody_,
        -lVel.x * stay, 0, -lVel.z * stay );
    }
    lVel -= front * (lVel / front);
    dBodyAddForce( chassisBody_, -lVel.x * TANK_STABILIZING_FORCE, -lVel.y * TANK_STABILIZING_FORCE,
        -lVel.z * TANK_STABILIZING_FORCE );
  }
}

bool MyTank::wantsContact( dContact & contact, MySystemBase * other, dGeomID you, dGeomID him, bool firstTest )
{
#if RENDER_TANK_CONTACTS
  memmove( &lines_[1], &lines_[0], sizeof(lines_[0])*99 );
  memmove( &normals_[1], &normals_[0], sizeof(normals_[0])*99 );
  lines_[0] = (Vector3 &)contact.geom.pos;
  normals_[0] = (Vector3 &)contact.geom.normal;
#endif

  contact.surface.mode |= dContactApprox1 | dContactFDir1 | dContactMu2;
  if( geomIsTreads( you ) ) {
    contact.surface.mode |= dContactSlip2;
    contact.surface.slip2 = TANK_TREAD_SLIP;
    float aTurn = fabsf( turn_ * 30 );
    float appliedPower = fabsf( gas_ ) + aTurn;
    if( appliedPower < 5 ) {
      contact.surface.mu = fabsf(appliedPower)*TANK_TREAD_MU/5 + (5-fabsf(appliedPower))*TANK_TREAD_MU2/5;
    }
    else {
      contact.surface.mu = TANK_TREAD_MU;   //  yay slippery sloppery!
    }
    aTurn = clamp( aTurn, 5.0f, 10.0f ) - 5;
    contact.surface.mu2 = (TANK_TREAD_MU2 * (5 - aTurn) + TANK_TREAD_MU * aTurn) / 5;
  }
  else {
    contact.surface.mu = TANK_MU;
    contact.surface.mu2 = TANK_MU2;
  }
  Matrix3 m3 = *(Matrix3 const *)R();
  Vector3 left, up, front;
  m3.transpose().leftUpFront( &left, &up, &front );
  (Vector3 &)contact.fdir1 = front;
  return true;
}

void MyTank::postStep()
{
  Matrix3 const & m3 = *(Matrix3 const *)R();
  Vector3 left, up, front;
  m3.transpose().leftUpFront( &left, &up, &front );
  Vector3 pos = P();
  Vector3 lPos = pos + left * 1.5f + up * -1.0f;
  Vector3 rPos = pos + left * -1.5f+ up * -1.0f;
  Vector3 lVel, rVel;
  dBodyGetPointVel( chassisBody_, lPos.x, lPos.y, lPos.z, &lVel.x );
  dBodyGetPointVel( chassisBody_, rPos.x, rPos.y, rPos.z, &rVel.x );
  if( contact_ ) {
    float junk1;
    lTreadOffset_ = modff( lTreadOffset_ + (lVel / front) * TIME_STEP * TANK_TREAD_SPEED, &junk1 );
  }
  if( contact_ ) {
    float junk2;
    rTreadOffset_ = modff( rTreadOffset_ + (rVel / front) * TIME_STEP * TANK_TREAD_SPEED, &junk2 );
  }

  switch( weapon_ ) {
  case GrenadeWeapon: {
      Vector3 p = *(Vector3 const *)dBodyGetPosition( gunBody_ );
      Matrix3 m3 = ((Matrix3 const *)dBodyGetRotation( gunBody_ ))->transpose();
      p += m3.front() * (GUN_SIZE/2 + GUN_RADIUS + GRENADE_RADIUS*2 );
      Vector3 v;
      dBodyGetPointVel( gunBody_, p.x, p.y, p.z, &v.x );
      grenadeWeapon_->setAim( m3, p, v );
    }
    break;
  case RocketWeapon: {
      Vector3 p = *(Vector3 const *)dBodyGetPosition( turretBody_ );
      Matrix3 m3 = Matrix3( Vector3::X, elevation_ * -0.2f ) * 
          ((Matrix3 const *)dBodyGetRotation( turretBody_ ))->transpose();
      p += m3.front() * MISSILE_TURRET_FRONT + m3.up() * MISSILE_TURRET_UP +
          m3.left() * MISSILE_TURRET_LEFT;
      Vector3 v;
      dBodyGetPointVel( turretBody_, p.x, p.y, p.z, &v.x );
      rocketWeapon_->setAim( m3, p, v );
    }
    break;
  case MineWeapon: {
      Vector3 p = *(Vector3 const *)dBodyGetPosition( chassisBody_ );
      Matrix3 m3 = ((Matrix3 const *)dBodyGetRotation( chassisBody_ ))->transpose();
      p += m3.up() * MINE_OFFSET_Y;
      Vector3 v;
      dBodyGetPointVel( chassisBody_, p.x, p.y, p.z, &v.x );
      mineWeapon_->setAim( m3, p, v );
    }
    break;
  }
}

void MyTank::recoverStep()
{
  Vector3 actualPos = P();
  Vector3 desiredPos = RayCastAgainstWorld( actualPos + Vector3::Y * 100 ) + Vector3::Y * 2;
  snapshot_.restore( this );
  Vector3 newPos = P();
  newPos = desiredPos - newPos;
  size_t c = bodyCount();
  for( size_t i = 0; i < c; ++i ) {
    dBodyID b = bodyAt( i );
    Vector3 bodyPos = *(Vector3 const *)dBodyGetPosition( b );
    bodyPos += newPos;
    dBodySetPosition( b, bodyPos.x, bodyPos.y, bodyPos.z );
  }
}

void MyTank::jumpImpulse()
{
  Vector3 l = L();
  float f = TANK_JUMP_FORCE - l.y * TANK_ANTIJUMP_VEL;
  Vector3 p = P();
  if( p.y > 200 ) {
    f /= (p.y - 199);
  }
  if( f > 0 ) {
    jumpForce_ = f;
  }
}

void MyTank::setTurn( float turn )
{
  turn_ = turn;
}

void MyTank::setGas( float gas, float target )
{
  gas_ = target > 0 ? gas : -gas;
  still_ = (target == 0);
}

bool MyTank::hasGeom( dGeomID g )
{
  return (std::find( geoms_.begin(), geoms_.end(), g ) != geoms_.end());
}

bool MyTank::geomIsTreads( dGeomID g ) const
{
  return g != geoms_[0] && g != geoms_[1] && g != geoms_[2];
}

void MyTank::setAim( float * aim, float * elevation )
{
  if( *aim < -1.0f ) {
    *aim = -1.0f;
  }
  else if( *aim > 1.0f ) {
    *aim = 1.0f;
  }
  dJointSetHingeParam( turretJoint_, dParamLoStop, -*aim );
  dJointSetHingeParam( turretJoint_, dParamHiStop, -*aim );
  dJointSetHingeParam( turretJoint_, dParamLoStop, -*aim );

  if( *elevation < -0.1f ) {
    *elevation = -0.1f;
  }
  else if( *elevation > 1.0f ) {
    *elevation = 1.0f;
  }
  dJointSetHingeParam( gunJoint_, dParamLoStop, *elevation );
  dJointSetHingeParam( gunJoint_, dParamHiStop, *elevation );
  dJointSetHingeParam( gunJoint_, dParamLoStop, *elevation );

  elevation_ = *elevation;
}


I_Weapon * MyTank::weapon()
{
  return weapon( weapon_ );
}

I_Weapon * MyTank::weapon( Weapon w )
{
  if( w == GrenadeWeapon ) {
    return grenadeWeapon_;
  }
  if( w == RocketWeapon ) {
    return rocketWeapon_;
  }
  if( w == MineWeapon ) {
    return mineWeapon_;
  }
  return 0;
}

Weapon MyTank::nextWeapon( int dir )
{
  assert( dir == -1 || dir == 1 );
  //fixme: this needs to separate when I support more than two weapons
  if( dir > 0 ) {
    switch( weapon_ ) {
    case GrenadeWeapon:
      weapon_ = RocketWeapon;
      break;
    default:
    case RocketWeapon:
      weapon_ = MineWeapon;
      break;
    case MineWeapon:
      weapon_ = GrenadeWeapon;
      break;
    }
  }
  else {
    switch( weapon_ ) {
    case GrenadeWeapon:
      weapon_ = MineWeapon;
      break;
    default:
    case RocketWeapon:
      weapon_ = GrenadeWeapon;
      break;
    case MineWeapon:
      weapon_ = RocketWeapon;
      break;
    }
  }
  return weapon_;
}

void MyTank::setWeapon( Weapon w )
{
  weapon_ = w;
}

