
#include "mywheeledcar.h"
#include "soundmanager.h"
#include "mytweakables.h"
#include "misc.h"
#include "myparticles.h"
#include "terrain.h"
#include "mymine.h"
#include "mymachinegun.h"
#include "mysystemsnapshot.h"

#include <assert.h>




MyWheeledCar::~MyWheeledCar() {
  gSoundManager->stopPlaying( engineSoundInstance_ );
  gSoundManager->unloadSound( engineSound_ );
  gSoundManager->unloadSound( crashSound_ );
  delete dust_[0];
  delete dust_[1];
  delete mineWeapon_;
}

MyWheeledCar::MyWheeledCar( Params const & params )
{
  exhaust_ = MakeParticleSystem( "exhaust", worldSpace() );

  memset( soundContact_, 0, sizeof( soundContact_ ) );
  memset( soundContactCnt_, 0, sizeof( soundContactCnt_ ) );
  memset( soundContactVol_, 0, sizeof( soundContactVol_ ) );

  dMass m;
  space_ = dSimpleSpaceCreate( worldSpace() );

  chassisBody_ = dBodyCreate( world() );
  dBodySetAutoDisableFlag( chassisBody_, 0 );
  dBodySetGravityMode( chassisBody_, 1 );
  dMassSetBoxTotal( &m, CHASSIS_MASS, 2.0f, 1.06f, 4.0f );
  dBodySetMass( chassisBody_, &m );
  addBody( chassisBody_, params.carShader, params.carMesh, Vector3( 0, SINK_AMT, 0 ) );

  //  Adjust the chassis up by the sink amount (I sink the 
  //  body to get a lower center of gravity).
  dGeomID tempGeom = dCreateBox( 0, 2.0f, 1.06f, 4.0f );
  dGeomSetPosition( tempGeom, 0, SINK_AMT, 0 );
  dGeomID tempGeom2 = dCreateGeomTransform( space_ );
  dGeomTransformSetCleanup( tempGeom2, 1 );
  dGeomTransformSetGeom( tempGeom2, tempGeom );
  dGeomTransformSetInfo( tempGeom2, 1 );
  addGeom( tempGeom2, chassisBody_ );

  tempGeom = dCreateBox( 0, 1.7f, 0.6f, 0.1f );
  dGeomSetPosition( tempGeom, 0, 1.3f, 0.8f );
  tempGeom2 = dCreateGeomTransform( space_ );
  dGeomTransformSetCleanup( tempGeom2, 1 );
  dGeomTransformSetGeom( tempGeom2, tempGeom );
  dGeomTransformSetInfo( tempGeom2, 1 );
  addGeom( tempGeom2, chassisBody_ );

  dBodySetPosition( chassisBody_, 0, -SINK_AMT, 0 );

  //wheel  X  Z
  //  0    1  1  LF
  //  1   -1  1  RF
  //  2    1 -1  LR
  //  3   -1 -1  RR
  for( int i = 0; i < 4; ++i ) {

    wheelBody_[ i ] = dBodyCreate( world() );

    dBodySetAutoDisableFlag( wheelBody_[i], 0 );
    dBodySetGravityMode( wheelBody_[i], 1 );
    dBodySetFiniteRotationMode( wheelBody_[i], 1 );
    dMassSetCylinderTotal( &m, WHEEL_MASS, 1, 0.45f, 0.3f );
    dBodySetMass( wheelBody_[i], &m );
    addBody( wheelBody_[ i ], params.wheelShader, params.wheelMesh );

    tempGeom = dCreateSphere( space_, WHEEL_RADIUS );
    addGeom( tempGeom, wheelBody_[i] );

    dBodySetPosition( wheelBody_[i], 0.9f * ((i&1) ? -1 : 1), -WHEEL_LOW, 1.2f * ((i&2) ? -1 : 1) );
    if( i & 1 ) {
      dMatrix3 R = {
        -1, 0, 0, 0,
        0, 1, 0, 0, 
        0, 0, -1, 0,
      };
      dBodySetRotation( wheelBody_[i], R );
    }

    hinges_[i] = dJointCreateHinge2( world(), 0 );
    addJoint( hinges_[i], chassisBody_, wheelBody_[i] );
    dJointSetHinge2Anchor( hinges_[i], 0.9f * ((i&1) ? -1 : 1), -WHEEL_LOW, 1.2f * ((i&2) ? -1 : 1) );
    dJointSetHinge2Axis1( hinges_[i], 0, 1, 0 );
    dJointSetHinge2Axis2( hinges_[i], ((i&1) ? 1.0f : -1.0f), 0, 0 );
    dJointSetHinge2Param( hinges_[i], dParamSuspensionERP, (!(i&2) ? FRONT_SUSPENSION_ERP : REAR_SUSPENSION_ERP) );
    dJointSetHinge2Param( hinges_[i], dParamSuspensionCFM, (!(i&2) ? FRONT_SUSPENSION_CFM : REAR_SUSPENSION_CFM) );
    dJointSetHinge2Param( hinges_[i], dParamLoStop, 0 );
    dJointSetHinge2Param( hinges_[i], dParamHiStop, 0 );
    dJointSetHinge2Param( hinges_[i], dParamLoStop, 0 );
    dJointSetHinge2Param( hinges_[i], dParamHiStop, 0 );
    dJointSetHinge2Param( hinges_[i], dParamStopERP, (!(i&2) ? FRONT_TURN_ERP : REAR_TURN_ERP) );
    dJointSetHinge2Param( hinges_[i], dParamStopCFM, (!(i&2) ? FRONT_TURN_CFM : REAR_TURN_CFM) );
  }

  machinegunShader_ = gShaderFactory->newShader( "machinegunShader" );
  machinegunMesh_ = ASEFromFile( asePath( "machinegun" ) );
  machinegunBody_ = dBodyCreate( world() );
  dMassSetCappedCylinderTotal( &m, 0.3f, 3, 0.2f, 0.6f );
  dBodySetMass( machinegunBody_, &m );
  dBodySetPosition( machinegunBody_, -0.3f, 1.4f, -1.2f );
  addBody( machinegunBody_, machinegunShader_, machinegunMesh_ );
  tempGeom = dCreateCCylinder( worldSpace(), 0.2f, 0.6f );
  addGeom( tempGeom, machinegunBody_ );

  machinegunJoint_ = dJointCreateUniversal( world(), 0 );
  addJoint( machinegunJoint_, chassisBody_, machinegunBody_ );
  dJointSetUniversalAnchor( machinegunJoint_, -0.3f, 1.4f, -1.2f );
  dJointSetUniversalAxis1( machinegunJoint_, 0, 1, 0 );
  dJointSetUniversalAxis2( machinegunJoint_, 1, 0, 0 );
  dJointSetUniversalParam( machinegunJoint_, dParamStopERP, 0.7f );
  dJointSetUniversalParam( machinegunJoint_, dParamStopCFM, 0.01f );
  dJointSetUniversalParam( machinegunJoint_, dParamStopERP|0x100, 0.7f );
  dJointSetUniversalParam( machinegunJoint_, dParamStopCFM|0x100, 0.01f );

  aim_ = 0;
  elevation_ = 0;
  dJointSetUniversalParam( machinegunJoint_, dParamLoStop, -aim_ );
  dJointSetUniversalParam( machinegunJoint_, dParamHiStop, -aim_ );
  dJointSetUniversalParam( machinegunJoint_, dParamLoStop, -aim_ );
  dJointSetUniversalParam( machinegunJoint_, dParamLoStop|0x100, elevation_ );
  dJointSetUniversalParam( machinegunJoint_, dParamHiStop|0x100, elevation_ );
  dJointSetUniversalParam( machinegunJoint_, dParamLoStop|0x100, elevation_ );
  
  swayBars_ = true;

  engineSound_ = gSoundManager->loadSound( wavPath( "engine" ) );
  crashSound_ = gSoundManager->loadSound( wavPath( "crash" ) );
  
  PlayParameters pp;
  pp.autoStop = false;
  pp.looping = true;
  pp.volL = 0.25f;
  pp.volR = 0.25f;
  pp.pitch = 0.5f;
  engineSoundInstance_ = gSoundManager->playInstance( engineSound_, &pp );
  deltaTime_ = 0;

  // dust trail
  terrain_ = params.terrain;
  contact_ = false;
  dust_[0] = MakeParticleSystem( "dustcloud", worldSpace() );
  dust_[1] = MakeParticleSystem( "dustcloud", worldSpace() );
  dust_[0]->setEmitting( false );
  dust_[1]->setEmitting( false );

  machinegunWeapon_ = NewMachinegunWeapon();
  mineWeapon_ = NewMineWeapon();
  weapon_ = MachinegunWeapon;

  snapshot_.capture( *this );
}

void MyWheeledCar::preStep()
{
  ++deltaTime_;
  //  Calculate finite rotation axis
  for( int ix = 0; ix < 4; ++ix ) {
    dVector3 axis;
    dJointGetHinge2Axis2( hinges_[ix], axis );
    dBodySetFiniteRotationAxis( wheelBody_[ix], axis[0], axis[1], axis[2] );
  }

  //  Speedometer
  dReal const * cv = dBodyGetLinearVel( chassisBody_ );
  float v2 = cv[0]*cv[0] + cv[1]*cv[1] + cv[2]*cv[2];
  float v = sqrtf( v2 ) / 1000 * 3600;
  debugVar( 1, 1, "vel = %-4.1f km/h", v );

  dReal const * r = R();
  Vector3 ori[3] = { Vector3( r[0], r[4], r[8] ), Vector3( r[1], r[5], r[9] ), Vector3( r[2], r[6], r[10] ) };
  Vector3 pos = P();
  Vector3 rPos = pos + ori[2] * -2.2f;
  exhaust_->setEmitterPos( rPos, ori );

  //  aerodynamics
  dBodyAddForce( chassisBody_, 0, AERO_LOAD * v2, 0 );

  //  engine sound
  float av = 0;
  for( int i = 0; i < 4; ++i ) {
    dReal const * aV = dBodyGetAngularVel( wheelBody_[i] );
    float aa = sqrtf( aV[0]*aV[0] + aV[1]*aV[1] + aV[2]*aV[2] );
    //  use the average of the wheels' angular velocities
    av += aa;
  }
  //fixme:  it's a little too often to do this 100 times a second...
  gSoundManager->setPitch( engineSoundInstance_, 0.75f + av * 0.005f );
  exhaust_->setEmitterAlpha( 0.20f + av * 0.002f );

  //  anti-sway bar forces (this is somewhat crude)
  if( swayBars_ ) {
    Vector3 bodyPoint;
    Vector3 hingePoint;
    Vector3 axis;
    float displacement;
    for( int ix = 0; ix < 4; ++ix ) {
      dJointGetHinge2Anchor2( hinges_[ix], &bodyPoint.x );
      dJointGetHinge2Anchor( hinges_[ix], &hingePoint.x );
      dJointGetHinge2Axis1( hinges_[ix], &axis.x );
      displacement = (hingePoint - bodyPoint) / axis;
      float amt = displacement * SWAY_FORCE;
      if( displacement > 0 ) {
        if( amt > 15 ) {
          amt = 15;
        }
        dBodyAddForce( wheelBody_[ix], -axis.x * amt, -axis.y * amt, -axis.z * amt );
        dReal const * wp = dBodyGetPosition( wheelBody_[ix] );
        dBodyAddForceAtPos( chassisBody_, axis.x*amt, axis.y*amt, axis.z*amt, wp[0], wp[1], wp[2] );
        dBodyAddForce( wheelBody_[ix^1], axis.x * amt, axis.y * amt, axis.z * amt );
        wp = dBodyGetPosition( wheelBody_[ix] );
        dBodyAddForceAtPos( chassisBody_, -axis.x*amt, -axis.y*amt, -axis.z*amt, wp[0], wp[1], wp[2] );
      }
    }
  }

  bool dust = false;
  if( contact_ ) {
    if( terrain_->groundAt( pos.x, pos.z ) == 1 ) {
      dust = true;
    }
  }
  v2 = clamp( v2/5, 0.0f, 45.0f );
  dust_[0]->setEmitterRate( v2 );
  dust_[1]->setEmitterRate( v2 );
  rPos = P() + ori[0] * -1.2f + ori[1] * (SINK_AMT-1.0f);
  dust_[0]->setEmitterPos( rPos, ori );
  dust_[0]->setEmitting( dust );
  rPos = P() - ori[0] * -1.2f + ori[1] * (SINK_AMT-1.0f);
  dust_[1]->setEmitterPos( rPos, ori );
  dust_[1]->setEmitting( dust );

  contact_ = false;
}

void MyWheeledCar::postStep()
{
  switch( weapon_ ) {
  case MachinegunWeapon: {
      Vector3 p = *(Vector3 const *)dBodyGetPosition( machinegunBody_ );
      Matrix3 m3 = ((Matrix3 const *)dBodyGetRotation( machinegunBody_ ))->transpose();
      Vector3 v;
      dBodyGetPointVel( machinegunBody_, p.x, p.y, p.z, &v.x );
      machinegunWeapon_->setAim( m3, p, v );
    }
    break;
  case MineWeapon: {
      Vector3 p = *(Vector3 const *)dBodyGetPosition( chassisBody_ );
      Matrix3 m3 = ((Matrix3 const *)dBodyGetRotation( chassisBody_ ))->transpose();
      p += m3.up() * (MINE_OFFSET_Y - SINK_AMT);
      Vector3 v;
      dBodyGetPointVel( chassisBody_, p.x, p.y, p.z, &v.x );
      mineWeapon_->setAim( m3, p, v );
    }
    break;
  }
}

bool MyWheeledCar::wantsContact( dContact & contact, MySystemBase * other, dGeomID you, dGeomID him, bool firstTest )
{
  // car
  float inv;
  dBodyID bid = dGeomGetBody( you );
  bool makeNoise = true;
  float noiseVolume = 0;
  dReal const * vel = dBodyGetLinearVel( bid );
  float colVel = ((Vector3 const &)*vel) / (Vector3 &)contact.geom.normal;
  if( bodyIsWheel( bid, &inv ) ) {
    Vector3 front;
    Vector3 up;
    //  Compute fDir1 if a wheel
    dReal const * R = dBodyGetRotation( bid );
    front = Vector3( R[2] * inv, R[6] * inv, R[10] * inv );
    up = Vector3( R[1] * inv, R[5] * inv, R[9] * inv );
    //  Set Slip2
    contact.surface.mode |= dContactSlip2 | dContactFDir1;
    float v = sqrtf( vel[0]*vel[0] + vel[1]*vel[1] + vel[2]*vel[2] );
    contact.surface.slip2 = TIRE_SLIP * v;
    //  Turn on Approx1
    contact.surface.mode |= dContactApprox1;
    //  re-tweak mu and mu2
    contact.surface.mu = contact.surface.mu2 = TIRE_MU;
    //  The theory is that it doesn't matter if "front" points "up" 
    //  because we want fdir2 to be orthogonal to "front" and contact 
    //  normal. Of course, if "front" points in the direction of the 
    //  contact, this can be a problem. In that case, choose "up".
    if( fabsf( dDot( contact.geom.normal, &front.x, 3 ) ) > 0.5f ) {
      ((Vector3 &)contact.fdir1) = up;
    }
    else {
      ((Vector3 &)contact.fdir1) = front;
    }
    makeNoise = (colVel > 2.0f);
    noiseVolume = colVel/6 - 0.2f;
    if( makeNoise ) {
      info( "wheel noise %.1f\n", noiseVolume );
    }
    contact_ = true;
  }
  else {
    noiseVolume = sqrtf( colVel * 0.4f );
    makeNoise = (noiseVolume >= 0.1f);
    //if( makeNoise ) {
      //info( "chassis noise %.1f %s\n", noiseVolume, makeNoise ? "ON" : "off" );
    //}
  }
  if( noiseVolume > 1 ) {
    noiseVolume = 1;
  }
  //  chassis
  if( makeNoise ) {
    for( int i = 0; i < MAX_SOUND_CONTACTS; ++i ) {
      if( !soundContact_[i] ) {
        info( "noiseVolume %.1f colVel %.1f\n", noiseVolume, colVel );
        gSoundManager->playInstance( crashSound_, noiseVolume * 0.5f + 0.2f );
        soundContact_[i] = him;
        soundContactCnt_[i] = 0;
        soundContactVol_[i] = noiseVolume;
        goto found;
      }
      if( soundContact_[i] == him ) {
        if( noiseVolume > soundContactVol_[i]+0.1f ) {
          info( "increased volume for 0x%lx to %.1f\n", him, noiseVolume );
          gSoundManager->playInstance( crashSound_, noiseVolume * 0.5f );
          soundContact_[i] = him;
          soundContactCnt_[i] = 0;
          soundContactVol_[i] = noiseVolume;
          goto found;
        }
        //  already have this guy -- don't do him again
        //info( "removed duplicate contact 0x%lx\n", him );
        goto found;
      }
    }
    info( "wanted noise, but no slot\n" );
  found:
    ;
  }
  return true;
}

void MyWheeledCar::render()
{
  MySystem::render();
  for( int i = 0; i < MAX_SOUND_CONTACTS; ++i ) {
    if( soundContact_[i] ) {
      soundContactCnt_[i] += deltaTime_;
      //info( "0x%lx adds %d to %d\n", soundContact_[i], deltaTime_, soundContactCnt_[i] );
      if( soundContactCnt_[i] > 50 ) {
        //  Remove contacts we haven't seen for a while.
        //  That way, we can collide with them again and
        //  make another sound.
        memmove( &soundContact_[i], &soundContact_[i+1], 
          sizeof( soundContact_[i] ) * (MAX_SOUND_CONTACTS-i-1) );
        memmove( &soundContactCnt_[i], &soundContactCnt_[i+1], 
          sizeof( soundContactCnt_[i] ) * (MAX_SOUND_CONTACTS-i-1) );
        soundContact_[MAX_SOUND_CONTACTS-1] = 0;
        soundContactCnt_[MAX_SOUND_CONTACTS-1] = 0;
        --i;
      }
    }
  }
  deltaTime_ = 0;
}

bool MyWheeledCar::bodyIsWheel( dBodyID bid, float * inv )
{
  if( bid == wheelBody_[0] || bid == wheelBody_[2] ) {
    *inv = -1;
    return true;
  }
  if( bid == wheelBody_[1] || bid == wheelBody_[3] ) {
    *inv = 1;
    return true;
  }
  return false;
}

void MyWheeledCar::doDampening()
{
  DampenBody( chassisBody_, -0.03f, -0.04f );
  for( int ix = 0; ix < 4; ++ix ) {
    DampenBody( wheelBody_[ix], -0.03f, -0.02f );
  }
}

void MyWheeledCar::recoverStep()
{
  Vector3 actualPos = P();
  Vector3 desiredPos = RayCastAgainstWorld( actualPos + Vector3::Y * 100 ) + Vector3::Y * 2;
  snapshot_.restore( this );
  Vector3 newPos = P();
  newPos = desiredPos - newPos;
  size_t c = bodyCount();
  for( size_t i = 0; i < c; ++i ) {
    dBodyID b = bodyAt( i );
    Vector3 bodyPos = *(Vector3 const *)dBodyGetPosition( b );
    bodyPos += newPos;
    dBodySetPosition( b, bodyPos.x, bodyPos.y, bodyPos.z );
  }
}

void MyWheeledCar::jumpImpulse()
{
  float scale = 1;
  if( P().y > 30 ) {
    scale = 30 / P().y;
  }
  dBodyAddForce( chassisBody_, 0, 5000 * scale, 0 );
  dBodyAddForce( wheelBody_[0], 0, 1000 * scale, 0 );
  dBodyAddForce( wheelBody_[1], 0, 1000 * scale, 0 );
  dBodyAddForce( wheelBody_[2], 0, 1000 * scale, 0 );
  dBodyAddForce( wheelBody_[3], 0, 1000 * scale, 0 );
}

void MyWheeledCar::setTurn( float turn )
{
  float leftAdjust = (turn > 0 ? turn : turn * TURN_INNER_SCALE);
  float rightAdjust = (turn < 0 ? turn : turn * TURN_INNER_SCALE);
  dJointSetHinge2Param( hinges_[0], dParamLoStop, leftAdjust );
  dJointSetHinge2Param( hinges_[0], dParamHiStop, leftAdjust );
  dJointSetHinge2Param( hinges_[0], dParamLoStop, leftAdjust );
  dJointSetHinge2Param( hinges_[0], dParamHiStop, leftAdjust );
  dJointSetHinge2Param( hinges_[1], dParamLoStop, rightAdjust );
  dJointSetHinge2Param( hinges_[1], dParamHiStop, rightAdjust );
  dJointSetHinge2Param( hinges_[1], dParamLoStop, rightAdjust );
  dJointSetHinge2Param( hinges_[1], dParamHiStop, rightAdjust );
}

void MyWheeledCar::setGas( float gas, float target )
{
  dJointSetHinge2Param( hinges_[0], dParamFMax2, gas * FRONTPCT );
  dJointSetHinge2Param( hinges_[1], dParamFMax2, gas * FRONTPCT );
  dJointSetHinge2Param( hinges_[2], dParamFMax2, gas * REARPCT );
  dJointSetHinge2Param( hinges_[3], dParamFMax2, gas * REARPCT );
  dJointSetHinge2Param( hinges_[0], dParamVel2, target );
  dJointSetHinge2Param( hinges_[1], dParamVel2, -target );
  dJointSetHinge2Param( hinges_[2], dParamVel2, target );
  dJointSetHinge2Param( hinges_[3], dParamVel2, -target );
}

void MyWheeledCar::setSwayBars( bool s )
{
  swayBars_ = s;
}

bool MyWheeledCar::swayBars() const
{
  return swayBars_;
}

void MyWheeledCar::setAim( float * aim, float * elevation )
{
  // negative goes to the right
  if( *aim < -1.5f ) {
    *aim = -1.5f;
  }
  else if( *aim > 1.5f ) {
    *aim = 1.5f;
  }
  aim_ = *aim;
  if( aim_ > -0.3f && aim_ < 0.62f ) {
    //  don't shoot the windshield!
    if( *elevation < -0.088f ) {
      *elevation = -0.088f;
    }
    else if( *elevation > 0.8f ) {
      *elevation = 0.8f;
    }
  }
  else {
    if( *elevation < -0.3f ) {
      *elevation = -0.3f;
    }
    else if( *elevation > 0.8f ) {
      *elevation = 0.8f;
    }
  }
  elevation_ = *elevation;

  dJointSetUniversalParam( machinegunJoint_, dParamLoStop, -aim_ );
  dJointSetUniversalParam( machinegunJoint_, dParamHiStop, -aim_ );
  dJointSetUniversalParam( machinegunJoint_, dParamLoStop, -aim_ );

  dJointSetUniversalParam( machinegunJoint_, dParamLoStop|0x100, elevation_ );
  dJointSetUniversalParam( machinegunJoint_, dParamHiStop|0x100, elevation_ );
  dJointSetUniversalParam( machinegunJoint_, dParamLoStop|0x100, elevation_ );
}

I_Weapon * MyWheeledCar::weapon()
{
  return weapon( weapon_ );
}

I_Weapon * MyWheeledCar::weapon( Weapon w )
{
  if( w == MineWeapon ) {
    return mineWeapon_;
  }
  if( w == MachinegunWeapon ) {
    return machinegunWeapon_;
  }
  return 0;
}

Weapon MyWheeledCar::nextWeapon( int dir )
{
  assert( dir == 1 || dir == -1 );
  if( dir > 0 ) {
    switch( weapon_ ) {
    case MineWeapon:
      weapon_ = MachinegunWeapon;
      break;
    default:
    case MachinegunWeapon:
      weapon_ = MineWeapon;
      break;
    }
  }
  else {
    switch( weapon_ ) {
    case MineWeapon:
      weapon_ = MachinegunWeapon;
      break;
    default:
    case MachinegunWeapon:
      weapon_ = MineWeapon;
      break;
    }
  }
  return weapon_;
}

void MyWheeledCar::setWeapon( Weapon w )
{
  weapon_ = w;
}

