
#if !defined( mywheeledcar_h )
#define mywheeledcar_h

#include "myvehicle.h"
#include "soundmanager.h"
#include "myparticles.h"
#include "mysystemsnapshot.h"

class I_Terrain;

class MyWheeledCar : public MyVehicle {
  public:
    struct Params {
      I_GL_ASE * carMesh;
      GLShader * carShader;
      I_GL_ASE * wheelMesh;
      GLShader * wheelShader;
      I_Terrain * terrain;
    };

    ~MyWheeledCar();
    MyWheeledCar( Params const & params );

    void preStep();
    void postStep();
    bool wantsContact( dContact & contact, MySystemBase * other, dGeomID you, dGeomID him, bool firstTest );
    void render();
    bool bodyIsWheel( dBodyID bid, float * inv );

    void doDampening();
    void recoverStep();
    void jumpImpulse();
    void setTurn( float turn );
    void setGas( float gas, float target );
    void setSwayBars( bool s );
    bool swayBars() const;
    void setAim( float * aim, float * elevation );

    I_Weapon * weapon();
    I_Weapon * weapon( Weapon w );
    Weapon nextWeapon( int dir );
    void setWeapon( Weapon w );

private:

    SoundHandle engineSound_;
    SoundHandle crashSound_;
    PlayingHandle engineSoundInstance_;

    enum {
      MAX_SOUND_CONTACTS = 8
    };
    dGeomID soundContact_[MAX_SOUND_CONTACTS];
    int soundContactCnt_[MAX_SOUND_CONTACTS];
    float soundContactVol_[MAX_SOUND_CONTACTS];
    int deltaTime_;

    ParticleOp exhaustOp_;
    MyParticles * exhaust_;
    bool contact_;
    I_Terrain * terrain_;
    MyParticles * dust_[2];

    float aim_;
    float elevation_;
    Weapon weapon_;
    I_Weapon * mineWeapon_;
    I_Weapon * machinegunWeapon_;
    GLShader * machinegunShader_;
    I_GL_ASE * machinegunMesh_;
    dJointID machinegunJoint_;
    dBodyID machinegunBody_;

    MySystemSnapshot snapshot_;

private:
    dBodyID wheelBody_[4];
    dJointID hinges_[4];
    bool swayBars_;
};

#endif  //  mywheeledcar_h

