
#if !defined( owavefile_h )
#define owavefile_h

#include "wavformat.h"

enum WaveOpenMode
{
  WaveOpenNull,
  WaveOpenRead,
  WaveOpenCreate,
  WaveOpenUpdate
};

//  This interface is deliberately WAVE file centric.
//  With some adjustment (abstracting format chunk, etc) it could be 
//  made to be sound file format agostic, but that's not worth it 
//  right now.

class OWaveFile
{
public:

  virtual void setPath( char const * path ) = 0;
  virtual void setOpenMode( WaveOpenMode mode ) = 0;
  virtual bool open() = 0;
  virtual void destroy() = 0;

  //  read/update functions
  virtual unsigned int countChunks() = 0;
  virtual unsigned int chunkType( unsigned int ix ) = 0;
  virtual unsigned int chunkSize( unsigned int ix ) = 0;
  virtual bool readChunkData( unsigned int ix, void * buffer, unsigned int offset, unsigned int * ioSize ) = 0;
  virtual bool fmtHeader( WAVFmtHeader * outHdr, void * extraData, unsigned int * extraDataSize ) = 0;
  virtual unsigned int countFrames() = 0;
  virtual unsigned int readFrames( float ** channels, unsigned int frameOffset, unsigned int frameCount ) = 0;

  //  update functions
  virtual bool writeChunkData( unsigned int ix, void const * buffer, unsigned int offset, unsigned int * ioSize ) = 0;

  //  write only functions
  virtual bool newChunk( unsigned int type ) = 0;
  virtual bool writeNewData( void const * data, unsigned int size ) = 0;
  virtual bool endChunk() = 0;
  virtual bool setFmtHeader( WAVFmtHeader const & hdr, void const * extraData, unsigned int extraSize ) = 0;
  virtual bool beginData() = 0;
  virtual bool writeFrames( float const * const * channels, unsigned int frameCount ) = 0;
  virtual bool endData() = 0;

  //  convenience
  bool addChunk( unsigned int type, void const * data, unsigned int size )
            {
              bool b = newChunk( type );
              b = b && writeNewData( data, size );
              b = b && endChunk();
              return b;
            }
protected:
  ~OWaveFile() { }
};

OWaveFile * NewWaveFile();


#endif  //  owavefile_h
