
#if !defined( owaveplayer_h )
#define owaveplayer_h

class OWavePlayer;
class MWavePlayer;

class MWaveBuffer {
public:
  virtual unsigned int size( ) = 0; //  in frames
  virtual unsigned int setData( float const * const * buffers, unsigned int frames ) = 0; //  in frames
};

class MWaveDataNotify {
public:
  //  called from within start()
  virtual void playbackStarted( MWavePlayer * wp ) = 0;

  //  called from another thread (!)
  virtual void bufferCallback( MWavePlayer * wp, MWaveBuffer * buf ) = 0;

  //  called from another thread (!)
  virtual void playbackStopped( MWavePlayer * wp ) = 0;

protected:
  ~MWaveDataNotify() { }
};

class MWavePlayer {
public:
  //  Start() will call you back for some data, and then start calling 
  //  you back at periodic intervals for more data.
  virtual bool start( MWaveDataNotify * notify ) = 0;
  //  Stop does NOT synchronously wait until everything's done!
  //  When playbackStopped() is called, you know it's done.
  virtual void stop() = 0;

protected:
  virtual ~MWavePlayer() { }
};

class OWavePlayer {
public:
  virtual void destroy() = 0;

  virtual void setFormat( unsigned int channelCount, float framesPerSecond ) = 0;
  virtual void setBuffers( unsigned int count, unsigned int size, bool tryFp ) = 0;
  virtual void setVolume( float mainVolume ) = 0;
  virtual bool open() = 0;
  virtual void close() = 0;

  virtual MWavePlayer * wavePlayer() = 0;

protected:
  ~OWavePlayer() { }
};

OWavePlayer * NewWavePlayer();

#endif  //  owaveplayer_h
