
#include "..\image.h"

#include <stdio.h>
#include <stdarg.h>


bool quiet = false;

char * pattern = "%s.tga";

int main( int argc, char * argv[] ) {
  if( argc > 1 && (!stricmp( argv[1], "-q" ) || !stricmp( argv[1], "/q" ) ) ) {
    quiet = true;
    ++argv;
    --argc;
  }
  if( argc > 1 && (!stricmp( argv[1], "-p" ) || !stricmp( argv[1], "/p" ) ) ) {
    if( argc < 3 ) {
      fprintf( stderr, "error: Pattern option needs an argument.\n" );
      exit( 1 );
    }
    pattern = argv[2];
    if( strlen( pattern ) > 50 ) {
      fprintf( stderr, "error: pattern can at most be 50 characters long.\n" );
      exit( 1 );
    }
    argc -= 2;
    argv += 2;
  }
  if( argc < 2 || argv[1][0] == '-' || argv[1][0] == '/' ) {
    fprintf( stderr, "usage: resizetga [-q] [-p pattern] file.tga [...]\n" );
    fprintf( stderr, "Will halve the size of the tga until it's no bigger than 256x256.\n" );
    exit( 1 );
  }
  int ret = 0;
  for( int j = 1; j < argc; ++j ) {
    if( strlen( argv[j] ) > 200 ) {
      fprintf( stderr, "error: input path '%s' is too long!\n", argv[j] );
      ++ret;
      continue;
    }
    Image i;
    if( !loadTgaImage( argv[j], &i ) ) {
      fprintf( stderr, "%s: cannot open or not a TGA file.\n", argv[j] );
      ++ret;
    }
    else {
      int nuWidth = i.width;
      int nuHeight = i.height;
      Image o;
      int scale = 1;
      int shift = 0;
      while( nuWidth > 256 || nuHeight > 256 ) {
        nuWidth >>= 1;
        nuHeight >>= 1;
        scale <<= 1;
        shift+=2;
      }
      o.allocate( i.bits, nuWidth, nuHeight );
      if( !quiet ) {
        fprintf( stderr, "%s: was %dx%d, now %dx%d.\n", argv[j], i.width, i.height, nuWidth, nuHeight );
      }
      int components = i.bits / 8;
      unsigned char * iptr = (unsigned char *)i.base;
      unsigned char * optr = (unsigned char *)o.base;
      for( int v = 0; v < nuHeight; ++v ) {
        for( int u = 0; u < nuWidth; ++u ) {
          int rgba[4] = { 0, 0, 0, 0 };
          for( int y = 0; y < scale; ++y ) {
            for( int x = 0; x < scale; ++x ) {
              for( int c = 0; c < components; ++c ) {
                rgba[c] += iptr[(u*scale+x)*components+c + (v*scale+y)*i.pitch];
              }
            }
          }
          for( int c = 0; c < components; ++c ) {
            optr[u*components+c + v*o.pitch] = (rgba[c]+scale/2)>>shift;
          }
        }
      }
      char outputName[256];
      char name[256];
      char bakName[256];
      char * inputName = argv[j];
      char * nameBegin = strrchr( inputName, '/' );
      if( nameBegin ) ++nameBegin; else nameBegin = inputName;
      char * nameEnd = strrchr( nameBegin, '.' );
      if( !nameEnd ) nameEnd = nameBegin+strlen( nameBegin );
      strncpy( name, nameBegin, nameEnd-nameBegin );
      name[nameEnd-nameBegin] = 0;
      //  we guard against overruns in the initial path check.
      sprintf( outputName, pattern, name );
      sprintf( bakName, "%s.bak", outputName );
      unlink( bakName );
      rename( outputName, bakName );
      if( !saveTgaImage( outputName, &o ) ) {
        fprintf( stderr, "%s: cannot save resized image.\n", outputName );
        ++ret;
        continue;
      }
    }
  }
  return ret;
}

