
#if !defined( scene_h )
#define scene_h

#include "mymath.h"

typedef struct _SceneRef_t * SceneRef;
class I_Scene;

//  You implement this interface for anything that goes into the scene
class I_SceneNode {
  protected:
    virtual ~I_SceneNode() {}
  public:
    //  properties are properties at identity posOri (i e, only calculated once)
    virtual void properties( Bounds * outBounds, float * outRadius ) = 0;
};

//  You implement this interface to receive nodes when iterating
class I_SceneVisitor {
  protected:
    virtual ~I_SceneVisitor() {}
  public:
    virtual void onSceneVisit( I_Scene * scene, SceneRef ref, I_SceneNode * node, Matrix4 const & posOri ) = 0;
};

//  I implement this interface (returned from MakeScene()) that allows you 
//  to add and manage I_SceneNode interfaces, and visit them according to 
//  location.
class I_Scene {
  protected:
    virtual ~I_Scene() {}
  public:
    virtual void dispose() = 0;

    virtual SceneRef addNode( I_SceneNode * node, Matrix4 const & posOri ) = 0;
    virtual void removeNode( SceneRef ref ) = 0;
    virtual void moveNode( SceneRef ref, Matrix4 const & posOri ) = 0;
    virtual void getNodePos( SceneRef ref, Matrix4 * outPosOri ) = 0;

    virtual void visitNodes( I_SceneVisitor * visitor ) = 0;
    virtual void visitNodesAABB( Vector3 const & low, Vector3 const & high, I_SceneVisitor * visitor ) = 0;
    virtual void visitNodesPlanes( Vector4 const * planes, uint numPlanes, I_SceneVisitor * visitor ) = 0;
};

I_Scene * MakeScene();

#endif  //  scene_h
