
#if !defined( terrain_h )
#define terrain_h

#include "mymath.h"
#include "ode/ode.h"

struct Image;

class I_TerrainChunk {
  public:
    virtual Vector3 center() = 0;
    virtual Vector3 origin() = 0;
    virtual Vector3 lower() = 0;
    virtual Vector3 upper() = 0;
    //  Mesurements of chunks are in a right-handed world (!)
    virtual void measure( Vector3 & oCenter, Vector3 & oOrigin, Vector3 & oLower, Vector3 & oUpper ) = 0;
};

class I_TerrainODE {
  public:
    virtual dTriMeshDataID triMeshDataID() = 0;
    virtual dGeomID makeGeom( dSpaceID w ) = 0;
};

class I_Terrain {
  public:
    virtual void dispose() = 0;

    //  Measurements are in a left-handed world(!)
    virtual void measure( float & west, float & south, float & width, float & height ) = 0;
    //  heightAt() isn't terribly efficient, but it's OK for things like 
    //  ray casts for character position or IK.
    //  Warning! heightAt() is LEFT-HANDED! (it's in bottom-to-top image space, 
    //  multiplied by the extent) To go from right-handed world space to u,v 
    //  space, negate the v component (which is usually Z in a Y-up world).
    virtual float heightAt( float u, float v ) = 0;
    virtual unsigned int groundAt( float u, float v ) = 0;
    //  issueAll() does culling and LOD for you, given the camera parameters
    virtual void issueAll( Vector3 const & camPos ) = 0;

    virtual unsigned int numChunksU() = 0;
    virtual unsigned int numChunksV() = 0;
    virtual unsigned int chunkWidth() = 0;
    virtual unsigned int chunkHeight() = 0;
    virtual I_TerrainChunk * chunkUV( unsigned int u, unsigned int v ) = 0;

    virtual I_TerrainODE * terrainODE() = 0;
};

//  The input image should be a multiple of 32, plus 3, in each dimension. 
//  It doesn't have to be square. The plus-3 is because of sampling for the 
//  cubic interpolation, taking a 257x257 and replicating edge pixels for 
//  a border to 259x259 is a fine way of creating a suitable image.
//  Scale and offset is for height values -- based on unsigned chars, so 
//  for a value of 255, the output height is 255 * scale + offset.
I_Terrain * NewTerrainFromImage( Image const & i, float scale, float offset );

#endif  //  terrain_h
