
#include "tweakable.h"
#include "log.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>


void doSaveTweakableManager()
{
  TweakableManager::manager().saveAll();
}

TweakableManager::TweakableManager()
{
  files_ = 0;
  atexit( doSaveTweakableManager );
}

void TweakableManager::saveAll()
{
  File * f = files_;

  while( f ) {

    Variable * v = f->vars_;
    char path[ 48 ];
    sprintf( path, "%s.twk", f->name_ );

    f = f->next_;

    FILE * ff = fopen( path, "wb" );
    if( !ff ) {
      error( "cannot create '%s'\n", path );
      continue;
    }

    //  reverse the order to make it stable
    Variable * v2 = NULL;
    while( v ) {
      Variable * tmp = v->next_;
      v->next_ = v2;
      v2 = v;
      v = tmp;
    }
    v = v2;
    while( v ) {
      fprintf( ff, "%s %g\n", v->name_, v->float_ );
      v = v->next_;
    }

    fclose( ff );
  }
}

void TweakableManager::registerFloat( char const * file, char const * name, float def, TweakableFloat * tf )
{
  Variable * v = makeVar( file, name );
  if( v->type_ == Init ) {
    v->type_ = Float;
    v->float_ = def;
  }
  tf->var_ = v;
}

TweakableManager::Variable * TweakableManager::makeVar( char const * file, char const * name )
{
  assert( strlen( file ) < 32 );
  assert( strlen( name ) < 32 );

  File * f = files_;
  while( f ) {
    if( !stricmp( file, f->name_ ) ) {
      goto found_file;
    }
    f = f->next_;
  }
  f = new File;
  strcpy( f->name_, file );
  f->vars_ = 0;
  f->next_ = files_;
  files_ = f;
  readFileVars( f );

found_file:
  Variable * v = f->vars_;
  while( v ) {
    if( !stricmp( name, v->name_ ) ) {
      return v;
    }
    v = v->next_;
  }
  v = new Variable;
  strcpy( v->name_, name );
  v->type_ = Init;
  v->next_ = f->vars_;
  f->vars_ = v;
  return v;
}

void TweakableManager::readFileVars( TweakableManager::File * f )
{
  char path[ 48 ];
  sprintf( path, "%s.twk", f->name_ );
  FILE * ff = fopen( path, "rb" );
  if( !ff ) {
    return;
  }
  char line[ 1024 ];
  int lno = 0;
  while( true ) {
    line[0] = 0;
    fgets( line, 1024, ff );
    if( !line[0] ) {
      break;
    }
    ++lno;
    float fv;
    if( 2 != sscanf( line, "%s %f", path, &fv ) ) {
      error( "Bad line in '%s': %d", path, lno );
    }
    else {
      Variable * v = makeVar( f->name_, path );
      v->type_ = Float;
      v->float_ = fv;
    }
  }
  fclose( ff );
}

TweakableManager & TweakableManager::manager()
{
  static TweakableManager sManager_;
  return sManager_;
}





